#include "ADIDGMethod.h"

namespace MaxwellProblem {

template <int dim>
ADIDGMethod<dim>::ADIDGMethod (
		Triangulation<dim>	&triangulation,
		unsigned int		deg,
		double				start_time,
		double				end_time,
		double				theta
		)
	:
	triangulation(triangulation),
	mapping(),
	degree(deg),
	fe (FESystem<dim>(FE_DGQ<dim>(degree), dim), 1,
		FESystem<dim>(FE_DGQ<dim>(degree), dim), 1),
	dof_handler(triangulation),
	quadrature (degree+1),
	face_quadrature (degree+1),
	assembler(fe, mapping, quadrature, face_quadrature, dof_handler),
	assembler2(fe, mapping, quadrature, face_quadrature, dof_handler),
	start_time(start_time),
	end_time(end_time),
	inhomogeneous(MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution"
		or 
		MaxwellProblem::global::config.get("Data.TestCase","") == "JumpSolution"),
	system_dir(10),
	current_dir(11),
	time_integrator(theta)
{
	/*
	 * reset timers
	 */
	rest_timer.reset();
	assemble_timer.reset();
	setup_timer.reset();

	rest_timer.start();
	std::cout << "Number of active cells:                 "
						<< triangulation.n_active_cells()
						<< std::endl;
	rest_timer.stop();

	/*
	 * set up the system, i.e., initialize dof_handler, matrices etc.
	 */
	setup_timer.start();
	setup_system ();
	setup_timer.stop();

	rest_timer.start();
	std::cout << "Number of degrees of freedom:           "
			<< dof_handler.n_dofs()
			<< std::endl
			<< std::endl;
	rest_timer.stop();

	/*
	 * assemble system matrices
	 */
	assemble_timer.start();
	std::cout << "Starting to assemble system" << std::endl << std::endl;
	assemble_system ();
	std::cout << "Finished assembling system" << std::endl << std::endl;
	assemble_timer.stop();

	rest_timer.start();
	std::cout << "                                        All"<< std::endl;

	std::cout << "Setting up System (CPU) :               " << std::setw(12) << setup_timer() << std::endl;
	std::cout << "Setting up System (wall):               " << std::setw(12) << setup_timer.wall_time() << std::endl << std::endl;

	std::cout << "Assembling System (CPU) :               " << std::setw(12) << assemble_timer() << std::endl;
	std::cout << "Assembling System (wall):               " << std::setw(12) << assemble_timer.wall_time() << std::endl << std::endl;
	rest_timer.stop();
}


template <int dim>
ADIDGMethod<dim>::~ADIDGMethod ()
{
	dof_handler.clear ();
}


template <int dim>
void ADIDGMethod<dim>::setup_system ()
{
	dof_handler.distribute_dofs (fe);

	unsigned int n_dofs = dof_handler.n_dofs();

	/*
	 * The overall system ordering will coincide
	 * with the one for which system_matrix_B has
	 * block-tridiagonal structure. This matrix
	 * corresponds to the negative part of the
	 * Maxwell-operator.
	 * system_matrix_A will internally reorder the
	 * degrees of freedom, when actions are performed
	 * on the state vectors.
	 *
	 * If this is changed, some other stuff has to be
	 * changed for the code to still run in a sensible
	 * way. E.g. the order of assemblage of the matrices
	 * has to be changed, etc.
	 */
	system_dir=1;

	geometry_info.initialize(dof_handler);
	
	//Blocktridiag
	expl_matrix_A.initialize (geometry_info, 0, system_dir);
	impl_matrix_A.initialize (geometry_info, 0, system_dir);
	stiffness_matrix_A.initialize (geometry_info, 0, system_dir);

	expl_matrix_B.initialize (geometry_info, 1, system_dir);
	impl_matrix_B.initialize (geometry_info, 1, system_dir);
	stiffness_matrix_B.initialize (geometry_info, 1, system_dir);
	mass_matrix.initialize(geometry_info, 1, system_dir);
	mass_matrix_inv.initialize(geometry_info, 1, system_dir);

	/*
	 * reinitialize solution vector
	 */
	solution.reinit (n_dofs);
	
	if ( MaxwellProblem::global::config.get("Error.ComputeDerivativeError", false) )
	{
		old_solution.reinit (n_dofs);
		half_step_solution.reinit (n_dofs);
	}

	/*
	 * reinitialize vector containing
	 * the matrix vector product between the
	 * current explicit matrix (i.e. (M + 0.5*timestep*A)
	 * or (M + 0.5*timestep*B)) and the current
	 * solution (in each halfstep)
	 */
	system_rhs.reinit (n_dofs);
	if ( inhomogeneous )
		rhs.reinit (n_dofs);

	/*
	 * reinitialize the vector
	 * representing the transformation
	 * between the two orderings of degrees
	 * of freedom
	 */
	transform.resize(n_dofs);
}


template<int dim>
void ADIDGMethod<dim>::reorder_dofs (
		const unsigned int ADI_dir
		)
{
	if (current_dir != ADI_dir)
	{
		geometry_info.compute_transformation(transform, dof_handler, ADI_dir);
		
// 		for (auto &ii : transform)
// 			std::cout << ii << " ";
// 		std::cout << std::endl << std::endl;

		dof_handler.renumber_dofs(transform);

		current_dir = ADI_dir;
	}
}


// template <int dim>
// void ADIDGMethod<dim>::assemble_system ()
// {
// 	std::cout << "Reordering DoFs A" << std::endl << std::endl;
// 	
// 	reorder_dofs (0);
// 
// 	std::cout << "Assembling Stiffness Matrix A" << std::endl << std::endl;
// 	
// 	assembler.assemble_stiffness_matrix(stiffness_matrix_A);
// 
// 	std::cout << "Reordering DoFs B" << std::endl << std::endl;
// 	
// 	reorder_dofs (1);
// 
// 	std::cout << "Assembling Mass Matrix" << std::endl << std::endl;
// 	
// 	assembler.assemble_mass_matrix(mass_matrix);
// 	
// 	std::cout << "Assembling Stiffness Matrix B" << std::endl << std::endl;
// 	
// 	assembler.assemble_stiffness_matrix(stiffness_matrix_B);
// }


template <int dim>
void ADIDGMethod<dim>::assemble_system ()
{
	std::cout << "Reordering DoFs A" << std::endl << std::endl;
	reorder_dofs (0);

	std::cout << "Reordering DoFs B" << std::endl << std::endl;
	reorder_dofs (1);
	
	std::cout << "Assembling mass and stiffness matrices";
	if ( MaxwellProblem::global::config.get("Error.ComputeDerivativeError", false) )
		std::cout << " and inverting mass matrix";
	
	std::cout << std::endl << "----------------------------------------------------------------" << std::endl;// << std::endl;
	 
	std::vector< std::thread > threads;
	if ( MaxwellProblem::global::config.get("Error.ComputeDerivativeError", false) )
		threads.push_back(std::thread([&]{
			assembler.assemble_mass_matrix(mass_matrix); 
			std::cout << " * Finished assembling mass matrix" << std::endl;// << std::endl;
			std::cout << " * Copying mass matrix into object holding the inverse" << std::endl;// << std::endl;
			mass_matrix_inv.copy_from(mass_matrix);
			std::cout << " * Finished copying mass matrix" << std::endl;// << std::endl;
			std::cout << " * Starting to invert mass matrix" << std::endl;// << std::endl;
			mass_matrix_inv.invert();
			std::cout << " * Finished inverting mass matrix" << std::endl;// << std::endl;
		}));
	else
		threads.push_back(std::thread([&]{
			assembler.assemble_mass_matrix(mass_matrix); 
			std::cout << " * Finished assembling mass matrix" << std::endl;// << std::endl;
		}));
	
	threads.push_back(std::thread([&]{ 
		assembler2.assemble_stiffness_matrix(stiffness_matrix_B); 
		std::cout << " * Finished assembling stiffness matrix" << std::endl;// << std::endl;
		std::cout << " * Copying stiffness Matrix A from B" << std::endl;// << std::endl;
		stiffness_matrix_A.copy_from(-1, stiffness_matrix_B);
		stiffness_matrix_A.set_reordering_flag(true);
		stiffness_matrix_A.set_direction(0);
	}));
		
	for (auto &ii : threads)
		ii.join();
		
// 	for (unsigned int ii = 0; ii < threads.size(); ++ii)
// 		threads[ii].join();
	
	std::cout << "----------------------------------------------------------------" << std::endl;// << std::endl;
}

template <int dim>
void ADIDGMethod<dim>::assemble_time_integrator(const double timestep)
{
	/*
	 * assemble the system matrices
	 * used to propagate in time, i.e.
	 * (M +/- 0.5*timestep*A) and
	 * (M +/- 0.5*timestep*B)
	 * +: explicit
	 * -: implicit
	 *
	 * To save memory, one could save the
	 * stiffness_matrices in the explicit_matrices
	 * and overwrite them after building the
	 * implicit_matrices (or vice versa) and hence
	 * ditch the stiffness_matrix-objects.
	 * The time_integrator class has corresponding
	 * routines, that overwrite the passed object.
	 *
	 * This is, however, not done here for the sake
	 * of readability.
	 */
	time_integrator.assemble_implicit_matrix(impl_matrix_A, stiffness_matrix_A, mass_matrix, transform, timestep);
	time_integrator.assemble_explicit_matrix(expl_matrix_A, stiffness_matrix_A, mass_matrix, transform, timestep);
	time_integrator.assemble_implicit_matrix(impl_matrix_B, stiffness_matrix_B, mass_matrix, transform, timestep);
	time_integrator.assemble_explicit_matrix(expl_matrix_B, stiffness_matrix_B, mass_matrix, transform, timestep);
}


template <int dim>
void ADIDGMethod<dim>::output_data (const unsigned int timestep_number, const Vector<double> &vec, const std::string & identifier) const
{
	/*
	 * save current solution
	 */
	std::stringstream filename;
	filename << "output/results/" << identifier << "_ADI-" << timestep_number << ".vtk";

	std::ofstream vtk_output ((filename.str()).c_str());

	std::vector<std::string> solution_names (dim, "H");
	for (unsigned int i=0; i < dim; ++i)
		solution_names.push_back ("E");
	std::vector<DataComponentInterpretation::DataComponentInterpretation> data_component_interpretation(2*dim, DataComponentInterpretation::component_is_part_of_vector);

	DataOut<dim> data_out;
	data_out.attach_dof_handler (dof_handler);
	data_out.add_data_vector (vec, solution_names, DataOut<dim>::type_dof_data, data_component_interpretation);

	data_out.build_patches (degree);

	data_out.write_vtk(vtk_output);
}


template <int dim>
void ADIDGMethod<dim>::output_results (
		const unsigned int timestep_number
		) const
{
	if ( MaxwellProblem::global::config.get("Output.Solution", false) )
		output_data(timestep_number, solution, "sol");
}


template <int dim>
void ADIDGMethod<dim>::output_exact_solution (
		const unsigned int timestep_number,
		const double current_time
		) const
{
	if ( MaxwellProblem::global::config.get("Output.ExactSolution", false) ) 
	{
		auto ex_sol(solution);
		if ( MaxwellProblem::global::config.get("Data.TestCase","") == "JumpSolution" )
			discretize(JumpSolutionLeft<dim>(current_time), JumpSolutionRight<dim>(current_time), ex_sol);
		else
			discretize(ExactSolution<dim>(current_time), ex_sol);
		output_data(timestep_number, ex_sol, "exsol");
	}
}


template <int dim>
void ADIDGMethod<dim>::output_rhs (const unsigned int timestep_number) const
{
	if(MaxwellProblem::global::config.get("Output.RHS", false))
		output_data(timestep_number, rhs, "rhs");
}


template <int dim>
double ADIDGMethod<dim>::compute_error (const double current_time, const double timestep)
{
	double total_error (0);

	/*
	 * determine exact solution at the current time
	 */
	ExactSolution<dim> exact_solution(current_time);

	const EpsMu<dim> epsmu_function;

	Vector<double> local_errors (triangulation.n_active_cells());

	const QGauss<dim> quadrature_error(degree+2);

	/*
	 * compute L2-error
	 *
	 * Remark: Here, the standard, not the weighted, L2 inner
     * product is computed. This is different in the leapfrog
     * routine and the weighted one would be the preferred one. 
     * However, in the example used in the paper, where eps=mu=1 
     * constant, both inner products are identical and there is
     * no difference. In addition, no error computation was performed
     * for the paper, so even though the weighted version should be
     * (slightly) more costly, the results in the paper are not 
     * polluted by this difference in runtime behavior.
     * Lastly, both inner products are equivalent. So, even if the
     * unweighted version is used, convergence behavior should be
     * identical (the bound can, however, be scaled by a factor).
	 *
	 */
	VectorTools::integrate_difference(mapping, dof_handler, solution, exact_solution,
									  local_errors, quadrature_error, VectorTools::L2_norm, &epsmu_function
									 );

	total_error = local_errors.l2_norm();
	
	
// 	ExactSolution<dim> exact_half_step_solution(current_time - .5 * timestep);
// 
// 	Vector<double> half_step_local_errors (triangulation.n_active_cells());
// 
// 	const QGauss<dim> half_step_quadrature_error(degree+2);
// 	
// 	VectorTools::integrate_difference(mapping, dof_handler, half_step_solution, exact_half_step_solution,
// 									  half_step_local_errors, half_step_quadrature_error, VectorTools::L2_norm);
// 
// 	if (half_step_local_errors.l2_norm() > total_error)
// 		total_error = half_step_local_errors.l2_norm();

	return total_error;
}


template <int dim>
double ADIDGMethod<dim>::compute_error_dt (const double current_time, const double timestep)
{
	double total_error (0);
	double half_time = current_time - .5*timestep;

	/*
	 * determine exact solution at the current time
	 */
	ExactSolutionDt<dim> exact_solution(half_time);


	Vector<double> local_errors (triangulation.n_active_cells());

	const QGauss<dim> quadrature_error(degree+2);
	
	Vector< double > dt_sol(solution);
	dt_sol -= old_solution;
	if (MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution")
		dt_sol /= -timestep;
	else
		dt_sol /= timestep;
	
	VectorTools::integrate_difference(mapping, dof_handler, dt_sol, exact_solution,
									  local_errors, quadrature_error, VectorTools::L2_norm);

	total_error = local_errors.l2_norm();

	return total_error;
}


template <int dim>
double ADIDGMethod<dim>::compute_error_curl (const double current_time, const double timestep)
{
	double total_error (0);
	double half_time = current_time - .5 * timestep;

	/*
	 * determine exact solution at the current time
	 */
	ExactSolutionCurl<dim> exact_solution(half_time);

	Vector<double> local_errors (triangulation.n_active_cells());

	const QGauss<dim> quadrature_error(degree+2);
	
	Vector< double > curl_sol(solution), tmp(half_step_solution);
	curl_sol += old_solution;
	curl_sol *= .5;
	stiffness_matrix_B.vmult(tmp, curl_sol, transform);
	stiffness_matrix_A.vmult(curl_sol, half_step_solution, transform);
	tmp += curl_sol;
	mass_matrix_inv.vmult(curl_sol, tmp, transform);
	
	
	VectorTools::integrate_difference(mapping, dof_handler, curl_sol, exact_solution,
									  local_errors, quadrature_error, VectorTools::L2_norm);

	total_error = local_errors.l2_norm();

	return total_error;
}


template <int dim>
void ADIDGMethod<dim>::discretize (const Function<dim> &function, Vector<double> &vec_out) const
{
// 	ConstraintMatrix constraints;
// 	constraints.close();
	if ( MaxwellProblem::global::config.get("Space_Discretisation.UseProjection",false) )
		VectorTools::project(mapping, dof_handler, ConstraintMatrix{}, quadrature, function, vec_out);
	else 
		VectorTools::interpolate(dof_handler, function, vec_out);
}


template <int dim>
void ADIDGMethod<dim>::discretize (const Function<dim> &function_zero, const Function<dim> &function_one, Vector<double> &vec_out) const
{
		std::map< types::material_id, const Function< dim > * > function_map;
		
		function_map.insert(std::pair< types::material_id, const Function< dim > * > (0, &function_zero) );
		function_map.insert(std::pair< types::material_id, const Function< dim > * > (1, &function_one) );
		
		VectorTools::interpolate_based_on_material_id(mapping, dof_handler, function_map, vec_out);
}


template <int dim>
void ADIDGMethod<dim>::discretize (const std::vector< Function<dim> > &function_vec, Vector<double> &vec_out) const
{
		std::map< types::material_id, const Function< dim > * > function_map;
		
		for (unsigned int ii = 0; ii < function_vec.size(); ++ii)
			function_map.insert(std::pair< types::material_id, const Function< dim > * > (ii, &function_vec[ii]) );
		
		VectorTools::interpolate_based_on_material_id(mapping, dof_handler, function_map, vec_out);
}


template <int dim>
void ADIDGMethod<dim>::discretize_init_val()
{
	Timer gen_timer;
	gen_timer.restart();
	
	if ( MaxwellProblem::global::config.get("Space_Discretisation.UseProjection",false) )
		std::cout << "Projecting initial value onto the grid." << std::endl << std::endl;
	else
		std::cout << "Interpolating initial value onto the grid." << std::endl << std::endl;
	
	if ( MaxwellProblem::global::config.get("Data.TestCase","") == "JumpSolution" )
		discretize(JumpSolutionLeft<dim>(start_time), JumpSolutionRight<dim>(start_time), solution);
	else
		discretize(ExactSolution<dim>(start_time), solution);
	
	std::cout << "Initial value discretized. This took " << gen_timer() << "/" << gen_timer.wall_time() << " seconds (CPU/wall time)." << std::endl << std::endl;
}


template <int dim>
void ADIDGMethod<dim>::run (const unsigned int total_timesteps)
{
	/*
	 * reset timers
	 */
	solve_timer_A.reset();
	solve_timer_B.reset();
	rhs_assemble_timer.reset();

	rest_timer.start();

	/*
	 * setup timesteps
	 */
	double time = start_time;
	unsigned int timestep_number = 0;
	double timestep = (end_time-start_time) / total_timesteps;
	const double timestep_half = 0.5*timestep;

	/*
	 * interpolate the initial value onto
	 * the grid.
	 */
	discretize_init_val ();
	

	output_results (timestep_number);
	output_exact_solution (timestep_number, start_time);
	
	
	double error(0), err_max(-12), err_dt_max(-12), err_curl_max(-12);
	const double resolution = MaxwellProblem::global::config.get("Error.Resolution", 0);
	
	if ( MaxwellProblem::global::config.get<bool>("Error.Error") )
	{
		error = compute_error(time, timestep);

		std::cout << "Discretization error for the initial value: " << error << std::endl << std::endl;

		err_max = error;
	}

	/*
	 * interpolate/project the right-hand side onto
	 * the grid.
	 */
		
	if( inhomogeneous )
	{
		if ( MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution" )
			discretize(ADIPolRHS<dim>(start_time), rhs);
		else if ( MaxwellProblem::global::config.get("Data.TestCase","") == "JumpSolution" )
			discretize(ADIJumpRHSLeft<dim>(start_time), ADIJumpRHSRight<dim>(start_time), rhs);
		
		output_rhs(timestep_number);
	
		if ( MaxwellProblem::global::config.get("Time_Integration.RHS_Version","1") == "1" )
		{
			rhs *= timestep_half;
			std::cout << "Using -=!" << std::endl << std::endl;
		}
		else
			std::cout << "Using sadd!" << std::endl << std::endl;
	}

	/*
	 * assemble system matrices
	 */
	std::cout << "Assembling the time integrator." << std::endl << std::endl;
	assemble_time_integrator(timestep);

	/*
	 * invert the implicit system matrices
	 */
	std::cout << "Computing inverses of the implicit system matrices." << std::endl;
	
	std::cout << "----------------------------------------------------------------" << std::endl;// << std::endl;

	rest_timer.stop();
	
	{
	
	std::vector< std::thread > threads;
	
	threads.push_back(std::thread([&]{ 
		solve_timer_A.start();
		impl_matrix_A.invert();
		solve_timer_A.stop();
		std::cout << " * Inverse of A computed. This took " << solve_timer_A() << "/" << solve_timer_A.wall_time() << " seconds (CPU/wall time)." << std::endl;// << std::endl;
	}));
	
	threads.push_back(std::thread([&]{ 
		solve_timer_B.start();
		impl_matrix_B.invert();
		solve_timer_B.stop();
		std::cout << " * Inverse of B computed. This took " << solve_timer_B() << "/" << solve_timer_B.wall_time() << " seconds (CPU/wall time)." << std::endl;// << std::endl;
	}));
		
	for (auto &ii : threads)
		ii.join();
	
	}
	
	
	std::cout << "----------------------------------------------------------------" << std::endl;// << std::endl;
		
// 	for (unsigned int ii = 0; ii < threads.size(); ++ii)
// 		threads[ii].join();
	
	/*
	 * time integration loop
	 */
	std::cout << std::endl << std::endl << std::endl << "Starting time integration with " << total_timesteps << " timesteps." << std::endl;
	
	unsigned int nn = 1;
	
	bool error_computed = false;

	rest_timer.start();

	for(timestep_number=1, time=start_time + timestep; timestep_number <= total_timesteps; time += timestep, ++timestep_number)
	{
		if ( MaxwellProblem::global::config.get("Error.ComputeDerivativeError", false) && MaxwellProblem::global::config.get("Data.TestCase", "") != "JumpSolution" && ( resolution==0 || fabs(nn*resolution - time) < timestep_half ) )
			old_solution = solution;
		
		rest_timer.stop();
		
		/*
		 * compute b[timestep,1] = (M + 0.5*timestep*B) * solution[timestep-1]
		 */
		rhs_assemble_timer.start();
		expl_matrix_B.vmult(system_rhs, solution, transform);
		rhs_assemble_timer.stop();

		/*
		 * solve (M - 0.5*timestep*A) * solution[timestep-1/2] = b[timestep,1]
		 */
		solve_timer_A.start();
		impl_matrix_A.inverse_vmult(solution, system_rhs, transform);
		solve_timer_A.stop();
		
		if ( inhomogeneous )
		{
			if ( MaxwellProblem::global::config.get("Time_Integration.RHS_Version","1") == "1" )
				solution -= rhs;
			else
				solution.sadd(1., -0.5*timestep, rhs);
		}
		
		if ( MaxwellProblem::global::config.get("Error.ComputeDerivativeError", false) && ( resolution==0 || fabs(nn*resolution - time) < timestep_half ) )
			half_step_solution = solution;
		
		
		if ( inhomogeneous )
		{
			if ( MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution" )
				discretize(ADIPolRHS<dim>(time), rhs);
			else if ( MaxwellProblem::global::config.get("Data.TestCase","") == "JumpSolution" )
				discretize(ADIJumpRHSLeft<dim>(time), ADIJumpRHSRight<dim>(time), rhs);
				
			output_rhs(timestep_number);
			
			if ( MaxwellProblem::global::config.get("Time_Integration.RHS_Version","1") == "1" )
			{
				rhs *= timestep_half;
			
				solution -= rhs;
			}
			else
				solution.sadd(1., -0.5*timestep, rhs);
		}

		/*
		 * compute b[timestep,2] = (M + 0.5*timestep*A) * solution[timestep-1/2]
		 */
		rhs_assemble_timer.start();
		expl_matrix_A.vmult(system_rhs, solution, transform);
		rhs_assemble_timer.stop();

		/*
		 * solve (M - 0.5*timestep*B) * solution[timestep] = b[timestep,2]
		 */
		solve_timer_B.start();
		impl_matrix_B.inverse_vmult(solution, system_rhs, transform);
		solve_timer_B.stop();

		rest_timer.start();
		output_results (timestep_number);
		output_exact_solution (timestep_number, time);

		if ( MaxwellProblem::global::config.get<bool>("Error.Error")  && ( resolution==0 || fabs(nn*resolution - time) < timestep_half ) )
		{
			error = compute_error(time, timestep);

			std::cout << "sol \t-- " << time << " -- " << error << std::endl;

			if (err_max < error)
				err_max = error;

			std::cout << time << " -- current err_max:\t" << err_max << std::endl;
			
			
			error_computed = true;
		}
	
		if ( MaxwellProblem::global::config.get("Error.ComputeDerivativeError", false) && ( resolution==0 || fabs(nn*resolution - time) < timestep_half ) )
		{
			
			error = compute_error_curl(time, timestep);
			
			std::cout << "curl \t-- " << time << " -- " << error << std::endl;

			if (err_curl_max < error)
				err_curl_max = error;

			std::cout << time << " -- current err_curl_max:\t" << err_curl_max << std::endl;
			
			
			error = compute_error_dt(time, timestep);

			std::cout << "dt \t-- " << time << " -- " << error << std::endl;

			if (err_dt_max < error)
				err_dt_max = error;

			std::cout << time << " -- current err_dt_max:\t" << err_dt_max << std::endl;
			
			
			error_computed = true;
		}
		
		if (error_computed)
		{
			error_computed = false;
			++nn;
		}
	}


	if (MaxwellProblem::global::config.get<bool>("Error.Error"))
	{
		std::ofstream out_err;
		out_err.open("output/error/error_ADI.txt", std::ios_base::app);
		out_err.setf( std::ios::left );

		error = compute_error(end_time, timestep);

		if (err_max < error)
			err_max = error;

		out_err << degree << "\t" << std::setw(12) << GridTools::maximal_cell_diameter(triangulation) << "\t" << std::setw(12) << timestep << "\t" << std::setw(12) << err_max << std::endl;
		out_err.close();
	}


	if ( MaxwellProblem::global::config.get("Error.ComputeDerivativeError", false) )
	{
		{
		std::ofstream out_err;
		out_err.open("output/error/error_ADI_curl.txt", std::ios_base::app);
		out_err.setf( std::ios::left );

		error = compute_error_curl(end_time, timestep);

		if (err_curl_max < error)
			err_curl_max = error;

		out_err << degree << "\t" << std::setw(12) << GridTools::maximal_cell_diameter(triangulation) << "\t" << std::setw(12) << timestep << "\t" << std::setw(12) << err_curl_max << std::endl;
		out_err.close();
		}{
		std::ofstream out_err;
		out_err.open("output/error/error_ADI_dt.txt", std::ios_base::app);
		out_err.setf( std::ios::left );

		error = compute_error_dt(end_time, timestep);

		if (err_dt_max < error)
			err_dt_max = error;

		out_err << degree << "\t" << std::setw(12) << GridTools::maximal_cell_diameter(triangulation) << "\t" << std::setw(12) << timestep << "\t" << std::setw(12) << err_dt_max << std::endl;
		out_err.close();
		}
	}

	rest_timer.stop();

	std::cout << std::endl << std::endl;

	std::cout << "                                        " << std::setw(12) << "All" << "\t" << std::setw(12) << "Mean"<< std::endl;

	std::cout << "Solving A-Systems (CPU) :               " << std::setw(12) << solve_timer_A() << "\t" << std::setw(12) << solve_timer_A()/total_timesteps << std::endl;
	std::cout << "Solving A-Systems (wall):               " << std::setw(12) << solve_timer_A.wall_time() << "\t" << std::setw(12) << solve_timer_A.wall_time()/total_timesteps << std::endl << std::endl;

	std::cout << "Solving B-Systems (CPU) :               " << std::setw(12) << solve_timer_B() << "\t" << std::setw(12) << solve_timer_B()/total_timesteps << std::endl;
	std::cout << "Solving B-Systems (wall):               " << std::setw(12) << solve_timer_B.wall_time() << "\t" << std::setw(12) << solve_timer_B.wall_time()/total_timesteps << std::endl << std::endl;

	std::cout << "Assembling system-rhs (CPU) :           " << std::setw(12) << rhs_assemble_timer() << "\t" << std::setw(12) << rhs_assemble_timer()/total_timesteps << std::endl;
	std::cout << "Assembling system-rhs (wall):           " << std::setw(12) << rhs_assemble_timer.wall_time() << "\t" << std::setw(12) << rhs_assemble_timer.wall_time()/total_timesteps << std::endl << std::endl;

	std::cout << "Overall Solving (CPU) :                 " << std::setw(12) << rhs_assemble_timer()+solve_timer_B()+solve_timer_A() << "\t" << std::setw(12) << (rhs_assemble_timer()+solve_timer_B()+solve_timer_A())/total_timesteps << std::endl;
	std::cout << "Overall Solving (wall):                 " << std::setw(12) << rhs_assemble_timer.wall_time()+solve_timer_B.wall_time()+solve_timer_A.wall_time() << "\t" << std::setw(12) << (rhs_assemble_timer.wall_time()+solve_timer_B.wall_time()+solve_timer_A.wall_time())/total_timesteps << std::endl << std::endl;

	std::cout << "All the rest (CPU) :                    " << std::setw(12) << rest_timer() << std::endl;
	std::cout << "All the rest (wall):                    " << std::setw(12) << rest_timer.wall_time() << std::endl << std::endl;

	if (MaxwellProblem::global::config.get("Output.Profiling", false))
	{
		std::stringstream filename;

		filename << "output/runtimes/ADI_solving_time.txt";

		std::ofstream out ((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer_A() << "\t" << solve_timer_B() << "\t" << solve_timer_A()+solve_timer_B() << std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_solving_time_wall_time.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer_A.wall_time() << "\t" << solve_timer_B.wall_time() << "\t" << solve_timer_A.wall_time()+solve_timer_B.wall_time() << std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_solving_time_mean.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer_A()/total_timesteps << "\t" << solve_timer_B()/total_timesteps << "\t" << solve_timer_A()/total_timesteps+solve_timer_B()/total_timesteps << std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_solving_time_mean_wall_time.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer_A.wall_time()/total_timesteps << "\t" << solve_timer_B.wall_time()/total_timesteps << "\t" << solve_timer_A.wall_time()/total_timesteps+solve_timer_B.wall_time()/total_timesteps << std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_rhs_and_solving_time.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer_A()+solve_timer_B()+rhs_assemble_timer() << std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_rhs_and_solving_time_wall_time.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer_A.wall_time()+solve_timer_B.wall_time()+rhs_assemble_timer.wall_time() << std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_rhs_and_solving_time_mean.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer_A()/total_timesteps+solve_timer_B()/total_timesteps+rhs_assemble_timer()/total_timesteps << std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_rhs_and_solving_time_mean_wall_time.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer_A.wall_time()/total_timesteps+solve_timer_B.wall_time()/total_timesteps+rhs_assemble_timer.wall_time()/total_timesteps << std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_rhs_assembling_time.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << rhs_assemble_timer() << std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_rhs_assembling_time_wall_time.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << rhs_assemble_timer.wall_time() <<  std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_rhs_assembling_time_mean.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << rhs_assemble_timer()/total_timesteps << std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_rhs_assembling_time_mean_wall_time.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t" << rhs_assemble_timer.wall_time()/total_timesteps << std::endl;

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_overall_time_wall_time.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t";

		out.close();

		filename.str("");
		filename.clear();

		filename << "output/runtimes/ADI_overall_time.txt";

		out.open((filename.str()).c_str(), std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << dof_handler.n_dofs() << "\t" << timestep << "\t";

		out.close();
	}
}

template class ADIDGMethod<3>;

} /* namespace MaxwellProblem */
