#ifndef ADIDGMETHOD_H_
#define ADIDGMETHOD_H_

#include <iostream>
#include <fstream>
#include <sstream>


#include <deal.II/grid/tria.h>
#include <deal.II/grid/grid_generator.h>
#include <deal.II/grid/grid_out.h>
#include <deal.II/grid/grid_tools.h>
#include <deal.II/grid/tria_accessor.h>
#include <deal.II/grid/tria_iterator.h>
#include <deal.II/fe/fe_values.h>
#include <deal.II/fe/fe_system.h>
#include <deal.II/fe/mapping_q1.h>
#include <deal.II/fe/fe_dgq.h>
#include <deal.II/dofs/dof_handler.h>
#include <deal.II/dofs/dof_accessor.h>
#include <deal.II/dofs/dof_tools.h>
#include <deal.II/dofs/dof_renumbering.h>

#include <deal.II/numerics/data_out.h>
#include <deal.II/numerics/vector_tools.h>
#include <deal.II/numerics/matrix_tools.h>

#include <deal.II/base/timer.h>

#include "assembling/ADIDGMaxwellGlobal.h"

#include "reordering/ADIGeometryInfo.h"
#include "reordering/Compare.h"
#include "data_types/ADIMatrix.h"
#include "data_types/ADIMassMatrix.h"
#include "time_integration/PeacemanRachford.h"
#include "data/ExactSolution.h"
// #include "data/ExactSolutionDt.h"
// #include "data/ExactSolutionCurl.h"
#include "data/EpsMu.h"
#include "data/PolRHS.h"
#include "data/JumpRHS.h"
#include "data/GridGeneration.cpp"
#include "global.h"

namespace MaxwellProblem {

using namespace dealii;
using namespace TimeIntegration;
using namespace Assembling;
using namespace Data;
using namespace global;

template <int dim>
class ADIDGMethod
{
public:

	/*
	 * constructor
	 *
	 * triangulation: triangulation of the domain
	 * deg: polynomial degree of the approximation
	 * start_time: start time of the approximation
	 * end_time: end time of the approximation
	 * theta: determines, on which time integration
	 * scheme the ADI-scheme is based. 0.5 equals
	 * the Peaceman-Rachford scheme, which is based
	 * on the Crank-Nicolson scheme.
	 */
	ADIDGMethod							(Triangulation<dim>	&triangulation,
										 unsigned int		deg,
										 double				start_time,
										 double				end_time,
								 	 	 double				theta = 0.5);

	/*
	 * destructor
	 */
	~ADIDGMethod						();

	/*
	 * main routine, performs simulation
	 *
	 * total_timesteps : total number of timesteps that should be used
	 */
	void run							(const unsigned int total_timesteps);

private:

	/*
	 * initializes matrices and vectors
	 */
	void setup_system 					();

	/*
	 * assembles the mass and stiffness matrices
	 */
	void assemble_system				();

	/*
	 * assembles the system matrices, i.e.
	 * the discretizations of
	 * (I +/- 0.5*timestep*A/B)
	 *
	 * timestep: timestep used for the scheme
	 */
	void assemble_time_integrator		(const double timestep);

	/*
	 * puts out the results of the simulation
	 */
	void output_results					(const unsigned int	timestep_number) const;
	void output_exact_solution			(const unsigned int	timestep_number, const double current_time) const;
	void output_rhs						(const unsigned int	timestep_number) const;
	
	void output_data					(const unsigned int	timestep_number, const Vector<double> &vec, const std::string &identifier) const;

	/*
	 * computes error to the given exact solution
	 *
	 * time: determines the point of simulation time at which
	 * the error should be computed
	 *
	 */
	double compute_error				(const double current_time,
										 const double timestep);
	double compute_error_curl			(const double current_time,
										 const double timestep);
	double compute_error_dt				(const double current_time,
										 const double timestep);
	
	
	void discretize						(const Function<dim> &function, Vector<double> &vec_out) const;
	void discretize						(const Function<dim> &function_zero, const Function<dim> &function_one, Vector<double> &vec_out) const;
	void discretize						(const std::vector< Function<dim> > &function_vec, Vector<double> &vec_out) const;
	void discretize_init_val			();

	/*
	 * reorders the global degrees of freedom
	 * this is needed to obtain the block-tridiagonal
	 * structure
	 */
	void reorder_dofs					(const unsigned int	normal_dir);


	/*
	 * triangulation
	 */
	Triangulation<dim>														&triangulation;

	/*
	 * mapping used for transformation from reference cell to general mesh cells
	 */
	const MappingQ1<dim>													mapping;

	/*
	 * polynomial degree of the finite elements
	 */
	const unsigned int														degree;

	/*
	 * finite element that is used
	 */
	FESystem<dim>															fe;

	/*
	 * DoFHandler, dealing with the degrees of freedom
	 */
	DoFHandler<dim>															dof_handler;


	/*
	 * cell quadrature formula
	 */
	const QGauss<dim>														quadrature;

	/*
	 * face quadrature formula
	 */
	const QGauss<dim-1>														face_quadrature;

	/*
	 * object assembling the matrices
	 */
	ADIDGMaxwellGlobal<dim>													assembler, assembler2;

	/**
	 * mass matrix
	 */
	ADIMassMatrix<dim> 														mass_matrix, mass_matrix_inv;

	/*
	 * stiffness matrices
	 */
	ADIMatrix<dim>															expl_matrix_A,
																			impl_matrix_A,
																			expl_matrix_B,
																			impl_matrix_B,
																			stiffness_matrix_A,
																			stiffness_matrix_B;

	/*
	 * vector storing the system rhs
	 * i.e. (I + 0.5*timestep*A/B)*solution
	 */
	Vector<double>															system_rhs;
	
	Vector<double>															rhs;

	/*
	 * solution vector
	 */
	Vector<double>															solution,
                                                                            half_step_solution,
                                                                            old_solution;

	/*
	 * start and endtime of the simulation
	 */
	const double															start_time,
																			end_time;
																			
	const bool																inhomogeneous;

	/*
	 * current_dir determines which direction the degrees
	 * of freedom are currently ordered
	 * system_dir determines with which direction all computations
	 * will be performed (reordering will be performed by the matrix
	 * classes)
	 */
	unsigned int															system_dir,
																			current_dir;
	/*
	 * vector storing the transformation between the two directions
	 * used to get A, resp. B into block-tridiagonal form
	 */
	std::vector<unsigned int>												transform;

	/*
	 * object storing geometrical info, which is used to determine
	 * the transformation between the formerly mentioned directions
	 */
	ADIGeometryInfo< dim >													geometry_info;

	/*
	 * time_integration object, mainly used to set up the system matrices.
	 */
	const PeacemanRachford<dim>												time_integrator;
};

} /* namespace MaxwellProblem */

#endif /* ADIDGMETHOD_H_ */
