#include "DGMethod.h"

namespace MaxwellProblem {

template <int dim>
DGMethod<dim>::DGMethod (
	Triangulation<dim>	&triangulation,
	unsigned int		deg,
	double				start_time,
	double				end_time)
	:
	triangulation(triangulation),
	mapping(),
	degree(deg),
	fe (FE_DGQ<dim>(degree), dim),
	dof_handler(triangulation),
	quadrature(degree+1),
	face_quadrature(degree+1),
	assembler(fe, mapping, quadrature, face_quadrature, dof_handler),
	start_time(start_time),
	end_time(end_time)
{}


template <int dim>
DGMethod<dim>::~DGMethod ()
{
	dof_handler.clear ();
}


template <int dim>
void DGMethod<dim>::setup_and_assemble_system ()
{
	/*
	 * reset timers
	 */
	setup_timer.reset();
	assemble_timer.reset();
	rest_timer.reset();

	rest_timer.start();
	std::cout << "Number of active cells:                 "
						<< triangulation.n_active_cells()
						<< std::endl;
	rest_timer.stop();

	/*
	 * set up the system, i.e., initialize dof_handler, matrices etc.
	 */
	setup_timer.start();
	setup_system ();
	setup_timer.stop();

	rest_timer.start();
	std::cout << "Number of degrees of freedom:           "
			<< 2 * dof_handler.n_dofs()
			<< std::endl << std::endl;
	rest_timer.stop();

	/*
	 * assemble system matrices
	 */
	assemble_timer.start();
	assemble_system ();
	assemble_timer.stop();

	rest_timer.start();
	std::cout << "                                        All"<< std::endl;

	std::cout << "Setting up System (CPU) :               " << std::setw(12) << setup_timer() << std::endl;
	std::cout << "Setting up System (wall):               " << std::setw(12) << setup_timer.wall_time() << std::endl << std::endl;

	std::cout << "Assembling System (CPU) :               " << std::setw(12) << assemble_timer() << std::endl;
	std::cout << "Assembling System (wall):               " << std::setw(12) << assemble_timer.wall_time() << std::endl << std::endl;
	rest_timer.stop();
}


template <int dim>
void DGMethod<dim>::setup_system ()
{
	dof_handler.distribute_dofs(fe);

    const unsigned int n_h = dof_handler.n_dofs();
    DynamicSparsityPattern dsp(n_h);

    /*
     * mass matrices:
     * create sparsity patterns and reinitialize
     * matrices
     */
	Table<2,DoFTools::Coupling> cell_coupling_M(fe.n_components(), fe.n_components());
	cell_coupling_M[0][0] = DoFTools::always;
	cell_coupling_M[1][1] = DoFTools::always;
	cell_coupling_M[2][2] = DoFTools::always;
	DoFTools::make_sparsity_pattern(dof_handler, cell_coupling_M, dsp);
	sparsity_pattern_M.copy_from(dsp);

	M_HH.reinit(sparsity_pattern_M);
	M_EE.reinit(sparsity_pattern_M);
	M_HH_inv.reinit(sparsity_pattern_M);
	M_EE_inv.reinit(sparsity_pattern_M);

    /*
     * curl/stiffness matrices:
     * create sparsity patterns and reinitialize
     * matrices
     */
	Table<2,DoFTools::Coupling> cell_coupling_C (fe.n_components(), fe.n_components());
	cell_coupling_C[0][1] = DoFTools::always;
	cell_coupling_C[0][2] = DoFTools::always;
	cell_coupling_C[1][0] = DoFTools::always;
	cell_coupling_C[1][2] = DoFTools::always;
	cell_coupling_C[2][0] = DoFTools::always;
	cell_coupling_C[2][1] = DoFTools::always;
	Table<2,DoFTools::Coupling> face_coupling_C (fe.n_components(), fe.n_components());
	face_coupling_C[0][1] = DoFTools::nonzero;
	face_coupling_C[0][2] = DoFTools::nonzero;
	face_coupling_C[1][0] = DoFTools::nonzero;
	face_coupling_C[1][2] = DoFTools::nonzero;
	face_coupling_C[2][0] = DoFTools::nonzero;
	face_coupling_C[2][1] = DoFTools::nonzero;
	dsp.reinit(n_h,n_h);
	DoFTools::make_flux_sparsity_pattern(dof_handler, dsp, cell_coupling_C, face_coupling_C);
	sparsity_pattern_C.copy_from(dsp);

	C_H.reinit(sparsity_pattern_C);
	C_E.reinit(sparsity_pattern_C);

	/*
	 * reinitialize solution vectors
	 */
	solution_H.reinit(n_h);
	solution_E.reinit(n_h);
	
	rhs_J.reinit(n_h);
}

template <int dim>
void DGMethod<dim>::assemble_system ()
{
	/*
	 * assemble mass and curl/stiffness matrices
	 */
	assembler.assemble_mass_matrix(M_HH, M_EE, M_HH_inv, M_EE_inv);
	assembler.assemble_discrete_curl_matrices(C_H, C_E);
}


template <int dim>
void DGMethod<dim>::run (const unsigned int total_timesteps)
{
	/*
	 * reset timers
	 */
	solve_timer.reset();

	rest_timer.start();

	/*
	 * setup timesteps
	 */
	const double timestep = (end_time-start_time) / total_timesteps;
	const double timestep_half = 0.5*timestep;
	double time = start_time;
	unsigned int timestep_number = 0;
	
	Vector<double> tmp_vec(solution_E.size()), tmp_vec2(solution_E.size());

	/*
	 * interpolate the initial value onto
	 * the grid.
	 */
	VectorTools::interpolate(dof_handler, ExactSolution_H<dim>(time), solution_H);
	VectorTools::interpolate(dof_handler, ExactSolution_E<dim>(time), solution_E);
	
	
// 	ConstraintMatrix constraints;
// 	constraints.close();
	
	if ( MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution" )
	{
// 		VectorTools::project(mapping, dof_handler, constraints, quadrature, PolRHS<dim>(time), tmp_vec);
		VectorTools::interpolate(dof_handler, PolRHS<dim>(time), tmp_vec2);
// 		// TODO delete
// 		output_rhs(timestep_number, tmp_vec2);
// 		// TODO delete
		M_EE.vmult(tmp_vec, tmp_vec2);
	}


	if(MaxwellProblem::global::config.get<bool>("Output.Solution"))
		output_results (timestep_number);


	/*
	 * initialize time integration
	 */
	Leapfrog<SparseMatrix<double>,SparseMatrix<double>> time_integrator(M_HH_inv, M_EE_inv, C_H, C_E, timestep);
	
	
	double error(0), err_max(-12), err_dt_max(-12), err_curl_max(-12), resolution(0.1);
	
	unsigned int nn = 1;
	
	bool error_computed = false;
	
	if ( MaxwellProblem::global::config.get<bool>("Error.Error") )
	{
		error = compute_error(time);

		std::cout << "sol \t-- " << time << " -- " << error << std::endl;

		err_max = error;
		
		std::cout << time << " -- current err_max:\t" << err_max << std::endl;
	}


	/*
	 * time integration loop
	 */
	std::cout << std::endl << std::endl << std::endl << "Starting time integration with " << total_timesteps << " time steps." << std::endl;

	for(timestep_number=1, time=start_time + timestep; timestep_number <= total_timesteps; time += timestep, ++timestep_number)
	{
		rest_timer.stop();
		
		if ( MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution" )
		{
			rhs_J.swap(tmp_vec);
			VectorTools::interpolate(dof_handler, PolRHS<dim>(time), tmp_vec2);
// 			// TODO delete
// 			output_rhs(timestep_number, tmp_vec2);
// 			// TODO delete
// 			VectorTools::project(mapping, dof_handler, constraints, quadrature, PolRHS<dim>(time), tmp_vec);
			M_EE.vmult(tmp_vec, tmp_vec2);
			rhs_J.sadd(0.5, 0.5, tmp_vec);
		}

		// compute one step of time integration
		solve_timer.start();
		if ( MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution" )
			time_integrator.integrate_step(solution_H, solution_E, rhs_J);
		else
			time_integrator.integrate_step(solution_H, solution_E);
		solve_timer.stop();

		rest_timer.start();
		if(MaxwellProblem::global::config.get<bool>("Output.Solution"))
			output_results (timestep_number);
		

		if ( MaxwellProblem::global::config.get<bool>("Error.Error")  && ( resolution==0 || fabs(nn*resolution - time) < timestep_half ) )
		{
			error = compute_error(time);

			std::cout << "sol \t-- " << time << " -- " << error << std::endl;

			if (err_max < error)
				err_max = error;

			std::cout << time << " -- current err_max:\t" << err_max << std::endl;
			
			error_computed = true;
		}
		
		if (error_computed)
		{
			error_computed = false;
			++nn;
		}
	}


	if (MaxwellProblem::global::config.get<bool>("Error.Error"))
	{
		std::ofstream out_err;
		out_err.open("output/error/error_std.txt", std::ios_base::app);
		out_err.setf( std::ios::left );

		error = compute_error(end_time);

		if (err_max < error)
			err_max = error;

		out_err << degree << "\t" << std::setw(12) << GridTools::maximal_cell_diameter(triangulation) << "\t" << std::setw(12) << timestep << "\t" << std::setw(12) << err_max << std::endl;
		out_err.close();
	}


	std::cout << std::endl << std::endl << std::endl;;

	std::cout << "                                        " << std::setw(12) << "All" << "\t" << std::setw(12) << "Mean"<< std::endl;

	std::cout << "Solving Systems (CPU) :                 " << std::setw(12) << solve_timer() << "\t" << std::setw(12) << solve_timer()/total_timesteps << std::endl;
	std::cout << "Solving Systems (wall):                 " << std::setw(12) << solve_timer.wall_time() << "\t" << std::setw(12) << solve_timer.wall_time()/total_timesteps << std::endl << std::endl;

	std::cout << "All the rest (CPU) :                    " << std::setw(12) << rest_timer() << std::endl;
	std::cout << "All the rest (wall):                    " << std::setw(12) << rest_timer.wall_time() << std::endl << std::endl;

	rest_timer.stop();

	if (MaxwellProblem::global::config.get<bool>("Output.Profiling"))
	{
		std::ofstream out ("output/runtimes/solving_time.txt", std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << 2 * dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer() << std::endl;

		out.close();

		out.open("output/runtimes/solving_time_wall_time.txt", std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << 2 * dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer.wall_time() <<  std::endl;

		out.close();

		out.open("output/runtimes/solving_time_mean.txt", std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << 2 * dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer()/total_timesteps << std::endl;

		out.close();

		out.open("output/runtimes/solving_time_mean_wall_time.txt", std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << 2 * dof_handler.n_dofs() << "\t" << timestep << "\t" << solve_timer.wall_time()/total_timesteps << std::endl;

		out.close();

		out.open("output/runtimes/overall_time_wall_time.txt", std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << 2 * dof_handler.n_dofs() << "\t" << timestep << "\t";

		out.close();

		out.open("output/runtimes/overall_time.txt", std::ios_base::app);

		out << triangulation.n_active_cells() << "\t" << 2 * dof_handler.n_dofs() << "\t" << timestep << "\t";

		out.close();
	}
}

template <int dim>
double DGMethod<dim>::compute_error (double time)
{
	double total_error (0);

	/*
	 * determine exact solution at the current time
	 */
	ExactSolution_H<dim> exsol_H_func(time);
	ExactSolution_E<dim> exsol_E_func(time);

	/*
	 * determine material parameters (used
	 * for weighted L2-product
	 */
	const Mu<dim> mu_function;
	const Eps<dim> eps_function;


	Vector<double> local_errors (triangulation.n_active_cells());

	const QGauss<dim> quadrature_formula(degree+2);


	/*
	 * compute L2-error
	 */
	VectorTools::integrate_difference(mapping, dof_handler, solution_H, exsol_H_func,
									  local_errors, quadrature_formula, VectorTools::L2_norm, &mu_function);
	total_error = local_errors.l2_norm();

	VectorTools::integrate_difference(mapping, dof_handler, solution_E, exsol_E_func,
									  local_errors, quadrature_formula, VectorTools::L2_norm, &eps_function);
	total_error = std::sqrt(std::pow(total_error,2) + std::pow(local_errors.l2_norm(),2));

	return total_error;
}


template <int dim>
void DGMethod<dim>::output_data (const unsigned int timestep_number, const Vector<double> &vec_H, const Vector<double> &vec_E, const std::string & identifier) const
{
	/*
	 * save current solution
	 */
	std::stringstream filename;
	filename << "output/results/" << identifier << "-" << timestep_number << ".vtk";

	std::ofstream vtk_output ((filename.str()).c_str());

	std::vector<DataComponentInterpretation::DataComponentInterpretation> data_component_interpretation(dim, DataComponentInterpretation::component_is_part_of_vector);

	DataOut<dim> data_out;
	data_out.attach_dof_handler (dof_handler);
	std::vector<std::string> solution_names (dim, "H");
	data_out.add_data_vector (vec_H, solution_names, DataOut<dim>::type_dof_data, data_component_interpretation);
	solution_names.assign (dim, "E");
	data_out.add_data_vector (vec_E, solution_names, DataOut<dim>::type_dof_data, data_component_interpretation);
	data_out.build_patches (degree);

	data_out.write_vtk(vtk_output);
}


template <int dim>
void DGMethod<dim>::output_results (const unsigned int timestep_number) const
{
	output_data(timestep_number, solution_H, solution_E, "sol");
}


template <int dim>
void DGMethod<dim>::output_rhs (const unsigned int timestep_number, const Vector<double> &vec_E) const
{
	output_data(timestep_number, Vector<double>(solution_H.size()), vec_E, "rhs");
}

template class DGMethod<3>;

} /* namespace MaxwellProblem */
