#ifndef DGMETHOD_H_
#define DGMETHOD_H_

#include <iostream>
#include <fstream>

#include <list>
#include <boost/lexical_cast.hpp>

#include <deal.II/grid/tria.h>
#include <deal.II/grid/grid_generator.h>
#include <deal.II/grid/grid_out.h>
#include <deal.II/grid/tria_accessor.h>
#include <deal.II/grid/tria_iterator.h>
#include <deal.II/fe/fe_values.h>
#include <deal.II/fe/fe_system.h>
#include <deal.II/fe/mapping_q1.h>
#include <deal.II/fe/fe_dgq.h>
#include <deal.II/dofs/dof_handler.h>
#include <deal.II/dofs/dof_accessor.h>
#include <deal.II/dofs/dof_tools.h>
#include <deal.II/dofs/dof_renumbering.h>

#include <deal.II/numerics/data_out.h>

#include <deal.II/lac/sparse_matrix.h>
#include <deal.II/lac/block_vector.h>
#include <deal.II/numerics/vector_tools.h>
#include <deal.II/numerics/matrix_tools.h>
#include <deal.II/grid/grid_tools.h>

#include <deal.II/base/timer.h>

#include "assembling/DGMaxwellGlobal.h"

#include "time_integration/Leapfrog.h"
#include "data/GridGeneration.cpp"
#include "data/ExactSolution.h"
#include "data/GradientField.h"
#include "data/Eps.h"
#include "data/Mu.h"
#include "data/PolRHS.h"
#include "global.h"

//#include "data_types/DG_MassMatrix.h"


namespace MaxwellProblem {

using namespace dealii;
using namespace TimeIntegration;
using namespace Assembling;
using namespace Data;
//using namespace DataTypes;
using namespace global;


template <int dim>
class DGMethod
{
public:

	/*
	 * constructor
	 *
	 * triangulation : triangulation of the domain
	 * deg : polynomial degree of the approximation
	 * start_time : start time of the approximation
	 * end_time : end time of the approximation
	 */
	DGMethod						(Triangulation<dim>	&triangulation,
									 unsigned int		deg,
									 double				start_time,
									 double				end_time);

	/*
	 * destructor
	 */
	~DGMethod						();

	/*
	 * main routine, performs simulation
	 *
	 * total_timesteps : total number of timesteps that should be used
	 */
	void run						(const unsigned int total_timesteps);

	/*
	 * setup of data structures and assembling of the mass and stiffness matrices.
	 */
	void setup_and_assemble_system	();




private:

	/*
	 * initializes matrices and vectors
	 */
	void setup_system 				();

	/*
	 * assembles the mass and stiffness matrices
	 */
	void assemble_system			();

	/*
	 * puts out the results of the simulation
	 */
	void output_results				(const unsigned int	timestep_number) const;
	void output_rhs					(const unsigned int	timestep_number, const Vector<double> &vec_E) const;
	
	void output_data				(const unsigned int	timestep_number, const Vector<double> &vec_H, const Vector<double> &vec_E, const std::string &identifier) const;

	/*
	 * computes error to the given exact solution
	 *
	 * time: determines the point of simulation time at which
	 * the error should be computed
	 *
	 */
	double compute_error			(double time);


	/*
	 * triangulation
	 */
	Triangulation<dim>				&triangulation;

	/*
	 * mapping used for transformation from reference cell to general mesh cells
	 */
	const MappingQ1<dim>			mapping;

	/*
	 * polynomial degree of the finite elements
	 */
	const unsigned int				degree;

	/*
	 * finite element that is used
	 */
	FESystem<dim>					fe;

	/*
	 * DoFHandler, dealing with the degrees of freedom
	 */
	DoFHandler<dim>					dof_handler;

	/*
	 * cell quadrature formula
	 */
	const QGauss<dim>				quadrature;

	/*
	 * face quadrature formula
	 */
	const QGauss<dim-1>				face_quadrature;

	/*
	 * object assembling the matrices
	 */
	DGMaxwellGlobal<dim> 			assembler;

	/*
	 * sparsity patterns of the matrices
	 */
	SparsityPattern					sparsity_pattern_C, sparsity_pattern_M;

	/*
	 * stiffness matrices
	 */
	SparseMatrix<double>			C_H, C_E;

	/**
	 * mass and inverse mass matrices
	 */
	SparseMatrix<double>			M_HH, M_EE, M_HH_inv, M_EE_inv;


	/*
	 * solution vectors
	 */
	Vector<double>					solution_H, solution_E;


	/*
	 * inhomogeneity
	 */
	Vector<double>					rhs_J;

	/*
	 * start and end time of the simulation
	 */
	const double					start_time,
									end_time;
};

} /* namespace MaxwellProblem */

#endif /* DGMETHOD_H_ */
