#include <deal.II/base/timer.h>
#include <deal.II/base/multithread_info.h>
#include <deal.II/grid/tria.h>

#include "global.h"
#include "DGMethod.h"
#include "ADIDGMethod.h"

#include <iostream>
#include <fstream>

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include <boost/foreach.hpp>
namespace pt = boost::property_tree;



void reset_profiling_files()
{
	std::ofstream out ("output/runtimes/overall_time.txt");

	out.close();

	out.open("output/runtimes/overall_time_wall_time.txt");

	out.close();

	out.open("output/runtimes/ADI_overall_time.txt");

	out.close();

	out.open("output/runtimes/ADI_overall_time_wall_time.txt");

	out.close();

	out.open("output/runtimes/solving_time.txt");

	out.close();

	out.open("output/runtimes/solving_time_wall_time.txt");

	out.close();

	out.open("output/runtimes/solving_time_mean.txt");

	out.close();

	out.open("output/runtimes/solving_time_mean_wall_time.txt");

	out.close();

	out.open("output/runtimes/overall_time_wall_time.txt");

	out.close();

	out.open("output/runtimes/overall_time.txt");

	out.close();

	out.open("output/runtimes/ADI_solving_time.txt");

	out.close();

	out.open("output/runtimes/ADI_solving_time_wall_time.txt");

	out.close();

	out.open("output/runtimes/ADI_solving_time_mean.txt");

	out.close();

	out.open("output/runtimes/ADI_solving_time_mean_wall_time.txt");

	out.close();

	out.open("output/runtimes/ADI_rhs_and_solving_time.txt");

	out.close();

	out.open("output/runtimes/ADI_rhs_and_solving_time_wall_time.txt");

	out.close();

	out.open("output/runtimes/ADI_rhs_and_solving_time_mean.txt");

	out.close();

	out.open("output/runtimes/ADI_rhs_and_solving_time_mean_wall_time.txt");

	out.close();

	out.open("output/runtimes/ADI_rhs_assembling_time.txt");

	out.close();

	out.open("output/runtimes/ADI_rhs_assembling_time_wall_time.txt");

	out.close();

	out.open("output/runtimes/ADI_rhs_assembling_time_mean.txt");

	out.close();

	out.open("output/runtimes/ADI_rhs_assembling_time_mean_wall_time.txt");

	out.close();

	out.open("output/runtimes/ADI_overall_time_wall_time.txt");

	out.close();

	out.open("output/runtimes/ADI_overall_time.txt");

	out.close();
}


int main ()
{
	try
	{
		using namespace dealii;
		using namespace MaxwellProblem;

		/*
		 * formatting output
		 */
		std::cout.setf( std::ios::left );

		/*
		 * set the maximum number of threads #num_threads that should be
		 * used by executing "export DEAL_II_NUM_THREADS=#num_threads"
		 * before executing this program
		 */
		MultithreadInfo::set_thread_limit();

		/*
		 * read config file
		 */
		pt::ini_parser::read_ini("config.ini", MaxwellProblem::global::config);

		/*
		 * timers
		 */
		dealii::Timer LeapFrogTimer, ADITimer;


		/*
		 * determine parameters for the refinement
		 * levels that should be run through.
		 * starting_lvl determines the first refinement
		 * level, where the actual refinement level is
		 * starting_lvl*refinement_factor.
		 * The refinement_factor determines the factor between
		 * the number of elements of two consecutive runs.
		 * The variable n_runs determines, how many runs should
		 * be processed.
		 */
		const unsigned int starting_lvl = MaxwellProblem::global::config.get<int>("Space_Discretisation.Starting_Lvl");
		const unsigned int n_runs = MaxwellProblem::global::config.get<int>("Space_Discretisation.N_Runs");
		const unsigned int refinement_factor = MaxwellProblem::global::config.get<int>("Space_Discretisation.Refinement_Factor");


		/*
		 * spatial dimension. This must not be changed.
		 */
		const unsigned int dim = 3;


		/*
		 *	determine the time step and ending time
		 */
		const double start_time = MaxwellProblem::global::config.get<double>("Time_Integration.Start_Time");
		const double end_time = start_time + MaxwellProblem::global::config.get<double>("Time_Integration.Simulation_Length");

		/*
		 *  flags: Should ADI or LF (or both?) be used for time integration?
		 */
		bool use_ADI = MaxwellProblem::global::config.get<bool>("Assembling.Use_ADI");
		bool use_LeapFrog = MaxwellProblem::global::config.get<bool>("Assembling.Use_LeapFrog");

		/*
		 *  create error-files if error is supposed to be put out
		 */
		if (MaxwellProblem::global::config.get<bool>("Error.Error"))
		{
			if(use_LeapFrog)
			{
				std::ofstream out_err ("output/error/error_std.txt");
				out_err.setf( std::ios::left );
				out_err << "degree" << "\t" << std::setw(12) << "h_max" << "\t" << std::setw(12) << "timestep" << "\t" << std::setw(12) << "error" <<  std::endl;
				out_err.close();
			}
			if(use_ADI)
			{
				std::ofstream out_err ("output/error/error_ADI.txt");
				out_err.setf( std::ios::left );
				out_err << "degree" << "\t" << std::setw(12) << "h_max" << "\t" << std::setw(12) << "timestep" << "\t" << std::setw(12) << "error" <<  std::endl;
				out_err.close();
			}
		}
		if ( MaxwellProblem::global::config.get("Error.ComputeDerivativeError", false) )
		{
			if(use_LeapFrog)
			{
				{
				std::ofstream out_err ("output/error/error_std_dt.txt");
				out_err.setf( std::ios::left );
				out_err << "degree" << "\t" << std::setw(12) << "h_max" << "\t" << std::setw(12) << "timestep" << "\t" << std::setw(12) << "error" <<  std::endl;
				out_err.close();
				}{
				std::ofstream out_err ("output/error/error_std_curl.txt");
				out_err.setf( std::ios::left );
				out_err << "degree" << "\t" << std::setw(12) << "h_max" << "\t" << std::setw(12) << "timestep" << "\t" << std::setw(12) << "error" <<  std::endl;
				out_err.close();
				}
			}
			if(use_ADI)
			{
				{
				std::ofstream out_err ("output/error/error_ADI_dt.txt");
				out_err.setf( std::ios::left );
				out_err << "degree" << "\t" << std::setw(12) << "h_max" << "\t" << std::setw(12) << "timestep" << "\t" << std::setw(12) << "error" <<  std::endl;
				out_err.close();
				}{
				std::ofstream out_err ("output/error/error_ADI_curl.txt");
				out_err.setf( std::ios::left );
				out_err << "degree" << "\t" << std::setw(12) << "h_max" << "\t" << std::setw(12) << "timestep" << "\t" << std::setw(12) << "error" <<  std::endl;
				out_err.close();
				}
			}
		}

		if (MaxwellProblem::global::config.get<bool>("Output.Profiling"))
			reset_profiling_files();


		/*
		 * loop over polynomial degrees
		 */
		std::stringstream degrees(MaxwellProblem::global::config.get<std::string>("Assembling.Pol_Degree"));
		std::string degree_string;
		while (std::getline(degrees, degree_string, ','))
		{
			unsigned int degree = std::atoi(degree_string.c_str());
			std::cout << std::endl << "------------------------------------polynomial_degree: " << degree << "-----------------------------------";
			if (degree < 10)
				std::cout << "-";
			std::cout << std::endl << std::endl;

			/*
			 * loop over refinement levels
			 */
			for (unsigned int refinement_lvl = starting_lvl; refinement_lvl < starting_lvl + n_runs; ++refinement_lvl)
			{
				std::cout << std::endl << "------------------------------------refinement_level: " << refinement_lvl*refinement_factor << "------------------------------------";
				if (refinement_lvl*refinement_factor < 10)
					std::cout << "-";
				std::cout << std::endl << std::endl;

				/*
				 * generation of the grid
				 */
				Triangulation<dim> triangulation(Triangulation<dim>::limit_level_difference_at_vertices);
				data::generate_cuboid(triangulation, refinement_lvl*refinement_factor);

				if(use_LeapFrog)
				{
					std::cout << std::endl << "------------------------------------LeapFrog-Scheme-----------------------------------------" << std::endl << std::endl;

	    			LeapFrogTimer.restart();

	    			/*
	    			 * initialize DGMethod object
	    			 */
					DGMethod<dim> dgmethod(triangulation, degree, start_time, end_time);

					/*
					 * assemble system matrices
					 */
					dgmethod.setup_and_assemble_system();

					/*
					 * loop over timesteps
					 */
					std::stringstream timesteps(MaxwellProblem::global::config.get<std::string>("Time_Integration.Timesteps"));
					std::string timestep_string;
					while (std::getline(timesteps, timestep_string, ','))
					{
						unsigned int total_timesteps = std::atoi(timestep_string.c_str());

						/*
						 * run simulation
						 */
						dgmethod.run (total_timesteps);
					}

					LeapFrogTimer.stop();

					/*
					 * profiling
					 */
					if (MaxwellProblem::global::config.get<bool>("Output.Profiling"))
					{
						std::stringstream filename;

						filename.str("");
						filename.clear();

						filename << "output/runtimes/overall_time.txt";

						std::ofstream out ((filename.str()).c_str(), std::ios_base::app);

						out << LeapFrogTimer() << std::endl;

						out.close();

						filename.str("");
						filename.clear();

						filename << "output/runtimes/overall_time_wall_time.txt";

						out.open((filename.str()).c_str(), std::ios_base::app);

						out << LeapFrogTimer.wall_time() << std::endl;

						out.close();
					}

					std::cout << std::endl << std::endl << "Elapsed CPU  time for LeapFrog-Scheme:  " << LeapFrogTimer() << " seconds." << std::endl;
					std::cout << "Elapsed wall time for LeapFrog-Scheme:  " << LeapFrogTimer.wall_time() << " seconds." << std::endl << std::endl;
				}
				if(use_ADI)
				{
					ADITimer.restart();

					std::cout << std::endl << "------------------------------------ADI-Scheme----------------------------------------------" << std::endl << std::endl;

	    			/*
	    			 * initialize DGMethod object and assemble matrices
	    			 */
					ADIDGMethod<dim> dgmethod(triangulation, degree, start_time, end_time);

					/*
					 * loop over timesteps
					 */
					std::stringstream timesteps(MaxwellProblem::global::config.get<std::string>("Time_Integration.Timesteps"));
					std::string timestep_string;
					while (std::getline(timesteps, timestep_string, ','))
					{
						unsigned int total_timesteps = std::atoi(timestep_string.c_str());

						/*
						 * run simulation
						 */
						dgmethod.run (total_timesteps);
					}

					ADITimer.stop();

					/*
					 * profiling
					 */
					if (MaxwellProblem::global::config.get<bool>("Output.Profiling"))
					{
						std::stringstream filename;

						filename.str("");
						filename.clear();

						filename << "output/runtimes/ADI_overall_time.txt";

						std::ofstream out ((filename.str()).c_str(), std::ios_base::app);

						out << ADITimer() << std::endl;

						out.close();

						filename.str("");
						filename.clear();

						filename << "output/runtimes/ADI_overall_time_wall_time.txt";

						out.open((filename.str()).c_str(), std::ios_base::app);

						out << ADITimer.wall_time() << std::endl;

						out.close();
					}

					std::cout << std::endl << std::endl << "Elapsed CPU  time for ADI-Scheme:       " << ADITimer() << " seconds." << std::endl;
					std::cout << "Elapsed wall time for ADI-Scheme:       " << ADITimer.wall_time() << " seconds." << std::endl << std::endl;
				}
				if(use_ADI&&use_LeapFrog)
				{
					std::cout << std::endl << "------------------------------------refinement_level: " << refinement_lvl*refinement_factor << " (Summary)--------------------------";
					if (refinement_lvl*refinement_factor < 10)
						std::cout << "-";
					std::cout << std::endl << std::endl << "Elapsed CPU  time for LeapFrog-Scheme:  " << LeapFrogTimer() << " seconds." << std::endl;
					std::cout << "Elapsed wall time for LeapFrog-Scheme:  " << LeapFrogTimer.wall_time() << " seconds.";
					std::cout << std::endl << std::endl << "Elapsed CPU  time for ADI-Scheme:       " << ADITimer() << " seconds." << std::endl;
					std::cout << "Elapsed wall time for ADI-Scheme:       " << ADITimer.wall_time() << " seconds." << std::endl << std::endl;
				}
	    	}
	    }
	}
	catch (std::exception &exc)
    {
		std::cerr << std::endl << std::endl
				<< "----------------------------------------------------"
				<< std::endl;
		std::cerr << "Exception on processing: " << std::endl
				<< exc.what() << std::endl
				<< "Aborting!" << std::endl
				<< "----------------------------------------------------"
				<< std::endl;
		return 1;
    }
	catch (...)
	{
		std::cerr << std::endl << std::endl
				<< "----------------------------------------------------"
				<< std::endl;
		std::cerr << "Unknown exception!" << std::endl
				<< "Aborting!" << std::endl
				<< "----------------------------------------------------"
				<< std::endl;
		return 1;
	};

	return 0;
}
