#include "ADIDGMaxwellGlobal.h"

namespace MaxwellProblem {
namespace Assembling{


template<int dim>
ADIDGMaxwellGlobal<dim>::ADIDGMaxwellGlobal(
		FESystem<dim>			&fe,
		const MappingQ1<dim>	&mapping,
		const QGauss<dim>		&quadrature,
		const QGauss<dim-1>		&face_quadrature,
		DoFHandler<dim>			&dof_handler
		)
:
mu_function (),
eps_function (),
fe(fe),
mapping(mapping),
quadrature(quadrature),
face_quadrature(face_quadrature),
dof_handler(dof_handler),
dg(),
fe_v (mapping, fe, quadrature, update_values| update_gradients| update_quadrature_points| update_JxW_values),
fe_v_face (mapping, fe, face_quadrature, update_values| update_quadrature_points| update_JxW_values| update_normal_vectors),
fe_v_subface (mapping, fe, face_quadrature, update_values| update_quadrature_points| update_JxW_values| update_normal_vectors),
fe_v_face_neighbor (mapping, fe, face_quadrature, update_values)
{}


template <int dim>
void ADIDGMaxwellGlobal<dim>::assemble_mass_matrix (
		ADIMassMatrix<dim>	&mass_matrix
		)
{
	const unsigned int dofs_per_cell = fe.dofs_per_cell;
	const unsigned int dofs_per_comp = dofs_per_cell/fe.n_blocks();

	std::vector<types::global_dof_index> dofs (dofs_per_cell);

	FullMatrix<double> cell_mass_matrix_H (dofs_per_comp, dofs_per_comp);
	FullMatrix<double> cell_mass_matrix_E (dofs_per_comp, dofs_per_comp);

	double eps, mu;
	
	bool jumpingEpsMu = (MaxwellProblem::global::config.get("Data.TestCase","") == "JumpSolution");
	bool lipschEpsMu = (MaxwellProblem::global::config.get("Data.TestCase","") == "LipschSolution");

	typename DoFHandler<dim>::active_cell_iterator
	cell = dof_handler.begin_active(),
	endc = dof_handler.end();
	for (; cell!=endc; ++cell)
	{
		cell_mass_matrix_H = 0;
		cell_mass_matrix_E = 0;

	    fe_v.reinit (cell);
		cell->get_dof_indices (dofs);
				
		if( jumpingEpsMu )
		{
			if ( cell->center()[0] < 0.5 )
				cell->set_material_id(0);
			else
				cell->set_material_id(1);
		}
		
		if( lipschEpsMu )
		{
			dg.assemble_cell_mass_matrices(cell_mass_matrix_H, cell_mass_matrix_E, fe_v);
		}
		else
		{
			eps = eps_function.value(cell->center());
			mu = mu_function.value(cell->center());
			dg.assemble_cell_mass_matrices(cell_mass_matrix_H, cell_mass_matrix_E, fe_v, eps, mu);
		}

		for (unsigned int i=0; i<dofs_per_comp; ++i)
			for (unsigned int j=0; j<dofs_per_comp; ++j)
			{
				if(fabs(cell_mass_matrix_H(i,j))>1e-17)
					mass_matrix(dofs[i],dofs[j]) += cell_mass_matrix_H(i,j);
				if(fabs(cell_mass_matrix_E(i,j))>1e-17)
					mass_matrix(dofs[i+dofs_per_comp],dofs[j+dofs_per_comp]) += cell_mass_matrix_E(i,j);
			}
	}
}


template <int dim>
void ADIDGMaxwellGlobal<dim>::assemble_stiffness_matrix (
		ADIMatrix<dim>		&mat
		)
{
	const unsigned int direction = mat.get_direction();

	const unsigned int dofs_per_cell = fe.dofs_per_cell;
	const unsigned int dofs_per_component = dofs_per_cell/fe.n_blocks();

	const unsigned int faces_per_cell = GeometryInfo<dim>::faces_per_cell;

	std::vector<types::global_dof_index>	dofs			(dofs_per_cell);
	std::vector<types::global_dof_index>	dofs_neighbor	(dofs_per_cell);

	FullMatrix<double>	cell_curl_matrix			(dofs_per_component, dofs_per_component);
	FullMatrix<double>	cell_face_matrix_ext_int	(dofs_per_component, dofs_per_component);
	FullMatrix<double>	cell_face_matrix_int_ext	(dofs_per_component, dofs_per_component);
	FullMatrix<double>	cell_curl_matrix_ext		(dofs_per_component, dofs_per_component);

	double eps, mu;
	double eps_neighbor, mu_neighbor;
	
	
	bool jumpingEpsMu = (MaxwellProblem::global::config.get("Data.TestCase","") == "JumpSolution");
// 	bool jumpingEpsMu = true;

	Vector<double> cell_vector (dofs_per_cell);

	typename DoFHandler<dim>::active_cell_iterator
	cell = dof_handler.begin_active(),
	endc = dof_handler.end();
	for (; cell!=endc; ++cell)
	{
		cell_curl_matrix = 0;
		cell_vector = 0;

		fe_v.reinit (cell);
		
		if ( jumpingEpsMu )
		{
			eps = eps_function.value(cell->center());
			mu = mu_function.value(cell->center());
		}

		
		dg.assemble_split_cell_curl(
				fe_v,
				cell_curl_matrix,
				cell_vector,
				direction
				);

		cell->get_dof_indices (dofs);

		for (unsigned int face_no = 0; face_no < faces_per_cell; ++face_no)
		{
			typename DoFHandler<dim>::face_iterator face = cell->face(face_no);

			if (!(face->at_boundary()))
			{
				Assert (cell->neighbor(face_no).state() == IteratorState::valid, ExcInternalError());
				typename DoFHandler<dim>::cell_iterator neighbor = cell->neighbor(face_no);

				if (face->has_children())
				{
					const unsigned int neighbor2 = cell->neighbor_face_no(face_no);
					for (unsigned int subface_no=0; subface_no<face->number_of_children(); ++subface_no)
					{
						typename DoFHandler<dim>::cell_iterator neighbor_child = cell->neighbor_child_on_subface (face_no, subface_no);
						Assert (!neighbor_child->has_children(), ExcInternalError());

						if ( jumpingEpsMu )
						{
							eps_neighbor = eps_function.value(neighbor_child->center());
							mu_neighbor = mu_function.value(neighbor_child->center());
						}

						cell_face_matrix_ext_int = 0;
						cell_face_matrix_int_ext = 0;
						cell_curl_matrix_ext = 0;

						fe_v_subface.reinit (cell, face_no, subface_no);
						fe_v_face_neighbor.reinit (neighbor_child, neighbor2);

						if ( jumpingEpsMu )
						{
							dg.assemble_split_face_curl(
									fe_v_face,
									fe_v_face_neighbor,
									cell_curl_matrix,
									cell_face_matrix_ext_int,
									cell_face_matrix_int_ext,
									cell_curl_matrix_ext,
									eps,
									mu,
									eps_neighbor,
									mu_neighbor,
									direction
									);
						}
						else
						{
							dg.assemble_split_face_curl(
									fe_v_face,
									fe_v_face_neighbor,
									cell_curl_matrix,
									cell_face_matrix_ext_int,
									cell_face_matrix_int_ext,
									cell_curl_matrix_ext,
									direction
									);
						}

						neighbor_child->get_dof_indices (dofs_neighbor);

						for (unsigned int i=0; i<dofs_per_component; ++i)
							for (unsigned int j=0; j<dofs_per_component; ++j)
							{
								// entries for split C_H
								mat.add(dofs[i+dofs_per_component], dofs_neighbor[j], cell_face_matrix_int_ext(i,j));
								mat.add(dofs_neighbor[i+dofs_per_component], dofs[j], cell_face_matrix_ext_int(i,j));
								mat.add(dofs_neighbor[i+dofs_per_component], dofs_neighbor[j], cell_curl_matrix_ext(i,j));

								// entries for split C_E
								mat.add(dofs[j], dofs_neighbor[i+dofs_per_component], -cell_face_matrix_ext_int(i,j));
								mat.add(dofs_neighbor[j], dofs[i+dofs_per_component], -cell_face_matrix_int_ext(i,j));
								mat.add(dofs_neighbor[j], dofs_neighbor[i+dofs_per_component], -cell_curl_matrix_ext(i,j));
							}
					}
				}
				else
				{
					if ( (neighbor->index() > cell->index() ||
							(neighbor->level() < cell->level() &&
							neighbor->index() == cell->index())))
					{
						const unsigned int neighbor2 = cell->neighbor_of_neighbor(face_no);

						if ( jumpingEpsMu )
						{
							eps_neighbor = eps_function.value(neighbor->center());
							mu_neighbor = mu_function.value(neighbor->center());
						}

						cell_face_matrix_ext_int = 0;
						cell_face_matrix_int_ext = 0;
						cell_curl_matrix_ext = 0;

						fe_v_face.reinit (cell, face_no);
						fe_v_face_neighbor.reinit (neighbor, neighbor2);

						if ( jumpingEpsMu )
						{
							dg.assemble_split_face_curl(
									fe_v_face,
									fe_v_face_neighbor,
									cell_curl_matrix,
									cell_face_matrix_ext_int,
									cell_face_matrix_int_ext,
									cell_curl_matrix_ext,
									eps,
									mu,
									eps_neighbor,
									mu_neighbor,
									direction
									);
						}
						else
						{
							dg.assemble_split_face_curl(
									fe_v_face,
									fe_v_face_neighbor,
									cell_curl_matrix,
									cell_face_matrix_ext_int,
									cell_face_matrix_int_ext,
									cell_curl_matrix_ext,
									direction
									);
						}

						neighbor->get_dof_indices (dofs_neighbor);

						for (unsigned int i=0; i<dofs_per_component; ++i)
							for (unsigned int j=0; j<dofs_per_component; ++j)
							{
								// entries for split C_H
								mat.add(dofs[i+dofs_per_component], dofs_neighbor[j], cell_face_matrix_int_ext(i,j));
								mat.add(dofs_neighbor[i+dofs_per_component], dofs[j], cell_face_matrix_ext_int(i,j));
								mat.add(dofs_neighbor[i+dofs_per_component], dofs_neighbor[j], cell_curl_matrix_ext(i,j));

								// entries for split C_E
								mat.add(dofs[j], dofs_neighbor[i+dofs_per_component], -cell_face_matrix_ext_int(i,j));
								mat.add(dofs_neighbor[j], dofs[i+dofs_per_component], -cell_face_matrix_int_ext(i,j));
								mat.add(dofs_neighbor[j], dofs_neighbor[i+dofs_per_component], -cell_curl_matrix_ext(i,j));
							}
					}
				}
			}
		}

		for (unsigned int i=0; i<dofs_per_component; ++i)
			for (unsigned int j=0; j<dofs_per_component; ++j)
			{
				// entries for H
				mat.add(dofs[i+dofs_per_component], dofs[j], cell_curl_matrix(i,j));

				// entries for E
				mat.add(dofs[j], dofs[i+dofs_per_component], -cell_curl_matrix(i,j));
			}
	}
}


template class ADIDGMaxwellGlobal<3>;

}
}
