#include "ADIDGMaxwellLocal.h"

namespace MaxwellProblem {
namespace Assembling{


template <int dim>
ADIDGMaxwellLocal<dim>::ADIDGMaxwellLocal ()
	:
	E(dim),
	H(0),
	mu_function (),
	eps_function (),
	index_deriv_pairs(2),
	sign(2)
{
	index_deriv_pairs[0].resize(3);
	index_deriv_pairs[0][0] = std::make_pair(2,1);
	index_deriv_pairs[0][1] = std::make_pair(0,2);
	index_deriv_pairs[0][2] = std::make_pair(1,0);
	sign[0] = 1;

	index_deriv_pairs[1].resize(3);
	index_deriv_pairs[1][0] = std::make_pair(1,2);
	index_deriv_pairs[1][1] = std::make_pair(2,0);
	index_deriv_pairs[1][2] = std::make_pair(0,1);
	sign[1] = -1;
}

template <int dim>
void ADIDGMaxwellLocal<dim>::assemble_split_cell_curl(
		const FEValues<dim>										&fe_v,
		FullMatrix<double>										&cell_curl_matrix,
		Vector<double>											&/*cell_vector*/,
		unsigned int											direction
		) const
{
	const std::vector< double > &JxW = fe_v.get_JxW_values ();
	std::vector< Tensor<1,dim,double> > rhs (fe_v.n_quadrature_points);

	const unsigned int dofs_per_cell = fe_v.dofs_per_cell/fe_v.get_fe().n_blocks();
	for (unsigned int q_index=0; q_index<fe_v.n_quadrature_points; ++q_index)
		for (unsigned int i=0; i<dofs_per_cell; ++i)
		{
			Tensor<1,dim,double> E_i = fe_v[E].value(i + dofs_per_cell ,q_index);

			if (direction == 0)
				for (unsigned int j=0; j<dofs_per_cell; ++j)
				{
					Tensor<1,dim,double> c1_H_j = fe_v[H].c1(j,q_index);

					cell_curl_matrix(i,j) += c1_H_j	*	E_i	* 	JxW[q_index];
				}

			else if (direction == 1)
				for (unsigned int j=0; j<dofs_per_cell; ++j)
				{
					Tensor<1,dim,double> c2_H_j = fe_v[H].c2(j,q_index);

					cell_curl_matrix(i,j) += c2_H_j	*	E_i	* 	JxW[q_index];
				}
		}
}

template <int dim>
void ADIDGMaxwellLocal<dim>::assemble_split_face_curl(
		const FEFaceValuesBase<dim>								&fe_v,
		const FEFaceValuesBase<dim>								&fe_v_neighbor,
		FullMatrix<double>										&cell_curl_matrix,
		FullMatrix<double>										&cell_face_matrix_ext_int,
		FullMatrix<double>										&cell_face_matrix_int_ext,
		FullMatrix<double>										&cell_curl_matrix_neighbor,
		unsigned int											direction
		) const
{
	double w = .5;
	double w_ext = .5;
	const std::vector<double> &JxW = fe_v.get_JxW_values ();
	const std::vector<Tensor <1,dim>> &normals = fe_v.get_all_normal_vectors ();

	const unsigned int dofs_cell = fe_v.dofs_per_cell/fe_v.get_fe().n_blocks();
	const unsigned int dofs_neighbor = fe_v_neighbor.dofs_per_cell/fe_v.get_fe().n_blocks();

	int loc_sign = sign[direction];

	for (unsigned int normal_direction=0; normal_direction < dim; ++normal_direction)
		if (fabs(normals[0][normal_direction]) > 1e-14)
		{
			loc_sign *= normals[0][normal_direction];

			for (unsigned int q_index=0; q_index<fe_v.n_quadrature_points; ++q_index)
			{
				for (unsigned int i=0; i<dofs_cell; ++i)
				{
					Tensor<1,dim,double> E_i = fe_v[E].value(i + dofs_cell,q_index);

					for (unsigned int j=0; j<dofs_neighbor; ++j)
					{
						Tensor<1,dim,double> H_j = fe_v[H].value(j,q_index);

						cell_curl_matrix(i,j) -= w * loc_sign * H_j[index_deriv_pairs[direction][normal_direction].second] *	E_i[index_deriv_pairs[direction][normal_direction].first ]	* 	JxW[q_index];
					}
				}

				for (unsigned int i=0; i<dofs_cell; ++i)
				{
					Tensor<1,dim,double> E_i = fe_v[E].value(i + dofs_cell,q_index);

					for (unsigned int l=0; l<dofs_neighbor; ++l)
					{
						Tensor<1,dim,double> H_neighbor_l = fe_v_neighbor[H].value(l,q_index);

						cell_face_matrix_int_ext(i,l) += w * loc_sign *	H_neighbor_l[index_deriv_pairs[direction][normal_direction].second] *	E_i[index_deriv_pairs[direction][normal_direction].first ]	* 	JxW[q_index];
					}
				}

				for (unsigned int k=0; k<dofs_neighbor; ++k)
				{
					Tensor<1,dim,double> E_neighbor_k = fe_v_neighbor[E].value(k + dofs_neighbor,q_index);

					for (unsigned int j=0; j<dofs_cell; ++j)
					{
						Tensor<1,dim,double> H_j = fe_v[H].value(j,q_index);

						cell_face_matrix_ext_int(k,j) -= w_ext * loc_sign *	H_j[index_deriv_pairs[direction][normal_direction].second] *	E_neighbor_k[index_deriv_pairs[direction][normal_direction].first ]	* 	JxW[q_index];
					}
				}

				for (unsigned int k=0; k<dofs_neighbor; ++k)
				{
					Tensor<1,dim,double> E_neighbor_k = fe_v_neighbor[E].value(k + dofs_neighbor,q_index);

					for (unsigned int l=0; l<dofs_neighbor; ++l)
					{
						Tensor<1,dim,double> H_neighbor_l = fe_v_neighbor[H].value(l,q_index);

						cell_curl_matrix_neighbor(k,l) += w_ext * loc_sign *	H_neighbor_l[index_deriv_pairs[direction][normal_direction].second] *	E_neighbor_k[index_deriv_pairs[direction][normal_direction].first ]	* 	JxW[q_index];
					}
				}
			}
			break;
		}
}

template <int dim>
void ADIDGMaxwellLocal<dim>::assemble_split_face_curl(
		const FEFaceValuesBase<dim>								&fe_v,
		const FEFaceValuesBase<dim>								&fe_v_neighbor,
		FullMatrix<double>										&cell_curl_matrix,
		FullMatrix<double>										&cell_face_matrix_ext_int,
		FullMatrix<double>										&cell_face_matrix_int_ext,
		FullMatrix<double>										&cell_curl_matrix_neighbor,
		double 													eps,
		double 													mu,
		double 													eps_neighbor,
		double													mu_neighbor,
		unsigned int											direction
		) const
{
	double w = std::sqrt(eps/mu);
	double w_ext = std::sqrt(eps_neighbor/mu_neighbor);
	double w_ges = w + w_ext;
	w /= w_ges;
	w_ext /= w_ges;
	const std::vector<double> &JxW = fe_v.get_JxW_values ();
	const std::vector<Tensor <1,dim>> &normals = fe_v.get_all_normal_vectors ();

	const unsigned int dofs_cell = fe_v.dofs_per_cell/fe_v.get_fe().n_blocks();
	const unsigned int dofs_neighbor = fe_v_neighbor.dofs_per_cell/fe_v.get_fe().n_blocks();

	int loc_sign = sign[direction];

	for (unsigned int normal_direction=0; normal_direction < dim; ++normal_direction)
		if (fabs(normals[0][normal_direction]) > 1e-14)
		{
			loc_sign *= normals[0][normal_direction];

			for (unsigned int q_index=0; q_index<fe_v.n_quadrature_points; ++q_index)
			{
				for (unsigned int i=0; i<dofs_cell; ++i)
				{
					Tensor<1,dim,double> E_i = fe_v[E].value(i + dofs_cell,q_index);

					for (unsigned int j=0; j<dofs_neighbor; ++j)
					{
						Tensor<1,dim,double> H_j = fe_v[H].value(j,q_index);

						cell_curl_matrix(i,j) -= w * loc_sign * H_j[index_deriv_pairs[direction][normal_direction].second] *	E_i[index_deriv_pairs[direction][normal_direction].first ]	* 	JxW[q_index];
					}
				}

				for (unsigned int i=0; i<dofs_cell; ++i)
				{
					Tensor<1,dim,double> E_i = fe_v[E].value(i + dofs_cell,q_index);

					for (unsigned int l=0; l<dofs_neighbor; ++l)
					{
						Tensor<1,dim,double> H_neighbor_l = fe_v_neighbor[H].value(l,q_index);

						cell_face_matrix_int_ext(i,l) += w * loc_sign *	H_neighbor_l[index_deriv_pairs[direction][normal_direction].second] *	E_i[index_deriv_pairs[direction][normal_direction].first ]	* 	JxW[q_index];
					}
				}

				for (unsigned int k=0; k<dofs_neighbor; ++k)
				{
					Tensor<1,dim,double> E_neighbor_k = fe_v_neighbor[E].value(k + dofs_neighbor,q_index);

					for (unsigned int j=0; j<dofs_cell; ++j)
					{
						Tensor<1,dim,double> H_j = fe_v[H].value(j,q_index);

						cell_face_matrix_ext_int(k,j) -= w_ext * loc_sign *	H_j[index_deriv_pairs[direction][normal_direction].second] *	E_neighbor_k[index_deriv_pairs[direction][normal_direction].first ]	* 	JxW[q_index];
					}
				}

				for (unsigned int k=0; k<dofs_neighbor; ++k)
				{
					Tensor<1,dim,double> E_neighbor_k = fe_v_neighbor[E].value(k + dofs_neighbor,q_index);

					for (unsigned int l=0; l<dofs_neighbor; ++l)
					{
						Tensor<1,dim,double> H_neighbor_l = fe_v_neighbor[H].value(l,q_index);

						cell_curl_matrix_neighbor(k,l) += w_ext * loc_sign *	H_neighbor_l[index_deriv_pairs[direction][normal_direction].second] *	E_neighbor_k[index_deriv_pairs[direction][normal_direction].first ]	* 	JxW[q_index];
					}
				}
			}
			break;
		}
}


template<int dim>
void ADIDGMaxwellLocal<dim>::assemble_cell_mass_matrix(
		FullMatrix<double> 			&cell_mass_matrix,
		FEValues<dim>				&fe_v,
		double						eps,
		double						mu
		) const
{
	const std::vector<double> &JxW = fe_v.get_JxW_values ();

	for (unsigned int q_index=0; q_index<fe_v.n_quadrature_points; ++q_index)
		for (unsigned int i=0; i<fe_v.dofs_per_cell; ++i)
		{
			Tensor<1,dim,double> E_i = fe_v[E].value(i,q_index);

			Tensor<1,dim,double> H_i = fe_v[H].value(i,q_index);

			for (unsigned int j=0; j<fe_v.dofs_per_cell; ++j)
			{
				Tensor<1,dim,double> E_j = fe_v[E].value(j,q_index);

				Tensor<1,dim,double> H_j = fe_v[H].value(j,q_index);

				cell_mass_matrix(i,j) += (eps * E_j * E_i	+ mu *  H_j * H_i	)	*	JxW[q_index];
			}
	}
}

template<int dim>
void ADIDGMaxwellLocal<dim>::assemble_cell_mass_matrices(
		FullMatrix<double> 			&cell_mass_matrix_H,
		FullMatrix<double> 			&cell_mass_matrix_E,
		FEValues<dim>				&fe_v,
		double						eps,
		double						mu
		) const
{
	const std::vector<double> &JxW = fe_v.get_JxW_values ();

	const unsigned int dofs_cell = fe_v.dofs_per_cell/fe_v.get_fe().n_blocks();

	for (unsigned int q_index=0; q_index<fe_v.n_quadrature_points; ++q_index)
		for (unsigned int i=0; i<dofs_cell; ++i)
		{
			Tensor<1,dim,double> E_i = fe_v[E].value(i + dofs_cell,q_index);

			Tensor<1,dim,double> H_i = fe_v[H].value(i,q_index);

			for (unsigned int j=0; j<dofs_cell; ++j)
			{
				Tensor<1,dim,double> E_j = fe_v[E].value(j + dofs_cell,q_index);

				Tensor<1,dim,double> H_j = fe_v[H].value(j,q_index);

				cell_mass_matrix_H(i,j) +=  mu *  H_j * H_i	*	JxW[q_index];
				cell_mass_matrix_E(i,j) +=  eps * E_j * E_i	*	JxW[q_index];
			}
		}
}

template<int dim>
void ADIDGMaxwellLocal<dim>::assemble_cell_mass_matrices(
		FullMatrix<double> 			&cell_mass_matrix_H,
		FullMatrix<double> 			&cell_mass_matrix_E,
		FEValues<dim>				&fe_v
		) const
{
	const std::vector<double> &JxW = fe_v.get_JxW_values ();

	const unsigned int dofs_cell = fe_v.dofs_per_cell/fe_v.get_fe().n_blocks();
	
	double eps, mu;

	for (unsigned int q_index=0; q_index<fe_v.n_quadrature_points; ++q_index)
	{
		const auto q_point = fe_v.quadrature_point(q_index);
		
		eps = eps_function.value(q_point);
		mu = mu_function.value(q_point);
		
		for (unsigned int i=0; i<dofs_cell; ++i)
		{
			Tensor<1,dim,double> E_i = fe_v[E].value(i + dofs_cell,q_index);

			Tensor<1,dim,double> H_i = fe_v[H].value(i,q_index);

			for (unsigned int j=0; j<dofs_cell; ++j)
			{
				Tensor<1,dim,double> E_j = fe_v[E].value(j + dofs_cell,q_index);

				Tensor<1,dim,double> H_j = fe_v[H].value(j,q_index);

				cell_mass_matrix_H(i,j) +=  mu *  H_j * H_i	*	JxW[q_index];
				cell_mass_matrix_E(i,j) +=  eps * E_j * E_i	*	JxW[q_index];
			}
		}
	}
}


template class ADIDGMaxwellLocal<3>;

}
} /* namespace MaxwellProblem */
