#include "DGMaxwellGlobal.h"

namespace MaxwellProblem {
namespace Assembling{

template<int dim>
DGMaxwellGlobal<dim>::DGMaxwellGlobal(FESystem<dim>	&fe,
		const MappingQ1<dim>		&mapping,
		const Quadrature<dim>		&quadrature,
		const Quadrature<dim-1>		&face_quadrature,
		DoFHandler<dim>				&dof_handler)
:
mu_function (),
eps_function (),
fe(fe),
mapping(mapping),
quadrature(quadrature),
face_quadrature(face_quadrature),
dof_handler(dof_handler),
dg(),
dofs_per_cell(fe.dofs_per_cell),
fe_v (mapping, fe, quadrature, update_values| update_gradients| update_quadrature_points| update_JxW_values),
fe_v_face (mapping, fe, face_quadrature, update_values| update_quadrature_points| update_JxW_values| update_normal_vectors),
fe_v_subface (mapping, fe, face_quadrature, update_values| update_quadrature_points| update_JxW_values| update_normal_vectors),
fe_v_face_neighbor (mapping, fe, face_quadrature, update_values)
{}

template<int dim>
template<class MatrixType>
void DGMaxwellGlobal<dim>::assemble_mass_matrix(MatrixType &mass_matrix_H,
		MatrixType &mass_matrix_E)
{
	std::vector<types::global_dof_index>	dofs (fe.dofs_per_cell);

	FullMatrix<double>	cell_mass_matrix_H (dofs_per_cell, dofs_per_cell);
	FullMatrix<double>	cell_mass_matrix_E (dofs_per_cell, dofs_per_cell);

	unsigned int comp_per_cell = fe.n_components();
	unsigned int dofs_per_comp_per_cell = dofs_per_cell / comp_per_cell;

	double eps, mu;

	typename DoFHandler<dim>::active_cell_iterator
	cell = dof_handler.begin_active(),
	endc = dof_handler.end();
	for (; cell!=endc; ++cell)
    {
		cell_mass_matrix_H = 0;
		cell_mass_matrix_E = 0;
		fe_v.reinit (cell);

		eps = eps_function.value(cell->center());
		mu = mu_function.value(cell->center());

		cell->get_dof_indices (dofs);
		dg.assemble_cell_mass_matrices(cell_mass_matrix_H, cell_mass_matrix_E, fe_v, eps, mu);

		unsigned int tmp_i, tmp_j = 0;
		for (unsigned int k=0; k<comp_per_cell; ++k)
		{
			for (unsigned int i=0; i<dofs_per_comp_per_cell; ++i)
			{
				tmp_i = k * dofs_per_comp_per_cell + i;
				for (unsigned int j=0; j<dofs_per_comp_per_cell; ++j)
				{
					tmp_j = k * dofs_per_comp_per_cell + j;
					mass_matrix_H.add(dofs[tmp_i],dofs[tmp_j],cell_mass_matrix_H(tmp_i,tmp_j));
					mass_matrix_E.add(dofs[tmp_i],dofs[tmp_j],cell_mass_matrix_E(tmp_i,tmp_j));
				}
			}
		}
    }
}

template<int dim>
template<class MatrixType>
void DGMaxwellGlobal<dim>::assemble_mass_matrix(MatrixType &mass_matrix_H,
						MatrixType &mass_matrix_E,
						MatrixType &mass_matrix_H_inv,
						MatrixType &mass_matrix_E_inv)
{
	std::vector<types::global_dof_index>	dofs (fe.dofs_per_cell);

	FullMatrix<double>	cell_mass_matrix_H (dofs_per_cell, dofs_per_cell);
	FullMatrix<double>	cell_mass_matrix_E (dofs_per_cell, dofs_per_cell);
	FullMatrix<double>	cell_mass_matrix_H_inv (dofs_per_cell, dofs_per_cell);
	FullMatrix<double>	cell_mass_matrix_E_inv (dofs_per_cell, dofs_per_cell);

	unsigned int comp_per_cell = fe.n_components();
	unsigned int dofs_per_comp_per_cell = dofs_per_cell / comp_per_cell;

	double eps, mu;

	typename DoFHandler<dim>::active_cell_iterator
	cell = dof_handler.begin_active(),
	endc = dof_handler.end();
	for (; cell!=endc; ++cell)
    {
		cell_mass_matrix_H = 0;
		cell_mass_matrix_E = 0;
		fe_v.reinit (cell);

		eps = eps_function.value(cell->center());
		mu = mu_function.value(cell->center());

		cell->get_dof_indices (dofs);
		dg.assemble_cell_mass_matrices(cell_mass_matrix_H, cell_mass_matrix_E, fe_v, eps, mu);

		cell_mass_matrix_H_inv = cell_mass_matrix_H;
		cell_mass_matrix_H_inv.gauss_jordan();

		cell_mass_matrix_E_inv = cell_mass_matrix_E;
		cell_mass_matrix_E_inv.gauss_jordan();

		unsigned int tmp_i, tmp_j = 0;
		for (unsigned int k=0; k<comp_per_cell; ++k)
		{
			for (unsigned int i=0; i<dofs_per_comp_per_cell; ++i)
			{
				tmp_i = k * dofs_per_comp_per_cell + i;
				for (unsigned int j=0; j<dofs_per_comp_per_cell; ++j)
				{
					tmp_j = k * dofs_per_comp_per_cell + j;
					mass_matrix_H.add(dofs[tmp_i],dofs[tmp_j],cell_mass_matrix_H(tmp_i,tmp_j));
					mass_matrix_E.add(dofs[tmp_i],dofs[tmp_j],cell_mass_matrix_E(tmp_i,tmp_j));
					mass_matrix_H_inv.add(dofs[tmp_i],dofs[tmp_j],cell_mass_matrix_H_inv(tmp_i,tmp_j));
					mass_matrix_E_inv.add(dofs[tmp_i],dofs[tmp_j],cell_mass_matrix_E_inv(tmp_i,tmp_j));
				}
			}
		}
    }
}

template<int dim>
template<class MatrixType>
void DGMaxwellGlobal<dim>::assemble_discrete_curl_matrices(MatrixType &C_H, MatrixType &C_E)
{
	std::vector<types::global_dof_index>	dofs (fe.dofs_per_cell);
	std::vector<types::global_dof_index>	dofs_neighbor (fe.dofs_per_cell);


	FullMatrix<double>	cell_curl_matrix (dofs_per_cell, dofs_per_cell);
	FullMatrix<double>	cell_face_matrix_ext_int (dofs_per_cell, dofs_per_cell);
	FullMatrix<double>	cell_face_matrix_int_ext (dofs_per_cell, dofs_per_cell);
	FullMatrix<double>	cell_curl_matrix_ext (dofs_per_cell, dofs_per_cell);

	double eps, mu;
	double eps_neighbor, mu_neighbor;

	typename DoFHandler<dim>::active_cell_iterator
	cell = dof_handler.begin_active(),
	endc = dof_handler.end();
	for (; cell!=endc; ++cell)
    {
		cell_curl_matrix = 0;
		fe_v.reinit (cell);

		eps = eps_function.value(cell->center());
		mu = mu_function.value(cell->center());

		dg.assemble_cell_curl(fe_v, cell_curl_matrix);

		cell->get_dof_indices (dofs);
		for (unsigned int face_no=0; face_no<GeometryInfo<dim>::faces_per_cell; ++face_no)
        {
			typename DoFHandler<dim>::face_iterator face = cell->face(face_no);
			if (!(face->at_boundary()))
            {
				Assert (cell->neighbor(face_no).state() == IteratorState::valid, ExcInternalError());
				typename DoFHandler<dim>::cell_iterator neighbor = cell->neighbor(face_no);
                if (face->has_children())
                  {
                    const unsigned int neighbor2 = cell->neighbor_face_no(face_no);
                    for (unsigned int subface_no=0; subface_no<face->number_of_children(); ++subface_no)
                      {
                        typename DoFHandler<dim>::cell_iterator neighbor_child
                          = cell->neighbor_child_on_subface (face_no, subface_no);
                        Assert (!neighbor_child->has_children(), ExcInternalError());


                        eps_neighbor = eps_function.value(neighbor_child->center());
                        mu_neighbor = mu_function.value(neighbor_child->center());

						cell_face_matrix_ext_int = 0;
						cell_face_matrix_int_ext = 0;
						cell_curl_matrix_ext = 0;
                        fe_v_subface.reinit (cell, face_no, subface_no);
                        fe_v_face_neighbor.reinit (neighbor_child, neighbor2);
                        dg.assemble_face_curl(fe_v_subface,
                                               fe_v_face_neighbor,
											   cell_curl_matrix,
											   cell_face_matrix_ext_int,
											   cell_face_matrix_int_ext,
											   cell_curl_matrix_ext,
											   eps,
											   mu,
											   eps_neighbor,
											   mu_neighbor);
                        neighbor_child->get_dof_indices (dofs_neighbor);
						for (unsigned int i=0; i<dofs_per_cell; ++i)
						{
							for (unsigned int j=0; j<dofs_per_cell; ++j)
							{
								C_H.add(dofs[i], dofs_neighbor[j], cell_face_matrix_int_ext(i,j));
								C_H.add(dofs_neighbor[i], dofs[j], cell_face_matrix_ext_int(i,j));
								C_H.add(dofs_neighbor[i], dofs_neighbor[j], cell_curl_matrix_ext(i,j));
								C_E.add(dofs_neighbor[j], dofs[i], cell_face_matrix_int_ext(i,j));
								C_E.add(dofs[j], dofs_neighbor[i], cell_face_matrix_ext_int(i,j));
								C_E.add(dofs_neighbor[j], dofs_neighbor[i], cell_curl_matrix_ext(i,j));
							}
						}
                      }
                  }
                else
				{
					if (!cell->neighbor_is_coarser(face_no) &&
						(neighbor->index() > cell->index() ||
						 (neighbor->level() < cell->level() &&
						  neighbor->index() == cell->index())))
					{
						const unsigned int neighbor2 = cell->neighbor_of_neighbor(face_no);

						eps_neighbor = eps_function.value(neighbor->center());
						mu_neighbor = mu_function.value(neighbor->center());

						cell_face_matrix_ext_int = 0;
						cell_face_matrix_int_ext = 0;
						cell_curl_matrix_ext = 0;

						fe_v_face.reinit (cell, face_no);
						fe_v_face_neighbor.reinit (neighbor, neighbor2);

						dg.assemble_face_curl(fe_v_face,
											  fe_v_face_neighbor,
											  cell_curl_matrix,
											  cell_face_matrix_ext_int,
											  cell_face_matrix_int_ext,
											  cell_curl_matrix_ext,
											  eps,
											  mu,
											  eps_neighbor,
											  mu_neighbor);

						neighbor->get_dof_indices (dofs_neighbor);
						for (unsigned int i=0; i<dofs_per_cell; ++i)
						{
							for (unsigned int j=0; j<dofs_per_cell; ++j)
							{
								C_H.add(dofs[i], dofs_neighbor[j], cell_face_matrix_int_ext(i,j));
								C_H.add(dofs_neighbor[i], dofs[j], cell_face_matrix_ext_int(i,j));
								C_H.add(dofs_neighbor[i], dofs_neighbor[j], cell_curl_matrix_ext(i,j));
								C_E.add(dofs_neighbor[j], dofs[i], cell_face_matrix_int_ext(i,j));
								C_E.add(dofs[j], dofs_neighbor[i], cell_face_matrix_ext_int(i,j));
								C_E.add(dofs_neighbor[j], dofs_neighbor[i], cell_curl_matrix_ext(i,j));
							}
						}
					}
                }
            }
        }
		for (unsigned int i=0; i<dofs_per_cell; ++i)
		{
			for (unsigned int j=0; j<dofs_per_cell; ++j)
			{
				C_H.add(dofs[i], dofs[j], cell_curl_matrix(i,j));
				C_E.add(dofs[j], dofs[i], cell_curl_matrix(i,j));
			}
		}
    }
}


template class DGMaxwellGlobal<3>;
template void DGMaxwellGlobal<3>::assemble_mass_matrix(SparseMatrix<double> &mass_matrix_H,	SparseMatrix<double> &mass_matrix_E);
template void DGMaxwellGlobal<3>::assemble_mass_matrix(SparseMatrix<double> &mass_matrix_H,	SparseMatrix<double> &mass_matrix_E,
														SparseMatrix<double> &mass_matrix_H_inv, SparseMatrix<double> &mass_matrix_E_inv);
template void DGMaxwellGlobal<3>::assemble_mass_matrix(BlockSparseMatrix<double> &mass_matrix_H, BlockSparseMatrix<double> &mass_matrix_E);
template void DGMaxwellGlobal<3>::assemble_mass_matrix(BlockSparseMatrix<double> &mass_matrix_H, BlockSparseMatrix<double> &mass_matrix_E,
														BlockSparseMatrix<double> &mass_matrix_H_inv, BlockSparseMatrix<double> &mass_matrix_E_inv);

template void DGMaxwellGlobal<3>::assemble_discrete_curl_matrices(SparseMatrix<double> &C_H, SparseMatrix<double> &C_E);
template void DGMaxwellGlobal<3>::assemble_discrete_curl_matrices(BlockSparseMatrix<double> &C_H, BlockSparseMatrix<double> &C_E);
} /* namespace Assembling */
} /* namespace MaxwellProblem */
