#include "DGMaxwellLocal.h"

namespace MaxwellProblem {
namespace Assembling{
template <int dim>
DGMaxwellLocal<dim>::DGMaxwellLocal () : V(0)
{}

template <int dim>
void DGMaxwellLocal<dim>::assemble_cell_curl(const FEValues<dim>	&fe_v,
											FullMatrix<double>	&cell_curl_matrix) const
{
	const std::vector< double > &JxW = fe_v.get_JxW_values ();
	std::vector< Tensor<1,dim,double> > rhs (fe_v.n_quadrature_points);

	const unsigned int dofs_per_cell = fe_v.dofs_per_cell;
	for (unsigned int q_point=0; q_point<fe_v.n_quadrature_points; ++q_point)
		for (unsigned int i=0; i<dofs_per_cell; ++i)
		{
			Tensor<1,dim,double> E_i = fe_v[V].value(i ,q_point);
			for (unsigned int j=0; j<dofs_per_cell; ++j)
			{
				Tensor<1,dim,double> curl_H_j = fe_v[V].curl(j,q_point);
				cell_curl_matrix(i,j) += 	curl_H_j	*	E_i	* 	JxW[q_point];
			}
		}
}

template <int dim>
void DGMaxwellLocal<dim>::assemble_face_curl(const FEFaceValuesBase<dim>	&fe_v,
												  const FEFaceValuesBase<dim>	&fe_v_neighbor,
												  FullMatrix<double>			&cell_curl_matrix,
												  FullMatrix<double>			&cell_face_matrix_ext_int,
												  FullMatrix<double>			&cell_face_matrix_int_ext,
												  FullMatrix<double>			&cell_curl_matrix_neighbor,
												  double 						eps,
												  double 						mu,
												  double 						eps_neighbor,
												  double						mu_neighbor) const
{
	double w = std::sqrt(eps/mu);
	double w_ext = std::sqrt(eps_neighbor/mu_neighbor);
	double w_ges = w + w_ext;
	w /= w_ges;
	w_ext /= w_ges;
	const std::vector<double> &JxW = fe_v.get_JxW_values ();
	const std::vector<Tensor <1,dim>> &normals = fe_v.get_all_normal_vectors ();

	const unsigned int dofs_cell = fe_v.dofs_per_cell;
	const unsigned int dofs_neighbor = fe_v_neighbor.dofs_per_cell;

	for (unsigned int q_point=0; q_point<fe_v.n_quadrature_points; ++q_point)
	{
		for (unsigned int i=0; i<dofs_cell; ++i)
		{
			Tensor<1,dim,double> E_i = fe_v[V].value(i,q_point);

			for (unsigned int j=0; j<dofs_cell; ++j)
			{
				Tensor<1,dim,double> H_j = fe_v[V].value(j,q_point);
				Tensor<1,dim,double> n_times_H_j;
				n_times_H_j = cross_product_3d(normals[q_point], H_j);

				cell_curl_matrix(i,j) -= w * 	n_times_H_j	*	E_i	 * JxW[q_point];
			}
		}

		for (unsigned int i=0; i<dofs_cell; ++i)
		{
			Tensor<1,dim,double> E_i = fe_v[V].value(i,q_point);

			for (unsigned int l=0; l<dofs_neighbor; ++l)
			{

				Tensor<1,dim,double> H_neighbor_l = fe_v_neighbor[V].value(l,q_point);
				Tensor<1,dim,double> n_times_H_neighbor_l;
				n_times_H_neighbor_l = cross_product_3d(normals[q_point], H_neighbor_l);

				cell_face_matrix_int_ext(i,l) += w * 	n_times_H_neighbor_l	*	E_i	 * JxW[q_point];
			}
		}

		for (unsigned int k=0; k<dofs_neighbor; ++k)
		{
			Tensor<1,dim,double> E_neighbor_k = fe_v_neighbor[V].value(k,q_point);

			for (unsigned int j=0; j<dofs_cell; ++j)
			{

				Tensor<1,dim,double> H_j = fe_v[V].value(j,q_point);
				Tensor<1,dim,double> n_times_H_j;
				n_times_H_j = cross_product_3d(normals[q_point], H_j);

				cell_face_matrix_ext_int(k,j) -= w_ext * 	n_times_H_j	* 	E_neighbor_k * JxW[q_point];
			}
		}

		for (unsigned int k=0; k<dofs_neighbor; ++k)
		{
			Tensor<1,dim,double> E_neighbor_k = fe_v_neighbor[V].value(k,q_point);

			for (unsigned int l=0; l<dofs_neighbor; ++l)
			{

				Tensor<1,dim,double> H_neighbor_l = fe_v_neighbor[V].value(l,q_point);
				Tensor<1,dim,double> n_times_H_neighbor_l;
				n_times_H_neighbor_l = cross_product_3d(normals[q_point], H_neighbor_l);

				cell_curl_matrix_neighbor(k,l) += w_ext * 	n_times_H_neighbor_l	*	E_neighbor_k	 * JxW[q_point];
			}
		}
	}
}


template<int dim>
void DGMaxwellLocal<dim>::assemble_cell_mass_matrices(
		FullMatrix<double> 			&cell_mass_matrix_H,
		FullMatrix<double> 			&cell_mass_matrix_E,
		FEValues<dim>				&fe_v,
		double						eps,
		double						mu) const
{
	const std::vector<double> &JxW = fe_v.get_JxW_values ();

	const unsigned int dofs_cell = fe_v.dofs_per_cell;

	for (unsigned int q_point=0; q_point<fe_v.n_quadrature_points; ++q_point)
		for (unsigned int i=0; i<dofs_cell; ++i)
		{
			Tensor<1,dim,double> E_i = fe_v[V].value(i, q_point);

			Tensor<1,dim,double> H_i = fe_v[V].value(i, q_point);

			for (unsigned int j=0; j<dofs_cell; ++j)
			{
				Tensor<1,dim,double> E_j = fe_v[V].value(j, q_point);

				Tensor<1,dim,double> H_j = fe_v[V].value(j, q_point);

				cell_mass_matrix_H(i,j) +=  mu *  H_j * H_i	*	JxW[q_point];
				cell_mass_matrix_E(i,j) +=  eps * E_j * E_i	*	JxW[q_point];
			}
		}
}


template class DGMaxwellLocal<3>;
}
} /* namespace MaxwellProblem */
