#include "CavitySolution.h"

#include "Eps.h"
#include "Mu.h"

namespace MaxwellProblem {
namespace Data{

template <int dim>
CavitySolution_H<dim>::CavitySolution_H(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time),
	pi (3.14159265358979)
{
    nx = 2; 		ny = 2; 		nz = 2;
    ax = 1;			ay = 1;			az = 1;
    kx = pi*nx/ax;	ky = pi*ny/ay;	kz = pi*nz/az;
    Ex = -1;		Ey = 0;         Ez = 1;
    k = sqrt(kx*kx + ky*ky + kz*kz);

    const Mu<dim> 	mu_function;
    const Eps<dim> 	eps_function;
    const Point<dim>  p;
    mu = mu_function.value(p);
    eps= eps_function.value(p);

    s_o_l = 1. / sqrt(mu * eps);
    omega = s_o_l*k;

    sin_omega_t = sin(omega*current_time);
}

template <int dim>
void CavitySolution_H<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) = -(s_o_l/k) * (Ez*ky-Ey*kz)	* sin(kx*p[0]) * cos(ky*p[1]) * cos(kz*p[2]) * sin_omega_t;
    values(1) = -(s_o_l/k) * (Ex*kz-Ez*kx)	* cos(kx*p[0]) * sin(ky*p[1]) *	cos(kz*p[2]) * sin_omega_t;
    values(2) = -(s_o_l/k) * (Ey*kx-Ex*ky)	* cos(kx*p[0]) * cos(ky*p[1]) *	sin(kz*p[2]) * sin_omega_t;
}

template <int dim>
void CavitySolution_H<dim>::vector_value_list(const std::vector<Point<dim>>	&points,
										   std::vector<Vector<double>>			&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		CavitySolution_H<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
void CavitySolution_H<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) = -(s_o_l/k) * (Ez*ky-Ey*kz)	* sin(kx*p[0]) * cos(ky*p[1]) * cos(kz*p[2]);
    values(1) = -(s_o_l/k) * (Ex*kz-Ez*kx)	* cos(kx*p[0]) * sin(ky*p[1]) *	cos(kz*p[2]);
    values(2) = -(s_o_l/k) * (Ey*kx-Ex*ky)	* cos(kx*p[0]) * cos(ky*p[1]) *	sin(kz*p[2]);
}

template <int dim>
void CavitySolution_H<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector<Vector<double>>				&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		CavitySolution_H<dim>::space_vector_value (points[p], value_list[p]);
}

template <int dim>
double CavitySolution_H<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;
	sin_omega_t = sin(omega*current_time);

	return sin_omega_t;
}


template <int dim>
CavitySolution_E<dim>::CavitySolution_E(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time),
	pi (3.14159265358979)
{
    nx = 2; 		ny = 2; 		nz = 2;
    ax = 1;			ay = 1;			az = 1;
    kx = pi*nx/ax;	ky = pi*ny/ay;	kz = pi*nz/az;
    Ex = -1;		Ey = 0;         Ez = 1;

    const Mu<dim> 	mu_function;
    const Eps<dim> 	eps_function;
    const Point<dim>  p;
    mu = mu_function.value(p);
    eps= eps_function.value(p);

    s_o_l = 1. / sqrt(mu * eps);
    omega = s_o_l*sqrt(kx*kx + ky*ky + kz*kz);

    cos_omega_t = cos(omega*current_time);
}

template <int dim>
void CavitySolution_E<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>		&values) const
{
    Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) = (1/eps) * Ex	* cos(kx*p[0]) * sin(ky*p[1]) *	sin(kz*p[2]) * cos_omega_t;
    values(1) = (1/eps) * Ey	* sin(kx*p[0]) * cos(ky*p[1]) *	sin(kz*p[2]) * cos_omega_t;
    values(2) = (1/eps) * Ez	* sin(kx*p[0]) * sin(ky*p[1]) *	cos(kz*p[2]) * cos_omega_t;
}

template <int dim>
void CavitySolution_E<dim>::vector_value_list(const std::vector<Point<dim>>	&points,
										   std::vector<Vector<double>>		&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		CavitySolution_E<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
void CavitySolution_E<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) = (1/eps) * Ex	* cos(kx*p[0]) * sin(ky*p[1]) *	sin(kz*p[2]);
    values(1) = (1/eps) * Ey	* sin(kx*p[0]) * cos(ky*p[1]) *	sin(kz*p[2]);
    values(2) = (1/eps) * Ez	* sin(kx*p[0]) * sin(ky*p[1]) *	cos(kz*p[2]);
}

template <int dim>
void CavitySolution_E<dim>::space_vector_value_list(const std::vector< Point<dim>>	&points,
										   std::vector< Vector<double>>				&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		CavitySolution_E<dim>::space_vector_value (points[p], value_list[p]);
}

template <int dim>
double CavitySolution_E<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;
	cos_omega_t = cos(omega*current_time);

	return cos_omega_t;
}

template <int dim>
CavitySolutionDt_H<dim>::CavitySolutionDt_H(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time),
	pi (3.14159265358979)
{
    nx = 2; 		ny = 2; 		nz = 2;
    ax = 1;			ay = 1;			az = 1;
    kx = pi*nx/ax;	ky = pi*ny/ay;	kz = pi*nz/az;
    Ex = -1;		Ey = 0;         Ez = 1;
    k = sqrt(kx*kx + ky*ky + kz*kz);

    const Mu<dim> 	mu_function;
    const Eps<dim> 	eps_function;
    const Point<dim>  p;
    mu = mu_function.value(p);
    eps= eps_function.value(p);

    s_o_l = 1. / sqrt(mu * eps);
    omega = s_o_l*k;

    cos_omega_t = cos(omega*current_time);
}

template <int dim>
void CavitySolutionDt_H<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) = - s_o_l * s_o_l * (Ez*ky-Ey*kz)	* sin(kx*p[0]) * cos(ky*p[1]) * cos(kz*p[2]) * cos_omega_t;
    values(1) = - s_o_l * s_o_l * (Ex*kz-Ez*kx)	* cos(kx*p[0]) * sin(ky*p[1]) *	cos(kz*p[2]) * cos_omega_t;
    values(2) = - s_o_l * s_o_l * (Ey*kx-Ex*ky)	* cos(kx*p[0]) * cos(ky*p[1]) *	sin(kz*p[2]) * cos_omega_t;
}

template <int dim>
void CavitySolutionDt_H<dim>::vector_value_list(const std::vector<Point<dim>>	&points,
										   std::vector<Vector<double>>			&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		CavitySolutionDt_H<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
void CavitySolutionDt_H<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) = - s_o_l * s_o_l * (Ez*ky-Ey*kz)	* sin(kx*p[0]) * cos(ky*p[1]) * cos(kz*p[2]);
    values(1) = - s_o_l * s_o_l * (Ex*kz-Ez*kx)	* cos(kx*p[0]) * sin(ky*p[1]) *	cos(kz*p[2]);
    values(2) = - s_o_l * s_o_l * (Ey*kx-Ex*ky)	* cos(kx*p[0]) * cos(ky*p[1]) *	sin(kz*p[2]);
}

template <int dim>
void CavitySolutionDt_H<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector<Vector<double>>				&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		CavitySolutionDt_H<dim>::space_vector_value (points[p], value_list[p]);
}

template <int dim>
double CavitySolutionDt_H<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;
	cos_omega_t = sin(omega*current_time);

	return cos_omega_t;
}


template <int dim>
CavitySolutionDt_E<dim>::CavitySolutionDt_E(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time),
	pi (3.14159265358979)
{
    nx = 2; 		ny = 2; 		nz = 2;
    ax = 1;			ay = 1;			az = 1;
    kx = pi*nx/ax;	ky = pi*ny/ay;	kz = pi*nz/az;
    Ex = -1;		Ey = 0;         Ez = 1;

    const Mu<dim> 	mu_function;
    const Eps<dim> 	eps_function;
    const Point<dim>  p;
    mu = mu_function.value(p);
    eps= eps_function.value(p);

    s_o_l = 1. / sqrt(mu * eps);
    omega = s_o_l*sqrt(kx*kx + ky*ky + kz*kz);

    omega_sin_omega_t = -omega*sin(omega*current_time);
}

template <int dim>
void CavitySolutionDt_E<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>		&values) const
{
    Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) = (1/eps) * Ex	* cos(kx*p[0]) * sin(ky*p[1]) *	sin(kz*p[2]) * omega_sin_omega_t;
    values(1) = (1/eps) * Ey	* sin(kx*p[0]) * cos(ky*p[1]) *	sin(kz*p[2]) * omega_sin_omega_t;
    values(2) = (1/eps) * Ez	* sin(kx*p[0]) * sin(ky*p[1]) *	cos(kz*p[2]) * omega_sin_omega_t;
}

template <int dim>
void CavitySolutionDt_E<dim>::vector_value_list(const std::vector<Point<dim>>	&points,
										   std::vector<Vector<double>>		&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		CavitySolutionDt_E<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
void CavitySolutionDt_E<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) = (1/eps) * Ex	* cos(kx*p[0]) * sin(ky*p[1]) *	sin(kz*p[2]);
    values(1) = (1/eps) * Ey	* sin(kx*p[0]) * cos(ky*p[1]) *	sin(kz*p[2]);
    values(2) = (1/eps) * Ez	* sin(kx*p[0]) * sin(ky*p[1]) *	cos(kz*p[2]);
}

template <int dim>
void CavitySolutionDt_E<dim>::space_vector_value_list(const std::vector< Point<dim>>	&points,
										   std::vector< Vector<double>>				&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		CavitySolutionDt_E<dim>::space_vector_value (points[p], value_list[p]);
}

template <int dim>
double CavitySolutionDt_E<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;
	omega_sin_omega_t = cos(omega*current_time);

	return omega_sin_omega_t;
}


template class CavitySolution_H<3>;
template class CavitySolution_E<3>;

template class CavitySolutionDt_H<3>;
template class CavitySolutionDt_E<3>;

} /* namespace Data */
} /* namespace MaxwellProblem */
