/*
 * ExactSolution.cpp
 *
 *  Created on: Mar 22, 2017
 *      Author: carle, leibold
 */

#include "ExactSolution.h"

namespace MaxwellProblem {
namespace Data{
	
// several functions here are just the three component zero-vector field, 
// as is the H-field of the polynomial solution.
// Hence, we introduce this more descriptive abbreviation.
template <int dim>
using ZeroFunction = PolSolution_H<dim>; 
	
template <int dim>
ExactSolution_H<dim>::ExactSolution_H(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time)
{
	if(MaxwellProblem::global::config.get("Data.TestCase","") == "CavitySolution")
	{
		solution_H = new CavitySolution_H<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "LipschSolution")
	{
		solution_H = new LipschSolution_H<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution")
	{
		solution_H = new ZeroFunction<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "JumpSolution")
	{
		solution_H = new JumpSolution_H<dim>(current_time);
	}
	else
	{
		solution_H = nullptr;
        DeclExceptionMsg(ExcExactSolution, "Error in parameter [Data.TestCase]: Unknown Input.");
		AssertThrow(false, ExcExactSolution());
	}
}

template <int dim>
void ExactSolution_H<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());

    solution_H->vector_value(p, values);
}

template <int dim>
void ExactSolution_H<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	solution_H->vector_value_list(points, value_list);
}

template <int dim>
void ExactSolution_H<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	solution_H->space_vector_value(p, values);
}

template <int dim>
void ExactSolution_H<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector<Vector<double>>				&value_list) const
{
	solution_H->space_vector_value_list(points, value_list);
}

template <int dim>
double ExactSolution_H<dim>::get_time_value(double current_time)
{
	return solution_H->get_time_value(current_time);
}



template <int dim>
ExactSolution_E<dim>::ExactSolution_E(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time)
{
	if(MaxwellProblem::global::config.get("Data.TestCase","") == "CavitySolution")
	{
		solution_E = new CavitySolution_E<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "LipschSolution")
	{
		solution_E = new LipschSolution_E<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution")
	{
		solution_E = new PolSolution_E<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "JumpSolution")
	{
		solution_E = new JumpSolution_E<dim>(current_time);
	}
	else
	{
		solution_E = nullptr;
        DeclExceptionMsg(ExcExactSolution, "Error in parameter [Data.TestCase]: Unknown Input.");
		AssertThrow(false, ExcExactSolution());
	}
}

template <int dim>
void ExactSolution_E<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());

    solution_E->vector_value(p, values);
}

template <int dim>
void ExactSolution_E<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	solution_E->vector_value_list(points, value_list);
}

template <int dim>
void ExactSolution_E<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	solution_E->space_vector_value(p, values);
}

template <int dim>
void ExactSolution_E<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector<Vector<double>>				&value_list) const
{
	solution_E->space_vector_value_list(points, value_list);
}

template <int dim>
double ExactSolution_E<dim>::get_time_value(double current_time)
{
	return solution_E->get_time_value(current_time);
}



template <int dim>
ExactSolution<dim>::ExactSolution(double current_time)
	:
	Function<dim>(2*dim, current_time), // Only for dim=3
	current_time(current_time),
	solution_H(current_time),
	solution_E(current_time)
{}

template <int dim>
void ExactSolution<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>	&values) const
{
//     Assert (values.size() == 2*dim, ExcDimensionMismatch (values.size(), 2*dim));
    Assert (dim >= 2, ExcNotImplemented());

    Vector<double> tmp_values(dim);
    solution_H.vector_value(p, tmp_values);
    for (int i=0; i<dim; i++)
    	values[i]=tmp_values[i];
    solution_E.vector_value(p, tmp_values);
    for (int i=0; i<dim; i++)
    	values[i+dim]=tmp_values[i];
}

template <int dim>
void ExactSolution<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector< Vector<double> >	&value_list) const
{
// 	Assert (value_list.size() == 2*points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		ExactSolution<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
ExactSolutionCurl_H<dim>::ExactSolutionCurl_H(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time)
{
	if(MaxwellProblem::global::config.get("Data.TestCase","") == "CavitySolution")
	{
		// we use CavitySolutionDt instead of ..Curl here, as they are equal
		// since we have a homogeneous problem
		solution_H = new CavitySolutionDt_H<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "LipschSolution")
	{
		// we use LipschSolutionDt instead of ..Curl here, as they are equal
		// since we have a homogeneous problem
		solution_H = new LipschSolutionDt_H<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution")
	{
		// The curl of the Polynomial solution is zero,
		// hence, we use the H-component of the solution
		// itself, which is zero as well.
 		solution_H = new ZeroFunction<dim>();
	}
	else
	{
		solution_H = nullptr;
        DeclExceptionMsg(ExcExactSolution, "Error in parameter [Data.TestCase]: Unknown Input.");
		AssertThrow(false, ExcExactSolution());
	}
}

template <int dim>
void ExactSolutionCurl_H<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());

    solution_H->vector_value(p, values);
}

template <int dim>
void ExactSolutionCurl_H<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	solution_H->vector_value_list(points, value_list);
}

template <int dim>
void ExactSolutionCurl_H<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	solution_H->space_vector_value(p, values);
}

template <int dim>
void ExactSolutionCurl_H<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector<Vector<double>>				&value_list) const
{
	solution_H->space_vector_value_list(points, value_list);
}

template <int dim>
double ExactSolutionCurl_H<dim>::get_time_value(double current_time)
{
	return solution_H->get_time_value(current_time);
}



template <int dim>
ExactSolutionCurl_E<dim>::ExactSolutionCurl_E(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time)
{
	if(MaxwellProblem::global::config.get("Data.TestCase","") == "CavitySolution")
	{
		// we use ExactSolutionDt instead of ..Curl here, as they are equal
		// since we have a homogeneous problem
		solution_E = new CavitySolutionDt_E<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "LipschSolution")
	{
		// we use ExactSolutionDt instead of ..Curl here, as they are equal
		// since we have a homogeneous problem
		solution_E = new LipschSolutionDt_E<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution")
	{
		// The curl of the Polynomial solution is zero,
		// hence, we use the H-component of the solution
		// itself, which is zero as well.
		solution_E = new ZeroFunction<dim>();
	}
	else
	{
		solution_E = nullptr;
        DeclExceptionMsg(ExcExactSolution, "Error in parameter [Data.TestCase]: Unknown Input.");
		AssertThrow(false, ExcExactSolution());
	}
}

template <int dim>
void ExactSolutionCurl_E<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());

    solution_E->vector_value(p, values);
}

template <int dim>
void ExactSolutionCurl_E<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	solution_E->vector_value_list(points, value_list);
}

template <int dim>
void ExactSolutionCurl_E<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	solution_E->space_vector_value(p, values);
}

template <int dim>
void ExactSolutionCurl_E<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector<Vector<double>>				&value_list) const
{
	solution_E->space_vector_value_list(points, value_list);
}

template <int dim>
double ExactSolutionCurl_E<dim>::get_time_value(double current_time)
{
	return solution_E->get_time_value(current_time);
}



template <int dim>
ExactSolutionCurl<dim>::ExactSolutionCurl(double current_time)
	:
	Function<dim>(2*dim, current_time), // Only for dim=3
	current_time(current_time),
	solution_H(current_time),
	solution_E(current_time)
{}

template <int dim>
void ExactSolutionCurl<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>	&values) const
{
    Assert (values.size() == 2*dim, ExcDimensionMismatch (values.size(), dim)); // Error for TM,TE?
    Assert (dim >= 2, ExcNotImplemented());

    Vector<double> tmp_values(dim);
    solution_H.vector_value(p, tmp_values);
    for (int i=0; i<dim; i++)
    	values[i]=tmp_values[i];
    solution_E.vector_value(p, tmp_values);
    for (int i=0; i<dim; i++)
    	values[i+dim]=tmp_values[i];
}

template <int dim>
void ExactSolutionCurl<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		ExactSolutionCurl<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
ExactSolutionDt_H<dim>::ExactSolutionDt_H(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time)
{
	if(MaxwellProblem::global::config.get("Data.TestCase","") == "CavitySolution")
	{
		solution_H = new CavitySolutionDt_H<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "LipschSolution")
	{
		solution_H = new LipschSolutionDt_H<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution")
	{
		// The H-field of the polynomial solution is zero
		// at all time. Hence, the same holds for its time
		// derivative, which is why we use the same class. 
// 		solution_H = new PolRHS<dim>(current_time);
		solution_H = new ZeroFunction<dim>(current_time);
	}
	else
	{
		solution_H = nullptr;
        DeclExceptionMsg(ExcExactSolution, "Error in parameter [Data.TestCase]: Unknown Input.");
		AssertThrow(false, ExcExactSolution());
	}
}

template <int dim>
void ExactSolutionDt_H<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());

    solution_H->vector_value(p, values);
}

template <int dim>
void ExactSolutionDt_H<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	solution_H->vector_value_list(points, value_list);
}

template <int dim>
void ExactSolutionDt_H<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	solution_H->space_vector_value(p, values);
}

template <int dim>
void ExactSolutionDt_H<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector<Vector<double>>				&value_list) const
{
	solution_H->space_vector_value_list(points, value_list);
}

template <int dim>
double ExactSolutionDt_H<dim>::get_time_value(double current_time)
{
	return solution_H->get_time_value(current_time);
}



template <int dim>
ExactSolutionDt_E<dim>::ExactSolutionDt_E(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time)
{
	if(MaxwellProblem::global::config.get("Data.TestCase","") == "CavitySolution")
	{
		solution_E = new CavitySolutionDt_E<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "LipschSolution")
	{
		solution_E = new LipschSolutionDt_E<dim>(current_time);
	}
	else if(MaxwellProblem::global::config.get("Data.TestCase","") == "PolSolution")
	{
// 		solution_E = new ZeroFunction<dim>(current_time);
		solution_E = new PolRHS<dim>(current_time);
	}
	else
	{
		solution_E = nullptr;
        DeclExceptionMsg(ExcExactSolution, "Error in parameter [Data.TestCase]: Unknown Input.");
		AssertThrow(false, ExcExactSolution());
	}
}

template <int dim>
void ExactSolutionDt_E<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());

    solution_E->vector_value(p, values);
}

template <int dim>
void ExactSolutionDt_E<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	solution_E->vector_value_list(points, value_list);
}

template <int dim>
void ExactSolutionDt_E<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	solution_E->space_vector_value(p, values);
}

template <int dim>
void ExactSolutionDt_E<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector<Vector<double>>				&value_list) const
{
	solution_E->space_vector_value_list(points, value_list);
}

template <int dim>
double ExactSolutionDt_E<dim>::get_time_value(double current_time)
{
	return solution_E->get_time_value(current_time);
}



template <int dim>
ExactSolutionDt<dim>::ExactSolutionDt(double current_time)
	:
	Function<dim>(2*dim, current_time), // Only for dim=3
	current_time(current_time),
	solution_H(current_time),
	solution_E(current_time)
{}

template <int dim>
void ExactSolutionDt<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>	&values) const
{
    Assert (values.size() == 2*dim, ExcDimensionMismatch (values.size(), dim)); // Error for TM,TE?
    Assert (dim >= 2, ExcNotImplemented());

    Vector<double> tmp_values(dim);
    solution_H.vector_value(p, tmp_values);
    for (int i=0; i<dim; i++)
    	values[i]=tmp_values[i];
    solution_E.vector_value(p, tmp_values);
    for (int i=0; i<dim; i++)
    	values[i+dim]=tmp_values[i];
}

template <int dim>
void ExactSolutionDt<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		ExactSolutionDt<dim>::vector_value (points[p], value_list[p]);
}

template class ExactSolution_H<3>;
template class ExactSolution_E<3>;

template class ExactSolution<3>;


template class ExactSolutionCurl_H<3>;
template class ExactSolutionCurl_E<3>;

template class ExactSolutionCurl<3>;


template class ExactSolutionDt_H<3>;
template class ExactSolutionDt_E<3>;

template class ExactSolutionDt<3>;

} /* namespace Data */
} /* namespace MaxwellProblem */
