/*
 * ExactSolution.h
 *
 *  Created on: Mar 22, 2017
 *      Author: carle, leibold
 */

#ifndef EXACTSOLUTION_H_
#define EXACTSOLUTION_H_


#include <deal.II/base/function.h>
#include <deal.II/lac/vector.h>

#include "FunctionInterface.h"

#include "CavitySolution.h"
#include "PolSolution.h"
#include "JumpSolution.h"
#include "PolRHS.h"
#include "LipschSolution.h"
#include "../global.h"


using namespace dealii;

namespace MaxwellProblem {
namespace Data{

/**
 * \brief Interface class for exact solutions.
 *
 * \note See ExactSolution_E for further information.
 */
template <int dim>
class ExactSolution_H : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 */
	ExactSolution_H 		(double current_time = 0.0);
	/**
	 * \brief Destructor.
	 */
	~ExactSolution_H		() {delete solution_H;}

	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;

	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
	/**
	 * \brief Evaluates the space-function at a single point.
	 *
	 *  This is the factor of the solution that depends only on the space variable.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the time function \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;

	/**
	 * pointer to the exact solution that is used
	 */
	FunctionInterface<dim>* solution_H;
};

/**
 * \brief Interface class for exact solutions.
 *
 * If an instance of this class is created, the constructor reads from the config file of
 * the project what exact solution should be used and generates an instance of this type.
 * Every time the exact solution is evaluated, this class then simply calls the corresponding
 * method of the exact solution that is used.
 *
 * Some solutions can be split in two parts that depends only on one variable (space variable resp. time variable).
 * Those two parts can be evaluated separately, which is used to make assembling of the exact solution more efficient.
 */
template <int dim>
class ExactSolution_E : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 */
	ExactSolution_E 		(double current_time = 0.0);
	/**
	 * \brief Destructor.
	 */
	~ExactSolution_E		() {delete solution_E;}
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
	/**
	 * \brief Evaluates the space-function at a single point.
	 *
	 *  This is the factor of the solution that depends only on the space variable.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the time function \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;

	/**
	 * pointer to the exact solution that is used
	 */
	FunctionInterface<dim>* solution_E;
};

/**
 * \brief Interface class for exact solutions.
 *
 * If an instance of this class is created, the constructor reads from the config file of
 * the project what exact solution should be used and generates an instance of this type.
 * Every time the exact solution is evaluated, this class then simply calls the corresponding
 * method of the exact solution that is used.
 *
 * \note Exact Solution is currently only used in ADI, otherwise directly ExactSolution_H and ExactSolution_E is used.
 */
template <int dim>
class ExactSolution : public Function<dim>
{
public:
	/**
	 * \brief Constructor.
	 */
	ExactSolution 			(double current_time = 0.0);
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;

private:
	/**
	 * variable of time
	 */
	double current_time;
	/**
	 * H-field of the exact solution
	 */
	ExactSolution_H<dim> solution_H;
	/**
	 * E-field of the exact solution
	 */
	ExactSolution_E<dim> solution_E;
};

/**
 * \brief Interface class for exact solutions.
 *
 * \note See ExactSolutionCurl_E for further information.
 */
template <int dim>
class ExactSolutionCurl_H : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 */
	ExactSolutionCurl_H 		(double current_time = 0.0);
	/**
	 * \brief Destructor.
	 */
	~ExactSolutionCurl_H		() {delete solution_H;}

	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;

	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
	/**
	 * \brief Evaluates the space-function at a single point.
	 *
	 *  This is the factor of the solution that depends only on the space variable.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the time function \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;

	/**
	 * pointer to the exact solution that is used
	 */
	FunctionInterface<dim>* solution_H;
};

/**
 * \brief Interface class for exact solutions.
 *
 * If an instance of this class is created, the constructor reads from the config file of
 * the project what exact solution should be used and generates an instance of this type.
 * Every time the exact solution is evaluated, this class then simply calls the corresponding
 * method of the exact solution that is used.
 *
 * Some solutions can be split in two parts that depends only on one variable (space variable resp. time variable).
 * Those two parts can be evaluated separately, which is used to make assembling of the exact solution more efficient.
 */
template <int dim>
class ExactSolutionCurl_E : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 */
	ExactSolutionCurl_E 		(double current_time = 0.0);
	/**
	 * \brief Destructor.
	 */
	~ExactSolutionCurl_E		() {delete solution_E;}
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
	/**
	 * \brief Evaluates the space-function at a single point.
	 *
	 *  This is the factor of the solution that depends only on the space variable.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the time function \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;

	/**
	 * pointer to the exact solution that is used
	 */
	FunctionInterface<dim>* solution_E;
};

/**
 * \brief Interface class for exact solutions.
 *
 * If an instance of this class is created, the constructor reads from the config file of
 * the project what exact solution should be used and generates an instance of this type.
 * Every time the exact solution is evaluated, this class then simply calls the corresponding
 * method of the exact solution that is used.
 *
 * \note Exact Solution is at the moment only used in ADI, otherwise directly ExactSolutionCurl_H and ExactSolutionCurl_E is used.
 */
template <int dim>
class ExactSolutionCurl : public Function<dim>
{
public:
	/**
	 * \brief Constructor.
	 */
	ExactSolutionCurl 			(double current_time = 0.0);
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;

private:
	/**
	 * variable of time
	 */
	double current_time;
	/**
	 * H-field of the exact solution
	 */
	ExactSolutionCurl_H<dim> solution_H;
	/**
	 * E-field of the exact solution
	 */
	ExactSolutionCurl_E<dim> solution_E;
};

/**
 * \brief Interface class for exact solutions.
 *
 * \note See ExactSolutionDt_E for further information.
 */
template <int dim>
class ExactSolutionDt_H : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 */
	ExactSolutionDt_H 		(double current_time = 0.0);
	/**
	 * \brief Destructor.
	 */
	~ExactSolutionDt_H		() {delete solution_H;}

	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;

	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
	/**
	 * \brief Evaluates the space-function at a single point.
	 *
	 *  This is the factor of the solution that depends only on the space variable.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the time function \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;

	/**
	 * pointer to the exact solution that is used
	 */
	FunctionInterface<dim>* solution_H;
};

/**
 * \brief Interface class for exact solutions.
 *
 * If an instance of this class is created, the constructor reads from the config file of
 * the project what exact solution should be used and generates an instance of this type.
 * Every time the exact solution is evaluated, this class then simply calls the corresponding
 * method of the exact solution that is used.
 *
 * Some solutions can be split in two parts that depends only on one variable (space variable resp. time variable).
 * Those two parts can be evaluated separately, which is used to make assembling of the exact solution more efficient.
 */
template <int dim>
class ExactSolutionDt_E : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 */
	ExactSolutionDt_E 		(double current_time = 0.0);
	/**
	 * \brief Destructor.
	 */
	~ExactSolutionDt_E		() {delete solution_E;}
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
	/**
	 * \brief Evaluates the space-function at a single point.
	 *
	 *  This is the factor of the solution that depends only on the space variable.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the time function \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;

	/**
	 * pointer to the exact solution that is used
	 */
	FunctionInterface<dim>* solution_E;
};

/**
 * \brief Interface class for exact solutions.
 *
 * If an instance of this class is created, the constructor reads from the config file of
 * the project what exact solution should be used and generates an instance of this type.
 * Every time the exact solution is evaluated, this class then simply calls the corresponding
 * method of the exact solution that is used.
 *
 * \note Exact Solution is at the moment only used in ADI, otherwise directly ExactSolutionDt_H and ExactSolutionDt_E is used.
 */
template <int dim>
class ExactSolutionDt : public Function<dim>
{
public:
	/**
	 * \brief Constructor.
	 */
	ExactSolutionDt 			(double current_time = 0.0);
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;

private:
	/**
	 * variable of time
	 */
	double current_time;
	/**
	 * H-field of the exact solution
	 */
	ExactSolutionDt_H<dim> solution_H;
	/**
	 * E-field of the exact solution
	 */
	ExactSolutionDt_E<dim> solution_E;
};

} /* namespace Data */
} /* namespace MaxwellProblem */

#endif /* EXACTSOLUTION_H_ */
