/*
 * FunctionInterface.h
 *
 *  Created on: 31.03.2017
 *      Author: carle
 */

#ifndef DATA_FUNCTIONINTERFACE_H_
#define DATA_FUNCTIONINTERFACE_H_


#include <deal.II/base/function.h>

using namespace dealii;

namespace MaxwellProblem {
namespace Data{

/**
 * \brief Base class of functions used in this project.
 *
 * This class is only a interface with the typical structure of functions used in this project.
 * Other functions can inherit from this base class, then this type can be used as placeholder if to compile time it is
 * not clear which function exactly should be used (e.g. which exact solution is used). Therefore all the methods are declared virtual.
 *
 * Why do not simply inherit from the deal.ii class Function?
 * Because some functions in this project can be splittet in two parts that only depends on the space variable resp. the time variable
 * what is not provided by the Function interface. Methods with support this are therefore provided here.
 */
template <int dim>
class FunctionInterface : public Function<dim>
{
public:
	/**
	 * \brief Constructor.
	 */
	FunctionInterface<dim>			(double current_time)
	:Function<dim>(dim, current_time)
	{};

	/**
	 * \brief Destructor.
	 */
	virtual ~FunctionInterface				() 	{};

	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	virtual void space_vector_value 		(const Point<dim>		&p,
	 	 	 	 	 	 	 	 	 	 	 Vector<double>			&values) 	 const = 0;
	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	virtual void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 	 	 std::vector<Vector<double>>	&value_list) const = 0;

	/**
	 * \brief Evaluates the polynomial \f$ p \f$ at \p current_time.
	 */
	virtual double get_time_value			(double current_time) = 0;

};

} /* namespace Data */
} /* namespace MaxwellProblem */

#endif /* DATA_FUNCTIONINTERFACE_H_ */
