/*
 * GradientField.h
 *
 *  Created on: 29.03.2017
 *      Author: leibold
 */

#ifndef DATA_GRADIENTFIELD_H_
#define DATA_GRADIENTFIELD_H_

#include <deal.II/base/function.h>
#include <deal.II/lac/vector.h>

using namespace dealii;

namespace MaxwellProblem {
namespace Data{


/**
 * \brief Gradient field function
 *
 * This function is the gradient field of \f$ \Phi:\mathbb{R}^3 \rightarrow \mathbb{R}, (x_1,x_2,x_3)\mapsto (x_1^2-2x1)(x_2^2-x_2)(x_3^2-x_3) \f$, i.e.
 * \f[
 * \nabla\Phi(x_1,x_2,x_3)=\begin{pmatrix} (2x_1 - 2)(x_2^2-x_2)(x_3^2-x_3)\\ (x_1^2-2x_1)(2x_2-1)(x_3^2-x_3) \\ (x_1^2-2x_1)(x_2^2-x_2)(2x_3-x_1) \end{pmatrix}.
 * \f]
 *
 * It satisfies \f$ n \times \nabla \Phi = 0\f$ on \f$ \partial \Omega \f$, i.e. the boundary conditions of the Maxwells equations.
 * Because it's a gradient field, it is divergence free.
 * This gradient field is used in PolSolution_H / PolSolution_E to construct a solution of the Maxwells equations which is polynomial in time.
 */
template <int dim>
class GradientField : public Function<dim> {
public:
	GradientField			();
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;

};



} /* namespace Data */
} /* namespace MaxwellProblem */

#endif /* DATA_GRADIENTFIELD_H_ */
