/*
 * GridGeneration.cpp
 *
 *  Created on: Jun 21, 2016
 *      Author: koehler
 */

#ifndef GRIDGEN
#define GRIDGEN

#include <deal.II/grid/tria.h>
#include <deal.II/grid/grid_generator.h>
#include <deal.II/grid/grid_out.h>
#include <deal.II/grid/grid_tools.h>
#include <deal.II/dofs/dof_renumbering.h>
#include <math.h>

#include "../global.h"


namespace MaxwellProblem {
namespace data{

using namespace dealii;

/**
 * \brief Generates the grid.
 * @param refinement_lvl: number of regular refinements starting from a cube
 */
template <int dim>
void generate_cuboid(Triangulation<dim> &	triangulation,
					 const unsigned int		refinement_lvl)
{
	Point<dim> p1,p2;
	p1(0)=0;
	
	if ( MaxwellProblem::global::config.get("Data.TestCase","") == "JumpSolution" )
		p2(0)=1.;
	else
		p2(0)=2.;
	p2(1)=1.;
	if (dim == 3)
		p2(2)=1.;

	std::vector<unsigned int> repetitions(dim,1);

	if ( MaxwellProblem::global::config.get("Data.TestCase","") == "JumpSolution" )
		repetitions[0]=4*refinement_lvl;
	else
		repetitions[0]=2*refinement_lvl;
	
	for (unsigned int i=1; i<dim; ++i)
		repetitions[i]=refinement_lvl;

	GridGenerator::subdivided_hyper_rectangle(triangulation,
											  repetitions,
											  p1,
											  p2);
}

/**
 * \brief Generates a local refined grid for grids with a few cells.
 *
 * @param refinement_lvl_coarse: number of regular grid refinements starting from a cube
 * (only even integers are allowed!))
 * @param refinement_factor_fine: number of fine refinements starting from the coarse mesh
 * (default is zero)
 * @param ratio_fine_total: ratio of fine cells to total cells (not exact!)
 */
template <int dim>
void generate_cuboid_local_refined_edge(Triangulation<dim> &	triangulation,
					 const unsigned int		refinement_lvl_coarse,
					 const unsigned int		refinement_factor_fine = 0,
					 const double 			ratio_fine_coarse = 0.1)
{
	Point<dim> p1,p2;
	p1(0)=0;

	p2(0)=2.;
	p2(1)=1.;
	if (dim == 3)
		p2(2)=1.;

	std::vector<unsigned int> repetitions(dim,1);

	repetitions[0]=2*refinement_lvl_coarse;
	for (unsigned int i=1; i<dim; ++i)
		repetitions[i]=refinement_lvl_coarse;

	GridGenerator::subdivided_hyper_rectangle ( triangulation,
												repetitions,
												p1,
												p2);

	Point<dim> p_refine;
	p_refine(0) = 1.;
	unsigned int start_cells = triangulation.n_active_cells();
	double h = 1./refinement_lvl_coarse;
	unsigned int n = ceil((ratio_fine_coarse*(start_cells - 1) - 8 - (refinement_factor_fine - 1)*7)/(14*pow(2,refinement_factor_fine - 1)-6+ratio_fine_coarse));
	p_refine(0) = n*h;
	for (unsigned int j=1; j<=refinement_factor_fine; ++j )
	{
		typename Triangulation<dim>::active_cell_iterator
			cell = triangulation.begin_active(),
			endc = triangulation.end();
		for (; cell!=endc; ++cell)
		{
			for (unsigned int v=0; v < GeometryInfo<dim>::vertices_per_cell; ++v)
			{
				if (cell->vertex(v)(0) < p_refine(0) && std::fabs(cell->vertex(v)(1)) + std::fabs(cell->vertex(v)(2)) < 1e-10)
				{
					cell->set_refine_flag();
					break;
				}
			}
		}

		triangulation.execute_coarsening_and_refinement();
	}
}

/**
 * \brief Generates a local refined grid for meshes with many cells.
 *
 * @param refinement_lvl_coarse: number of regular grid refinements starting from a cube
 * (only even integers are allowed!))
 * @param refinement_factor_fine: number of fine refinements starting from the coarse mesh
 * (default is zero)
 * @param ratio_fine_total: ratio of fine cells to total cells (not exact!)
 */
template <int dim>
void generate_cuboid_local_refined_face(Triangulation<dim> &	triangulation,
					const unsigned int		refinement_lvl_coarse,
					const unsigned int		refinement_factor_fine = 0,
					const double			ratio_fine_total = 0.05)
{
	// coarse grid
	Point<dim> p1,p2;
	p1(0)=0;

	p2(0)=1.;
	p2(1)=1.;
	if (dim == 3)
		p2(2)=2.;

	std::vector<unsigned int> repetitions(dim,1);
	repetitions[dim-1]=(refinement_lvl_coarse/2)*2;
	for (unsigned int i=0; i<dim-1; ++i)
		repetitions[i]= refinement_lvl_coarse/2;

	GridGenerator::subdivided_hyper_rectangle(triangulation,
			repetitions, p1, p2);
	triangulation.refine_global(1);

	// refinement of the coarse grid
	if (refinement_factor_fine != 0)
	{
		unsigned int total_cells = triangulation.n_active_cells();
		unsigned int fine_cells = ratio_fine_total * total_cells;

		int difference(fine_cells);
		unsigned int computed_cells(0);
		unsigned int fine_cells_real(0);
		unsigned int m_cells(0);
		unsigned int n_cells(0);

		const unsigned int &f = refinement_factor_fine;

		for(unsigned int m=1; m<(refinement_lvl_coarse-1); ++m)
			for(unsigned int n=1; n<(refinement_lvl_coarse-1); ++n)
			{
				if (f==1)
					computed_cells = 2*(8*m*n);
				else
				{
					computed_cells
					= 2*( ((7*int(std::pow(4.,f))-4)/3) * (m*n)
						 + (7*int(std::pow(2.,f-1))-6) * (m+n)
						 + 7*f-6 );
				}

				if (std::fabs(int(fine_cells-computed_cells)) < difference)
				{
					difference = std::fabs(fine_cells-computed_cells);
					fine_cells_real = computed_cells;
					m_cells = m;
					n_cells = n;
				}
			}

		{
			DeclException2 (ExcLocalRefinedGrid, int, int,
					<< "Local refinement of grid with that ratio of fine "
					<< "total cells is not possible: expected fine cells = " << arg1
					<< ", real fine cells = " << arg2);
			AssertThrow((std::fabs(int(fine_cells_real-fine_cells)) < (0.1*fine_cells)),
						ExcLocalRefinedGrid(fine_cells, fine_cells_real) );
		}

		double cell_length = 1./refinement_lvl_coarse;
		double p_refine_z = 1.5;
		double p_refine_x = m_cells*cell_length;
		double p_refine_y = n_cells*cell_length;

		for(unsigned int j=1; j<=refinement_factor_fine; ++j)
		{
			typename Triangulation<dim>::active_cell_iterator
				cell = triangulation.begin_active(),
				endc = triangulation.end();
			for(; cell!=endc; ++cell)
			{
				for(unsigned int v=0; v < GeometryInfo<dim>::faces_per_cell; ++v)
				{
					typename Triangulation<dim>::face_iterator face = cell->face(v);
					if ( std::fabs(face->center()(2)- p_refine_z) < 1e-10
							&& face->center()(0) < p_refine_x
							&& face->center()(1) < p_refine_y )
					{
						cell->set_refine_flag();
						break;
					}
				}
			}

			triangulation.execute_coarsening_and_refinement();
		}

		// correct the number of total cells
		{
			unsigned int cells_to_coarsening(0);
			if (refinement_factor_fine==1)
				cells_to_coarsening = fine_cells_real;
			else
				cells_to_coarsening = ((fine_cells_real-2*(m_cells+1)*(n_cells+1))*8)/7;
			unsigned int j=0;
			typename Triangulation<dim>::active_cell_iterator
				cell = triangulation.begin_active(),
				endc = triangulation.end();
			for(; cell!=endc; ++cell)
			{
				if (cell->center()(2) < 1. && j < cells_to_coarsening)
				{
					cell->set_coarsen_flag();
					++j;
				}
			}

			triangulation.execute_coarsening_and_refinement();
		}
	}
}

template <int dim>
void set_cells_explicit_implicit(Triangulation<dim>& triangulation,
								 const double min_dia_explicit)
{
	// implicit cells (fine and coarse)
	typename Triangulation<dim>::active_cell_iterator
				cell = triangulation.begin_active(),
				endc = triangulation.end();
	for (; cell!=endc; ++cell)
	{
		if ( (min_dia_explicit - cell->diameter()) > 1e-6)
		{
			cell->set_user_index(4);
			// Only for checks with paraview
			cell->set_material_id(4);
		}
		else 
		{
			// set cell explicit first; if necessary, it is changed 
			cell->set_user_index(1);
			// Only for checks with paraview
			cell->set_material_id(1);
			
			for (unsigned int face_no=0; face_no < GeometryInfo<dim>::faces_per_cell; ++face_no)
			{
				typename Triangulation<dim>::face_iterator face = cell->face(face_no);
				if (!(face->at_boundary()))
				{
					Assert (cell->neighbor(face_no).state() == IteratorState::valid, ExcInternalError());
					typename Triangulation<dim>::cell_iterator neighbor = cell->neighbor(face_no);
					if (face->has_children())
					{
						for (unsigned int subface_no=0; subface_no < face->number_of_children(); ++subface_no)
						{
							typename Triangulation<dim>::cell_iterator neighbor_child
								= cell->neighbor_child_on_subface (face_no, subface_no);
							Assert (!neighbor_child->has_children(), ExcInternalError());
							
							if ( (min_dia_explicit - neighbor_child->diameter()) > 1e-6 )
							{
								cell->set_user_index(3);
								// Only for checks with paraview
								cell->set_material_id(3);
								
								break;
							}
						}
					}
					else if ( (min_dia_explicit - neighbor->diameter()) > 1e-6 )
					{
						cell->set_user_index(3);
						// Only for checks with paraview
						cell->set_material_id(3);
					}

					if (cell->user_index()==3)
						break;
				}
			}
		}
	}

	// explicit cells with implicit neighbors
	typename Triangulation<dim>::active_cell_iterator
				cell2 = triangulation.begin_active(),
				endc2 = triangulation.end();
	for (; cell2!=endc2; ++cell2)
	{
		if (cell2->user_index()==1)
		{
			for (unsigned int face_no=0; face_no<GeometryInfo<dim>::faces_per_cell; ++face_no)
			{
				typename Triangulation<dim>::face_iterator face = cell2->face(face_no);
				if (!(face->at_boundary()))
				{
					Assert (cell2->neighbor(face_no).state() == IteratorState::valid, ExcInternalError());
					if (face->has_children())
					{
						for (unsigned int subface_no=0; subface_no < face->number_of_children(); ++subface_no)
						{
							typename Triangulation<dim>::cell_iterator neighbor_child
								= cell2->neighbor_child_on_subface (face_no, subface_no);
							Assert (!neighbor_child->has_children(), ExcInternalError());
							
							if ( neighbor_child->user_index() == 3)
							{
								cell2->set_user_index(2);
								// Only for checks with paraview
								cell2->set_material_id(2);
								
								break;
							}
						}
					}
					else if (cell2->neighbor(face_no)->user_index() == 3)
					{
						cell2->set_user_index(2);
						// Only for checks with paraview
						cell2->set_material_id(2);
					}

					if (cell2->user_index()==2)
						break;
				}
			}
		}
	}
}

template <int dim>
void get_number_dofs_exp_imp (Triangulation<dim>& 			triangulation,
							DoFHandler<dim>& 				dof_handler,
							std::vector<unsigned int>&		number_imp_exp_dofs)
{
	// Resize vector
	number_imp_exp_dofs.resize (4, 0);

	typename Triangulation<dim>::active_cell_iterator
			cell = triangulation.begin_active(),
			endc = triangulation.end();
	for (; cell!=endc; ++cell)
		number_imp_exp_dofs[(cell->user_index()) - 1]++;

	std::cout << "Number of \n"
			 "implicit fine cells: " 	<< number_imp_exp_dofs[3] << "\n"
			 "implicit coarse cells: "	<< number_imp_exp_dofs[2] << "\n"
			 "explicit cells with implicit neighbor cells: " << number_imp_exp_dofs[1] << "\n"
			 "explicit cells without implicit neighbor cells: " << number_imp_exp_dofs[0] << std::endl << std::endl;

	// Change in case of upwind fluxes and different polynomial degrees!
	for(auto &j : number_imp_exp_dofs)
		j *= dof_handler.get_fe().dofs_per_cell;
}


namespace comparator
{
	struct
	{
		template <class DHCellIterator>
		bool operator()(const DHCellIterator &c1, const DHCellIterator &c2) const
		{
			return (c1->user_index() < c2->user_index() );
		}
	}comp;
}

template <int dim>
void renumbering_dofs_explicit_to_implicit (
		DoFHandler<dim>& 			dof_handler)
{
	std::vector<typename DoFHandler<dim>::active_cell_iterator> ordered_cells;
	ordered_cells.reserve(dof_handler.get_triangulation().n_active_cells());

	typename DoFHandler<dim>::active_cell_iterator
		p = dof_handler.begin_active(),
		end = dof_handler.end();
	while(p!=end)
	{
		ordered_cells.push_back(p);
		++p;
	}

	std::sort(ordered_cells.begin(), ordered_cells.end(), comparator::comp);
	// stable_sort preserves the order of the elements -> better for DG?
	//std::stable_sort(ordered_cells.begin(), ordered_cells.end(), comparator::comp);

	DoFRenumbering::cell_wise(dof_handler, ordered_cells);
}

template<int dim>
void output_grid(Triangulation<dim>& triangulation)
{
	std::ofstream out("output/grid.vtk");
	GridOut grid_out;
	grid_out.write_vtk(triangulation, out);
	std::cout << "Grid written to output/grid.vtk" << std::endl << std::endl;
}


} /* namespace Data */
} /* namespace MaxwellProblem */

#endif /* GRIDGEN */
