/*
 * JumpRHS.h
 *
 *  Created on: Mar 26, 2017
 *      Author: leibold
 */

#ifndef JUMPRHS_H_
#define JUMPRHS_H_


#include <deal.II/lac/vector.h>

#include "FunctionInterface.h"
#include "../global.h"
#include "JumpSolution.h"

using namespace dealii;

namespace MaxwellProblem {
namespace Data{

template <int dim>
class JumpRHSLeft : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 *
	 * \param current_time: time at which the function should is evaluated
	 */
	JumpRHSLeft 					(double current_time = 0.0);
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the polynomial \f$ p' \f$ at \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;
	
	const double pi;
	
	double sin_t;
	
	JumpSolutionLeft_E<dim> sol_E;
};


template <int dim>
class JumpRHSRight : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 *
	 * \param current_time: time at which the function should is evaluated
	 */
	JumpRHSRight 					(double current_time = 0.0);
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the polynomial \f$ p' \f$ at \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;
	
	const double pi;
	
	double sin_t;
	
	JumpSolutionRight_E<dim> sol_E;
};


template <int dim>
class ADIJumpRHSLeft : public Function<dim>
{
public:
	/**
	 * \brief Constructor.
	 *
	 * \param current_time: time at which the function should is evaluated
	 */
	ADIJumpRHSLeft 					(double current_time = 0.0);
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
// 	/**
// 	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at a single point.
// 	 *
// 	 * 	This routine takes a point in the dim-dimensional space and
// 	 * 	writes the corresponding function-values at that point into
// 	 * 	the values-parameter.
// 	 */
// 	void space_vector_value 		(const Point<dim>				&p,
// 	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
// 	/**
// 	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at multiple points.
// 	 *
// 	 * 	This routine takes a point in the dim-dimensional space and
// 	 * 	writes the corresponding function-values at that point into
// 	 * 	the values-parameter.
// 	 */
// 	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
// 									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the polynomial \f$ p' \f$ at \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;
	
	JumpRHSLeft<dim> rhs_E;
};


template <int dim>
class ADIJumpRHSRight : public Function<dim>
{
public:
	/**
	 * \brief Constructor.
	 *
	 * \param current_time: time at which the function should is evaluated
	 */
	ADIJumpRHSRight 					(double current_time = 0.0);
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
// 	/**
// 	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at a single point.
// 	 *
// 	 * 	This routine takes a point in the dim-dimensional space and
// 	 * 	writes the corresponding function-values at that point into
// 	 * 	the values-parameter.
// 	 */
// 	void space_vector_value 		(const Point<dim>				&p,
// 	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
// 	/**
// 	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at multiple points.
// 	 *
// 	 * 	This routine takes a point in the dim-dimensional space and
// 	 * 	writes the corresponding function-values at that point into
// 	 * 	the values-parameter.
// 	 */
// 	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
// 									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the polynomial \f$ p' \f$ at \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;
	
	JumpRHSRight<dim> rhs_E;
};

} /* namespace Data */
} /* namespace MaxwellProblem */

#endif /* JUMPRHS_H_ */
