#include "LipschSolution.h"

namespace MaxwellProblem {
namespace Data{

template <int dim>
LipschSolution_H<dim>::LipschSolution_H(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time),
	pi (3.14159265358979)
{
    sin_omega_t = sin((sqrt(2)*pi*current_time));
}

template <int dim>
void LipschSolution_H<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) =  0;
    values(1) = -1./sqrt(2)	* sin(pi*p[1]) * cos(pi*p[2]) * sin_omega_t;
    values(2) =  1./sqrt(2)	* cos(pi*p[1]) * sin(pi*p[2]) * sin_omega_t;
}

template <int dim>
void LipschSolution_H<dim>::vector_value_list(const std::vector<Point<dim>>	&points,
										   std::vector<Vector<double>>			&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		LipschSolution_H<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
void LipschSolution_H<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) =  0;
    values(1) = -1./sqrt(2)	* sin(pi*p[1]) * cos(pi*p[2]);
    values(2) =  1./sqrt(2)	* cos(pi*p[1]) * sin(pi*p[2]);
}

template <int dim>
void LipschSolution_H<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector<Vector<double>>				&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		LipschSolution_H<dim>::space_vector_value (points[p], value_list[p]);
}

template <int dim>
double LipschSolution_H<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;
	sin_omega_t = sin(omega*current_time);

	return sin_omega_t;
}


template <int dim>
LipschSolution_E<dim>::LipschSolution_E(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time),
	pi (3.14159265358979)
{
    cos_omega_t = cos(sqrt(2)*pi*current_time);
}

template <int dim>
void LipschSolution_E<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>		&values) const
{
    Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
	
    double eps = eps_function.value(p);
	
    values(0) = (1/eps) * sin(pi*p[1]) * sin(pi*p[2]) * cos_omega_t;
    values(1) = 0;
    values(2) = 0;
}

template <int dim>
void LipschSolution_E<dim>::vector_value_list(const std::vector<Point<dim>>	&points,
										   std::vector<Vector<double>>		&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		LipschSolution_E<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
void LipschSolution_E<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    double eps = eps_function.value(p);

    values(0) = (1/eps) * sin(pi*p[1]) * sin(pi*p[2]);
    values(1) = 0;
    values(2) = 0;
}

template <int dim>
void LipschSolution_E<dim>::space_vector_value_list(const std::vector< Point<dim>>	&points,
										   std::vector< Vector<double>>				&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		LipschSolution_E<dim>::space_vector_value (points[p], value_list[p]);
}

template <int dim>
double LipschSolution_E<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;
    cos_omega_t = cos(sqrt(2)*pi*current_time);

	return cos_omega_t;
}

template <int dim>
LipschSolutionDt_H<dim>::LipschSolutionDt_H(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time),
	pi (3.14159265358979)
{
    cos_omega_t = sqrt(2) * pi * cos((sqrt(2)*pi*current_time));
}

template <int dim>
void LipschSolutionDt_H<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) =  0;
    values(1) = -1./sqrt(2)	* sin(pi*p[1]) * cos(pi*p[2]) * cos_omega_t;
    values(2) =  1./sqrt(2)	* cos(pi*p[1]) * sin(pi*p[2]) * cos_omega_t;
}

template <int dim>
void LipschSolutionDt_H<dim>::vector_value_list(const std::vector<Point<dim>>	&points,
										   std::vector<Vector<double>>			&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		LipschSolutionDt_H<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
void LipschSolutionDt_H<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    values(0) =  0;
    values(1) = -1./sqrt(2)	* sin(pi*p[1]) * cos(pi*p[2]);
    values(2) =  1./sqrt(2)	* cos(pi*p[1]) * sin(pi*p[2]);
}

template <int dim>
void LipschSolutionDt_H<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector<Vector<double>>				&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		LipschSolutionDt_H<dim>::space_vector_value (points[p], value_list[p]);
}

template <int dim>
double LipschSolutionDt_H<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;
	cos_omega_t = sqrt(2) * pi * cos((sqrt(2)*pi*current_time));

	return cos_omega_t;
}


template <int dim>
LipschSolutionDt_E<dim>::LipschSolutionDt_E(double current_time)
	:
	FunctionInterface<dim>(current_time),
	current_time(current_time),
	pi (3.14159265358979)
{
    sin_omega_t = -sqrt(2) * pi * sin(sqrt(2)*pi*current_time);
}

template <int dim>
void LipschSolutionDt_E<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>		&values) const
{
    Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
	
    double eps = eps_function.value(p);
	
    values(0) = (1/eps) * sin(pi*p[1]) * sin(pi*p[2]) * sin_omega_t;
    values(1) = 0;
    values(2) = 0;
}

template <int dim>
void LipschSolutionDt_E<dim>::vector_value_list(const std::vector<Point<dim>>	&points,
										   std::vector<Vector<double>>		&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		LipschSolutionDt_E<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
void LipschSolutionDt_E<dim>::space_vector_value(const Point<dim> &p,
											Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    double eps = eps_function.value(p);

    values(0) = (1/eps) * sin(pi*p[1]) * sin(pi*p[2]);
    values(1) = 0;
    values(2) = 0;
}

template <int dim>
void LipschSolutionDt_E<dim>::space_vector_value_list(const std::vector< Point<dim>>	&points,
										   std::vector< Vector<double>>				&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		LipschSolutionDt_E<dim>::space_vector_value (points[p], value_list[p]);
}

template <int dim>
double LipschSolutionDt_E<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;
    sin_omega_t = -sqrt(2) * pi * sin(sqrt(2)*pi*current_time);

	return sin_omega_t;
}

template class LipschSolution_H<3>;
template class LipschSolution_E<3>;

template class LipschSolutionDt_H<3>;
template class LipschSolutionDt_E<3>;

} /* namespace Data */
} /* namespace MaxwellProblem */
