/*
 * PolRHS.cpp
 *
 *  Created on: Mar 26, 2017
 *      Author: leibold
 */

#include "PolRHS.h"

namespace MaxwellProblem {
namespace Data{

template<int dim>
PolRHS<dim>::PolRHS(double current_time)
:
FunctionInterface<dim>(current_time), // Only for dim=3
current_time(current_time),
gradient_field()
{
    std::stringstream ss_c(MaxwellProblem::global::config.get<std::string>("Data.Coefficients"));
    std::string c;
	// Horner scheme
    while (std::getline(ss_c, c, ','))
    {
    	c.erase(remove(c.begin(),c.end(),' '),c.end());
    	coeffs.push_back(boost::lexical_cast<double>(c));
    }
    coeffs.pop_back();
    deg = coeffs.size();
}

template <int dim>
void PolRHS<dim>::vector_value(const Point<dim>	&p,
								Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());

    double value = 0;
    unsigned int n = deg;
    for (std::list<double>::const_iterator coeff=coeffs.begin(); coeff != coeffs.end(); ++coeff)
    {
    	value *= current_time;
    	value -= n*(*coeff);
    	n--;
    }

    gradient_field.vector_value(p,values);
    values *= value;
}

template <int dim>
void PolRHS<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
									std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		PolRHS<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
void PolRHS<dim>::space_vector_value(const Point<dim>	&p,
									Vector<double>	&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    gradient_field.vector_value(p,values);
}

template <int dim>
void PolRHS<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
									std::vector<Vector<double>>			&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		PolRHS<dim>::space_vector_value (points[p], value_list[p]);
}

template <int dim>
double PolRHS<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;

    double time_value = 0;
    unsigned int n = deg;
    for (std::list<double>::const_iterator coeff=coeffs.begin(); coeff != coeffs.end(); ++coeff)
    {
    	time_value *= current_time;
    	time_value -= n*(*coeff);
    	n--;
    }

	return time_value;
}


template<int dim>
ADIPolRHS<dim>::ADIPolRHS(double current_time)
	:
	Function<dim>(2*dim, current_time), // Only for dim=3
	current_time(current_time),
	rhs_E(current_time)
{}

template <int dim>
void ADIPolRHS<dim>::vector_value(const Point<dim>	&p,
								Vector<double>	&values) const
{
    Assert (values.size() == 2*dim, ExcDimensionMismatch (values.size(), 2*dim));
    Assert (dim >= 2, ExcNotImplemented());
	
    Vector<double> tmp_values(dim);
    rhs_E.vector_value(p, tmp_values);
    for (unsigned int ii=0; ii<dim; ++ii)
	{
    	values[ii]=0;
    	values[ii+dim]=tmp_values[ii];
	}
}

template <int dim>
void ADIPolRHS<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
									std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		ADIPolRHS<dim>::vector_value (points[p], value_list[p]);
}

// template <int dim>
// void ADIPolRHS<dim>::space_vector_value(const Point<dim>	&p,
// 									Vector<double>	&values) const
// {
// 	Assert (dim > 2, ExcNotImplemented());
// 	Assert (values.size() == 2*dim, ExcDimensionMismatch (values.size(), 2*dim));
// 
//     Vector<double> tmp_values(dim);
//     rhs_E.space_vector_value(p, tmp_values);
//     for (unsigned int ii=0; ii<dim; ++ii)
// 	{
//     	values[ii]=0;
//     	values[ii+dim]=tmp_values[ii];
// 	}
//     
// }
// 
// template <int dim>
// void ADIPolRHS<dim>::space_vector_value_list(const std::vector< Point<dim> >	&points,
// 									std::vector<Vector<double>>			&value_list) const
// {
// 	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));
// 
// 	const unsigned int n_points = points.size();
// 
// 	for (unsigned int p=0; p<n_points; ++p)
// 		ADIPolRHS<dim>::space_vector_value (points[p], value_list[p]);
// }

template <int dim>
double ADIPolRHS<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;

	return rhs_E.get_time_value(current_time);
}


template class PolRHS<3>;
template class ADIPolRHS<3>;

} /* namespace Data */
} /* namespace MaxwellProblem*/
