/*
 * PolRHS.h
 *
 *  Created on: Mar 26, 2017
 *      Author: leibold
 */

#ifndef POLRHS_H_
#define POLRHS_H_


#include <deal.II/lac/vector.h>

#include <list>
#include <boost/lexical_cast.hpp>

#include "FunctionInterface.h"
#include "../global.h"
#include "GradientField.h"

using namespace dealii;

namespace MaxwellProblem {
namespace Data{

/**
 * \brief Right-hand-side of the polynomial solution of the Maxwells equations
 *
 * It is given by
 * \f[
 * J(x,t) = p'(t)\nabla\Phi(x).
 * \f]
 *
 * See the documentation of PolSolution_E for further information.
 */
template <int dim>
class PolRHS : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 *
	 * \param current_time: time at which the function should is evaluated
	 */
	PolRHS 					(double current_time = 0.0);
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the polynomial \f$ p' \f$ at \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;
	/**
	 * Degree of the polynomial \f$ p \f$ (NOT \f$ p' \f$!)
	 */
	unsigned int deg;
	/**
	 * coefficients of the polynomial \f$ p \f$
	 */
	std::list<double> coeffs;
	/**
	 * gradient field that is used
	 */
	GradientField<dim> gradient_field;
};


template <int dim>
class ADIPolRHS : public Function<dim>
{
public:
	/**
	 * \brief Constructor.
	 *
	 * \param current_time: time at which the function should is evaluated
	 */
	ADIPolRHS 					(double current_time = 0.0);
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;
// 	/**
// 	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at a single point.
// 	 *
// 	 * 	This routine takes a point in the dim-dimensional space and
// 	 * 	writes the corresponding function-values at that point into
// 	 * 	the values-parameter.
// 	 */
// 	void space_vector_value 		(const Point<dim>				&p,
// 	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
// 	/**
// 	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at multiple points.
// 	 *
// 	 * 	This routine takes a point in the dim-dimensional space and
// 	 * 	writes the corresponding function-values at that point into
// 	 * 	the values-parameter.
// 	 */
// 	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
// 									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the polynomial \f$ p' \f$ at \p current_time.
	 */
	double get_time_value			(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;
	
	PolRHS<dim> rhs_E;
};

} /* namespace Data */
} /* namespace MaxwellProblem */

#endif /* POLRHS_H_ */
