/*
 * PolSolution.cpp
 *
 *  Created on: Mar 23, 2017
 *      Author: leibold
 */

#include "PolSolution.h"

namespace MaxwellProblem {
namespace Data{

template<int dim>
PolSolution_H<dim>::PolSolution_H(double current_time)
:
FunctionInterface<dim>(current_time), // Only for dim=3
current_time(current_time)
{}

template <int dim>
void PolSolution_H<dim>::vector_value(const Point<dim>&	/*p*/,
									  Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());

    values = 0.;
}

template <int dim>
void PolSolution_H<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		PolSolution_H<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
void PolSolution_H<dim>::space_vector_value(const Point<dim>&	/*p*/,
											Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

	values = 0.;
}

template <int dim>
void PolSolution_H<dim>::space_vector_value_list(const std::vector<Point<dim>>	&points,
											std::vector<Vector<double>>			&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		PolSolution_H<dim>::space_vector_value (points[p], value_list[p]);
}


template <int dim>
double PolSolution_H<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;

	return 0.;
}


template<int dim>
PolSolution_E<dim>::PolSolution_E(double current_time)
:
FunctionInterface<dim>(current_time), // Only for dim=3
current_time(current_time),
gradient_field()
{
    std::stringstream ss_c(MaxwellProblem::global::config.get<std::string>("Data.Coefficients"));
    std::string c;
    while (std::getline(ss_c, c, ','))
    {
    	c.erase(remove(c.begin(),c.end(),' '),c.end());
    	coeffs.push_back(boost::lexical_cast<double>(c));
    }
}

template <int dim>
void PolSolution_E<dim>::vector_value(const Point<dim>	&p,
									  Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());

    double value = 0;

	// Horner scheme
    for (std::list<double>::const_iterator coeff=coeffs.begin(); coeff != coeffs.end(); ++coeff)
    {
    	value *= current_time;
    	value += *coeff;
    }
    gradient_field.vector_value(p,values);
    values *= value;
}

template <int dim>
void PolSolution_E<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
										   std::vector< Vector<double> >	&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		PolSolution_E<dim>::vector_value (points[p], value_list[p]);
}

template <int dim>
void PolSolution_E<dim>::space_vector_value(const Point<dim>	&p,
											Vector<double>		&values) const
{
	Assert (dim > 2, ExcNotImplemented());
	Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));

    gradient_field.vector_value(p,values);
}

template <int dim>
void PolSolution_E<dim>::space_vector_value_list(const std::vector<Point<dim>>	&points,
											std::vector<Vector<double>>			&value_list) const
{
	Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));

	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		PolSolution_E<dim>::space_vector_value (points[p], value_list[p]);
}

template <int dim>
double PolSolution_E<dim>::get_time_value(double current_time)
{
	this->current_time = current_time;

    double time_value = 0.;

	//evaluation of polynomial in time
    for (std::list<double>::const_iterator coeff=coeffs.begin(); coeff != coeffs.end(); ++coeff)
    {
    	time_value *= current_time;
    	time_value += *coeff;
    }

	return time_value;
}


template class PolSolution_H<3>;
template class PolSolution_E<3>;

} /* namespace Data */
} /* namespace MaxwellProblem*/
