/*
 * PolSolution.h
 *
 *  Created on: Mar 23, 2017
 *      Author: leibold
 */

#ifndef POLSOLUTION_H_
#define POLSOLUTION_H_


#include <deal.II/lac/vector.h>

#include <list>
#include <boost/lexical_cast.hpp>

#include "FunctionInterface.h"

#include "../global.h"
#include "GradientField.h"

using namespace dealii;

namespace MaxwellProblem {
namespace Data{


/**
 * \brief Polynomial solution of the Maxwells equations.
 *
 * See documentation of PolSolution_E for further information.
 *
 * \note Because of the form of this solution, the H-component is always 0!
 */
template <int dim>
class PolSolution_H : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 *
	 * \param current_time: time at which the function should is evaluated
	 */
	PolSolution_H 			(double current_time = 0.0);
	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;

	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;

	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the polynomial \f$ p \f$ at \p current_time.
	 */
	double get_time_value	(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;
};

/**
 * \brief Polynomial solution of the Maxwells equations.
 *
 * \note Only for the domain  \f$ \Omega = [0,2]\times[0,1]\times[0,1] \f$.
 *
 * <b>Form of the solution \f$ u \f$: </b>
 *
 * Let \f$ \Phi:\mathbb{R}^3 \rightarrow \mathbb{R}\f$ be a \f$ C^2- \f$function. Then \f$ \nabla\Phi \f$ is divergence free.
 * We assume \f$ n \times \nabla \Phi = 0\f$ on \f$ \partial \Omega \f$, i.e. \f$ \nabla\Phi \f$ satisfies the boundary conditions of the Maxwells equations.
 *
 * Furthermore let \f$ p(t)=\sum_{i=0}^p a_i t^i \f$ be a polynomial.
 *
 * The polynomial solution is then defined via
 * \f[
 *  u(x,t) = \begin{pmatrix} 0 \\ \nabla \Phi (x) \end{pmatrix}p(t).
 * \f]
 * It has its name, because it is polynomial in time and solves the maxwells equations with the current
 * \f[
 * J(x,t) = p'(t)\nabla\Phi(x).
 * \f]
 *
 * The coefficients of the polynomial can be set in the config file of this project, the gradient field is defined in GradientField.
 */
template <int dim>
class PolSolution_E : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 *
	 * \param current_time: time at which the function should is evaluated
	 */
	PolSolution_E 			(double current_time = 0.0);

	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;

	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list	(const std::vector<Point<dim> >	&points,
							 std::vector<Vector<double> >	&value_list) const;

	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;

	/**
	 * \brief Evaluates the space-function \f$ \nabla \Phi \f$ at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;

	/**
	 * \brief Evaluates the polynomial \f$ p \f$ at \p current_time.
	 */
	double get_time_value	(double current_time);

private:
	/**
	 * variable of time
	 */
	double current_time;
	/**
	 * coefficients of \f$ p \f$
	 */
	std::list<double> coeffs;
	/**
	 * gradient-field which is used
	 */
	GradientField<dim> gradient_field;
};

} /* namespace Data */
} /* namespace MaxwellProblem */

#endif /* POLSOLUTION_H_ */
