/*
 * RHS.h
 *
 *  Created on: Feb 2, 2016
 *      Author: koehler, carle, leibold
 */

#ifndef RHS_H_
#define RHS_H_


#include <deal.II/lac/vector.h>

#include "FunctionInterface.h"

#include "PolRHS.h"

using namespace dealii;

namespace MaxwellProblem {
namespace Data{

/**
 * \brief This class implements the rhs-function of the Maxwell problem.
 *
 * This class is only an interface:
 * If an instance of this class is created, the constructor reads from the config file of
 * the project what right hand side should be used and generates an instance of this type.
 * Every time the right hand side is evaluated, this class then simply calls the corresponding
 * method of the right hand side that is used.
 *
 * Some right hand sides can be split in two parts that depends only on one variable (space variable resp. time variable).
 * Those two parts can be evaluated separately, which is used to make assembling of the rhs in time-integration more efficient.
 */

template <int dim>
class RHS : public FunctionInterface<dim>
{
public:
	/**
	 * \brief Constructor.
	 *
	 * \param current_time: time at which the function should is evaluated
	 */
	RHS(double time = 0.0);

	/**
	 * \brief Destructor.
	 */
	~RHS() {delete rhs_func; delete rhs_func_ts_sep;}

	/**
	 *  \brief Evaluates the function at a single point.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void vector_value(
			const Point<dim>				&p,
			Vector<double>					&values) const;

	/**
	 * \brief Evaluates the function at multiple points.
	 *
	 * This routine takes a list of points in the dim-dimensional space and
	 * writes the corresponding function-values at that points into
	 * the value_list-parameter.
	 */
	void vector_value_list(
			const std::vector<Point<dim>>	&points,
			std::vector<Vector<double>>		&value_list) const;

	/**
	 * \brief Evaluates the space-function at a single point.
	 *
	 *  This is the factor of the solution that depends only on the space variable.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value 		(const Point<dim>				&p,
	 	 	 	 	 	 	 	 	 Vector<double>					&values) const;
	/**
	 * \brief Evaluates the space-function at multiple points.
	 *
	 * 	This routine takes a point in the dim-dimensional space and
	 * 	writes the corresponding function-values at that point into
	 * 	the values-parameter.
	 */
	void space_vector_value_list	(const std::vector<Point<dim>>	&points,
									 std::vector<Vector<double>>	&value_list) const;
	/**
	 * \brief Evaluates the time function \p current_time.
	 */
	double get_time_value			(double current_time);


private:
	/**
	 * concrete right hand side function if it is time-space separated
	 */
	FunctionInterface<dim>* rhs_func_ts_sep;
	/**
	 * concrete right hand side function if it is not time-space separated
	 */
	Function<dim>*			rhs_func;
	/**
	 * current time
	 */
	double time;
};

} /* namespace Data */
} /* namespace MaxwellProblem */

#endif /* RHS_H_ */
