/*
 * ADIBlockTridiag.cpp
 *
 *  Created on: Feb 16, 2016
 *      Author: koehler
 */

#include "ADIBlockTridiag.h"

namespace MaxwellProblem {

template <int dim>
ADIBlockTridiag<dim>::ADIBlockTridiag ()
:
dofs_per_element(0),
n_elements(0),
dim_n(0),
n_threads(MultithreadInfo::n_cores())
// n_threads(1)
{}

template <int dim>
ADIBlockTridiag<dim>::ADIBlockTridiag (const ADIBlockTridiag<dim> &mat)
:
dofs_per_element(mat.dofs_per_element),
n_elements(mat.n_elements),
dim_n(mat.dim_n),
diagonal(mat.diagonal),
subdiagonal(mat.subdiagonal),
superdiagonal(mat.superdiagonal),
n_threads(MultithreadInfo::n_cores())
// n_threads(1)
{}

template <int dim>
ADIBlockTridiag<dim> & ADIBlockTridiag<dim>::operator*= (
		const double coeff
		)
{
	for (unsigned int ii = 0; ii < n_elements-1; ++ii)
	{
		diagonal[ii] *= coeff;
		subdiagonal[ii] *= coeff;
		superdiagonal[ii] *= coeff;
	}
	diagonal[n_elements-1] *= coeff;

	return *this;
}


template <int dim>
void ADIBlockTridiag<dim>::initialize (
		const ADIGeometryInfo< dim >	&geometry_info,
		const unsigned int				n_el
		)
{
	dofs_per_element = 2*geometry_info.get_comp_dofs_per_cell();
	n_elements=n_el;

	diagonal.resize(n_elements, FullMatrix<double>(dofs_per_element));
	subdiagonal.resize(n_elements-1, FullMatrix<double>(dofs_per_element));
	superdiagonal.resize(n_elements-1, FullMatrix<double>(dofs_per_element));

	dim_n = n_elements*dofs_per_element;
	
// 	std::cout << std::endl << "Blocktridiag initialized with " << n_elements << " elements." << std::endl << std::endl;
// 	if (n_elements <= 1)
// 		std::cout << "Problem!!" << std::endl << std::endl; 
}


template <int dim>
void ADIBlockTridiag<dim>::copy_from (
		const ADIBlockTridiag<dim>	&mat
		)
{
	dofs_per_element = mat.dofs_per_element;
	n_elements = mat.n_elements;
	dim_n = mat.dim_n;

	diagonal = mat.diagonal;
	subdiagonal = mat.subdiagonal;
	superdiagonal = mat.superdiagonal;
}


template <int dim>
void ADIBlockTridiag<dim>::add (
		const std::vector< types::global_dof_index >	&dofs,
		const std::vector< types::global_dof_index >	&neighbor_dofs,
		const FullMatrix< double >						&mat
		)
{
// 	add(dofs[0], neighbor_dofs[0], mat);
	add(dofs.at(0), neighbor_dofs.at(0), mat);
}


//template <int dim>
//void ADIBlockTridiag<dim>::add (
//		const std::vector< types::global_dof_index >	&dofs,
//		const FullMatrix< double >						&mat
//		)
//{
//	add(dofs[0], dofs[0], mat);
//}


template <int dim>
void ADIBlockTridiag<dim>::add (
		const std::vector< types::global_dof_index >	&dofs,
		const FullMatrix< double >						&mat
		)
{
	add(dofs.at(0), mat);
// 	add(dofs[0], mat);
}


template <int dim>
void ADIBlockTridiag<dim>::add (
		const types::global_dof_index					dof_offset,
		const FullMatrix< double >						&mat
		)
{
	types::global_dof_index matrix = dof_offset/dofs_per_element;
	types::global_dof_index offset=dof_offset%dofs_per_element;

// 	diagonal[matrix].add(mat, 1, offset, offset);
	diagonal.at(matrix).add(mat, 1, offset, offset);
}


template <int dim>
void ADIBlockTridiag<dim>::add (
		const types::global_dof_index	dof_offset_i,
		const types::global_dof_index	dof_offset_j,
		const FullMatrix< double >		&mat
		)
{
	types::global_dof_index	matrix_i=dof_offset_i/dofs_per_element,
							matrix_j=dof_offset_j/dofs_per_element;

	Assert (((matrix_i>matrix_j)?(matrix_i-matrix_j):(matrix_j-matrix_i))<2, ExcInternalError());

	//original without check:
	if (matrix_i==matrix_j)
		diagonal.at(matrix_i).add(mat, 1, dof_offset_i%dofs_per_element, dof_offset_j%dofs_per_element);
	else if (matrix_i>matrix_j)
		subdiagonal.at(matrix_i-1).add(mat, 1, dof_offset_i%dofs_per_element, dof_offset_j%dofs_per_element);
	else
		superdiagonal.at(matrix_i).add(mat, 1, dof_offset_i%dofs_per_element, dof_offset_j%dofs_per_element);
}


//template <int dim>
//void ADIBlockTridiag<dim>::add (
//		const std::vector< types::global_dof_index >	&dofs,
//		const FullMatrix< double >						&mat
//		)
//{
//	//**********************************************************************
//	// This is the same routine as the one with another parameter,
//	// neighbor_dofs, but for quadratic FullMatrices.
//	//**********************************************************************
//
//	Assert (mat.n()==dofs.size() && mat.m()==dofs.size(), ExcInternalError());
//
//	unsigned int component = dof_to_component(dofs[0]);
//
//	unsigned int matrix_index = dof_to_matrix(dofs[0]);
//
//	std::vector< types::global_dof_index > new_dofs(dofs);
//
//	for (unsigned int i=0; i<new_dofs.size(); ++i)
//		new_dofs[i] = dofs[i];
//
//	dofs_to_matrix_dofs(new_dofs);
//
//	matrices[component][matrix_index].add(new_dofs, mat);
//}


template <int dim>
void ADIBlockTridiag<dim>::add (
		types::global_dof_index	dof_i,
		types::global_dof_index	dof_j,
		const double			val
		)
{
	types::global_dof_index	matrix_i=dof_i/dofs_per_element,
							matrix_j=dof_j/dofs_per_element;

	if(((matrix_i>matrix_j)?(matrix_i-matrix_j):(matrix_j-matrix_i))>=2)
            std::cout << "Problem!" << std::endl << std::endl;
	
	
	// original without check:
	if (matrix_i==matrix_j)
		diagonal.at(matrix_i).add(dof_i%dofs_per_element, dof_j%dofs_per_element, val);
	else if (matrix_i>matrix_j)
		subdiagonal.at(matrix_i-1).add(dof_i%dofs_per_element, dof_j%dofs_per_element, val);
	else
		superdiagonal.at(matrix_i).add(dof_i%dofs_per_element, dof_j%dofs_per_element, val);
}


template <int dim>
void ADIBlockTridiag<dim>::add (
		const double					coeff,
		const ADIBlockTridiag< dim >	&mat
		)
{
	Assert(n_elements==mat.n_elements && dofs_per_element==mat.dofs_per_element, ExcInternalError());

	for (unsigned int ii = 0; ii < n_elements-1; ++ii)
	{
		diagonal.at(ii).add(coeff,mat.diagonal.at(ii));
		subdiagonal.at(ii).add(coeff,mat.subdiagonal.at(ii));
		superdiagonal.at(ii).add(coeff,mat.superdiagonal.at(ii));
	}
	diagonal.at(n_elements-1).add(coeff,mat.diagonal.at(n_elements-1));

// 	for (unsigned int ii = 0; ii < n_elements-1; ++ii)
// 	{
// 		diagonal[ii].add(coeff,mat.diagonal[ii]);
// 		subdiagonal[ii].add(coeff,mat.subdiagonal[ii]);
// 		superdiagonal[ii].add(coeff,mat.superdiagonal[ii]);
// 	}
// 	diagonal[n_elements-1].add(coeff,mat.diagonal[n_elements-1]);
}


template <int dim>
void ADIBlockTridiag<dim>::add (
		const ADIBlockTridiag< dim > &mat
		)
{
	//**********************************************************************
	// see the corresponding routine with variable coefficient.
	//**********************************************************************

	add(1,mat);
}


//template <int dim>
//void ADIBlockTridiag<dim>::mmult (
//		ADIBlockTridiag< dim > 			&C,
//		const ADIBlockTridiag< dim > 	&B
//		) const
//{
//	Assert(B.direction == direction, ExcInternalError());
//	Assert(B.dim_n == dim_n, ExcInternalError());
//	Assert(C.direction == direction, ExcInternalError());
//	Assert(C.dim_n == dim_n, ExcInternalError());
//
//	// loop over all matrices, i.e., subproblems
//	for (unsigned int component=0; component < dim; ++component)
//		for (unsigned int i=0; i < matrices[component].size(); ++i)
//			matrices[component][i].mmult(C.matrices[component][i], B.matrices[component][i], Vector< double >(), false);
//}
//
//
//template <int dim>
//void ADIBlockTridiag<dim>::Tmmult (
//		ADIBlockTridiag< dim >		&C,
//		const ADIBlockTridiag< dim > &B
//		) const
//{
//	Assert(B.direction == direction, ExcInternalError());
//	Assert(B.dim_n == dim_n, ExcInternalError());
//	Assert(C.direction == direction, ExcInternalError());
//	Assert(C.dim_n == dim_n, ExcInternalError());
//
//	// loop over all matrices, i.e., subproblems
//	for (unsigned int component=0; component < dim; ++component)
//		for (unsigned int i=0; i < matrices[component].size(); ++i)
//			matrices[component][i].Tmmult(C.matrices[component][i], B.matrices[component][i], Vector< double >(), false);
//}



template <int dim>
void ADIBlockTridiag<dim>::vmult (
		Vector< double > 					&vec_out,
		const Vector< double > 				&vec_in
		) const
{
	Vector<double>::const_iterator curr_it_in = vec_in.begin()+dofs_per_element;
	unsigned int curr_pos(0);
	Vector<double> 	tmp_vec_sol(dofs_per_element),
					tmp_vec1(Vector< double >(curr_it_in - dofs_per_element, curr_it_in)),
					tmp_vec2(Vector< double >(curr_it_in, curr_it_in + dofs_per_element));

	for (unsigned int ii = 0; ii < n_elements-2; ++ii)
	{
		diagonal[ii].vmult_add(tmp_vec_sol, tmp_vec1);
		superdiagonal[ii].vmult_add(tmp_vec_sol, tmp_vec2);

		for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
			vec_out[curr_pos + jj]=tmp_vec_sol[jj];

		subdiagonal[ii].vmult(tmp_vec_sol, tmp_vec1);

		curr_it_in += dofs_per_element;
		curr_pos += dofs_per_element;

		tmp_vec2.swap(tmp_vec1);
		tmp_vec2 = Vector< double >(curr_it_in, curr_it_in + dofs_per_element);
	}

	diagonal[n_elements-2].vmult_add(tmp_vec_sol, tmp_vec1);
	superdiagonal[n_elements-2].vmult_add(tmp_vec_sol, tmp_vec2);

	for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
		vec_out[curr_pos + jj]=tmp_vec_sol[jj];

	curr_pos += dofs_per_element;

	subdiagonal[n_elements-2].vmult(tmp_vec_sol, tmp_vec1);
	diagonal[n_elements-1].vmult_add(tmp_vec_sol, tmp_vec2);

	for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
		vec_out[curr_pos + jj]=tmp_vec_sol[jj];
}


//template <int dim>
//void ADIBlockTridiag<dim>::vmult_old (
//		Vector< double > 					&vec_out,
//		const Vector< double > 				&vec_in
//		) const
//{
//	Vector<double> tmp_vec_sol(dofs_per_element);
//	Vector<double>::const_iterator curr_it_in = vec_in.begin();
//	unsigned int curr_pos(0);
//
//	diagonal[0].vmult(tmp_vec_sol, Vector< double >(curr_it_in, curr_it_in + dofs_per_element));
//	superdiagonal[0].vmult_add(tmp_vec_sol, Vector< double >(curr_it_in+dofs_per_element, curr_it_in+2*dofs_per_element));
//
//	for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
//		vec_out[jj]=tmp_vec_sol[jj];
//
//	for (unsigned int ii = 1; ii < n_elements-1; ++ii)
//	{
//		curr_it_in += dofs_per_element;
//		curr_pos += dofs_per_element;
//
//		subdiagonal[ii-1].vmult(tmp_vec_sol, Vector< double >(curr_it_in - dofs_per_element,curr_it_in));
//		diagonal[ii].vmult_add(tmp_vec_sol, Vector< double >(curr_it_in, curr_it_in + dofs_per_element));
//		superdiagonal[ii].vmult_add(tmp_vec_sol, Vector< double >(curr_it_in + dofs_per_element, curr_it_in + 2*dofs_per_element));
//
//		for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
//			vec_out[curr_pos + jj]=tmp_vec_sol[jj];
//	}
//
//	curr_it_in += dofs_per_element;
//	curr_pos += dofs_per_element;
//
//	subdiagonal[n_elements-2].vmult(tmp_vec_sol, Vector< double >(curr_it_in - dofs_per_element, curr_it_in));
//	diagonal[n_elements-1].vmult_add(tmp_vec_sol, Vector< double >(curr_it_in, curr_it_in + dofs_per_element));
//
//	for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
//		vec_out[curr_pos + jj]=tmp_vec_sol[jj];
//}



//template <int dim>
//void ADIBlockTridiag<dim>::vmult (
//		Vector< double > 					&vec_out,
//		const Vector< double > 				&vec_in
//		) const
//{
//	if (MaxwellProblem::global::config.get<bool>("Assembling.Threading"))
//		threaded_vmult(vec_out, vec_in);
//	else
//		sequential_vmult(vec_out, vec_in);
//}
//
//
//
//template <int dim>
//void ADIBlockTridiag<dim>::threaded_vmult (
//		Vector< double > 					&vec_out,
//		const Vector< double > 				&vec_in
//		) const
//{
//	Assert(vec_in.size() == vec_out.size(), ExcInternalError());
//
//	// TODO nur fuer Threading, sollte man evtl als abgeleitete Klasse oder so machen..
//	for (unsigned int task=0; task < n_elements; ++task)
//		tasks.push(task);
//	// TODO nur fuer Threading
//
//	auto thread_func = &ADIBlockTridiag<dim>::vmult_function;
//
//	std::vector< std::thread > threads;
//
//	for (unsigned int ii=0; ii<n_threads; ++ii)
//		threads.push_back(std::thread(thread_func, this, std::ref(vec_out), std::cref(vec_in)));
//
//	for (unsigned int ii=0; ii<n_threads; ++ii)
//		threads[ii].join();
//}
//
//
//template <int dim>
//void ADIBlockTridiag<dim>::vmult_function (
//		Vector< double > 					&vec_out,
//		const Vector< double > 				&vec_in
//		) const
//{
//	while (true)
//	{
//		mtx.lock();
//
//		if(tasks.empty())
//		{
//			mtx.unlock();
//			break;
//		}
//
//		unsigned int current_task = tasks.top();
//		tasks.pop();
//
//		mtx.unlock();
//
//		Vector<double> tmp_vec_sol(dofs_per_element);
//		Vector<double>::const_iterator curr_it_in = vec_in.begin();
//		unsigned int curr_pos(0);
//
//		if (current_task==0)
//		{
//			diagonal[0].vmult(tmp_vec_sol, Vector< double >(curr_it_in, curr_it_in + dofs_per_element));
//			superdiagonal[0].vmult_add(tmp_vec_sol, Vector< double >(curr_it_in+dofs_per_element, curr_it_in+2*dofs_per_element));
//
//			for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
//				vec_out[jj]=tmp_vec_sol[jj];
//		}
//		else if (current_task < n_elements-1)
//		{
//			curr_it_in += current_task*dofs_per_element;
//			curr_pos = current_task*dofs_per_element;
//
//			subdiagonal[current_task-1].vmult(tmp_vec_sol, Vector< double >(curr_it_in - dofs_per_element,curr_it_in));
//			diagonal[current_task].vmult_add(tmp_vec_sol, Vector< double >(curr_it_in, curr_it_in + dofs_per_element));
//			superdiagonal[current_task].vmult_add(tmp_vec_sol, Vector< double >(curr_it_in + dofs_per_element, curr_it_in + 2*dofs_per_element));
//
//			for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
//				vec_out[curr_pos + jj]=tmp_vec_sol[jj];
//		}
//		else
//		{
//			curr_it_in += current_task*dofs_per_element;
//			curr_pos = current_task*dofs_per_element;
//
//			subdiagonal[current_task-1].vmult(tmp_vec_sol, Vector< double >(curr_it_in - dofs_per_element, curr_it_in));
//			diagonal[current_task].vmult_add(tmp_vec_sol, Vector< double >(curr_it_in, curr_it_in + dofs_per_element));
//
//			for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
//				vec_out[curr_pos + jj]=tmp_vec_sol[jj];
//		}
//	}
//}


template <int dim>
void ADIBlockTridiag<dim>::invert ()
{
	FullMatrix< double > tmp(dofs_per_element);

	for (unsigned int ii = 0; ii < n_elements-1; ++ii)
	{
		diagonal[ii].gauss_jordan();

		diagonal[ii].mmult(tmp, superdiagonal[ii]);
		superdiagonal[ii]=tmp;

		subdiagonal[ii].mmult(tmp, superdiagonal[ii]);
		diagonal[ii+1].add(-1, tmp);
	}

	diagonal[n_elements-1].gauss_jordan();
}


template <int dim>
void ADIBlockTridiag<dim>::inverse_vmult (
		Vector< double > 							&vec_out,
 		const Vector< double > 						&vec_in
		) const
{
	Vector<double>::const_iterator curr_it = vec_in.begin() + dofs_per_element;
	unsigned int curr_pos(0);
	Vector<double>	tmp_vec_sol(dofs_per_element),
					tmp_vec(dim_n),
					tmp_vec1(Vector< double >(curr_it - dofs_per_element, curr_it)),
					tmp_vec2(Vector< double >(curr_it, curr_it + dofs_per_element));

	unsigned int ii = 0;

	// forward substitution:
	for (; ii < n_elements-2; ++ii)
	{
		diagonal[ii].vmult(tmp_vec_sol, tmp_vec1);

		for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
			tmp_vec[curr_pos + jj] = tmp_vec_sol[jj];

		subdiagonal[ii].vmult(tmp_vec1, tmp_vec_sol);
		tmp_vec2 -= tmp_vec1;

		curr_it += dofs_per_element;
		curr_pos += dofs_per_element;

		tmp_vec2.swap(tmp_vec1);
		tmp_vec2 = Vector< double >(curr_it, curr_it + dofs_per_element);
	}

	diagonal[ii].vmult(tmp_vec_sol, tmp_vec1);

	for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
		tmp_vec[curr_pos + jj]=tmp_vec_sol[jj];

//	tmp_vec_sol.print(std::cout,3,true,false);
//	std::cout << std::endl << std::endl;

	subdiagonal[ii].vmult(tmp_vec1, tmp_vec_sol);
	tmp_vec2 -= tmp_vec1;

//	diagonal[ii+1].print(std::cout);
//	std::cout << std::endl << std::endl;

	diagonal[ii+1].vmult(tmp_vec_sol, tmp_vec2);
	curr_pos += dofs_per_element;

	for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
		vec_out[curr_pos + jj] = tmp_vec_sol[jj];

	// backward substitution
	curr_it = tmp_vec.end() - dofs_per_element;
	curr_pos -= dofs_per_element;

	for (; ii+1 != 0; --ii)
	{
		superdiagonal[ii].vmult(tmp_vec1, tmp_vec_sol);

		tmp_vec_sol = Vector< double >(curr_it - dofs_per_element, curr_it);

		tmp_vec_sol -= tmp_vec1;

		for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
			vec_out[curr_pos + jj] = tmp_vec_sol[jj];

		curr_it -= dofs_per_element;
		curr_pos -= dofs_per_element;
	}
//	superdiagonal[0].vmult(tmp_vec1, tmp_vec_sol);

//	tmp_vec_sol = Vector< double >(curr_it - dofs_per_element, curr_it);

//	tmp_vec_sol -= tmp_vec1;

//	for (unsigned int jj = 0; jj < dofs_per_element; ++jj)
//		vec_out[curr_pos + jj] = tmp_vec_sol[jj];
}


template <int dim>
void ADIBlockTridiag<dim>::print_norms () const
{
	for (unsigned int ii=0; ii<n_elements; ++ii)
	{
		diagonal[ii].frobenius_norm();
	}
}


template <int dim>
void ADIBlockTridiag<dim>::save_matrix (
		std::ofstream &out
		) const
{
	out.precision(std::numeric_limits< double >::max_digits10);
	out << std::scientific;
	
	out << dofs_per_element << "\t" << n_elements << "\t" << dim_n << std::endl;
	
	
	auto diag_it = diagonal.begin(), end_diag_it = diagonal.end(),
		 subd_it = subdiagonal.begin(), end_subd_it = subdiagonal.end(),
		 supd_it = superdiagonal.begin(), end_supd_it = superdiagonal.end();
		 
	for (; diag_it != end_diag_it; ++diag_it) 
		for ( unsigned int row = 0; row < (*diag_it).m(); ++row )
		{
			for ( unsigned int col = 0; col < (*diag_it).n(); ++col )
				out << (*diag_it)[row][col] << "\t";
			out << std::endl;
		}
	out << std::endl;
	
	for (; subd_it != end_subd_it; ++subd_it) 
		for ( unsigned int row = 0; row < (*subd_it).m(); ++row )
		{
			for ( unsigned int col = 0; col < (*subd_it).n(); ++col )
				out << (*subd_it)[row][col] << "\t";
			out << std::endl;
		}
	out << std::endl;
	
	for (; supd_it != end_supd_it; ++supd_it) 
		for ( unsigned int row = 0; row < (*supd_it).m(); ++row )
		{
			for ( unsigned int col = 0; col < (*supd_it).n(); ++col )
				out << (*supd_it)[row][col] << "\t";
			out << std::endl;
		}
	out << std::endl;
}


template <int dim>
void ADIBlockTridiag<dim>::load_matrix (
		std::ifstream &in
		)
{
	char buff[256];
	char delim = '\t';
	
	in.getline(buff, 256, delim);
	dofs_per_element = atoi(buff);
	
	in.getline(buff, 256, delim);
	n_elements = atoi(buff);
	
	in.getline(buff, 256, '\n');
	dim_n = atoi(buff);
	
	diagonal.resize(n_elements, FullMatrix<double>(dofs_per_element));
	subdiagonal.resize(n_elements-1, FullMatrix<double>(dofs_per_element));
	superdiagonal.resize(n_elements-1, FullMatrix<double>(dofs_per_element));
	
	
	auto diag_it = diagonal.begin(), end_diag_it = diagonal.end(),
		 subd_it = subdiagonal.begin(), end_subd_it = subdiagonal.end(),
		 supd_it = superdiagonal.begin(), end_supd_it = superdiagonal.end();
		 
	for (; diag_it != end_diag_it; ++diag_it) 
		for ( unsigned int row = 0; row < dofs_per_element; ++row )
		{
			for ( unsigned int col = 0; col < dofs_per_element; ++col )
			{
				in.getline(buff, 256, delim);
				(*diag_it)[row][col] = atof(buff);
			}
			while(!strcmp(buff, ""))
				in.getline(buff, 256);
		}
		 
	for (; subd_it != end_subd_it; ++subd_it) 
		for ( unsigned int row = 0; row < dofs_per_element; ++row )
		{
			for ( unsigned int col = 0; col < dofs_per_element; ++col )
			{
				in.getline(buff, 256, delim);
				(*subd_it)[row][col] = atof(buff);
			}
			while(!strcmp(buff, ""))
				in.getline(buff, 256);
		}
		 
	for (; supd_it != end_supd_it; ++supd_it) 
		for ( unsigned int row = 0; row < dofs_per_element; ++row )
		{
			for ( unsigned int col = 0; col < dofs_per_element; ++col )
			{
				in.getline(buff, 256, delim);
				(*supd_it)[row][col] = atof(buff);
			}
			while(!strcmp(buff, ""))
				in.getline(buff, 256);
		}
}


//template <int dim>
//void ADIBlockTridiag<dim>::output_matrix (
//		std::stringstream &filename
//		) const
//{
//	//**********************************************************************
//	// This routine puts out the object in a format readable by visit or
//	// paraview, but without values, i.e. only as a spy pattern.
//	// Nevertheless, in the file itself, the values are written down.
//	//**********************************************************************
//
//	filename << ".vtk";
//
//	std::ofstream out ((filename.str()).c_str());
//
//	unsigned int offseti=0, offsetj=0;
//
//	for (unsigned int component=0; component < matrices.size(); ++component)
//		for (unsigned int ii=0; ii < matrices[component].size(); ++ii)
//		{
////			for (unsigned int i=0; i<matrix_sizes[component][ii]; ++i)
////				for (unsigned int j=0; j<matrix_sizes[component][ii]; ++j)
////					if (fabs(matrices[component][ii].el(i,j)) > 1e-17)
////						out << i+offseti << " -" << j+offsetj << "\t" << matrices[component][ii].el(i,j) << std::endl;
//
//			for (unsigned int row=0; row<matrix_sizes[component][ii]; ++row)
//			{
//				SparseMatrix<double>::const_iterator it = matrices[component][ii].begin(row);
//				SparseMatrix<double>::const_iterator end = matrices[component][ii].end(row);
//				for (; it!=end; ++it)
//					if (fabs((*it).value()) > 10e-17)
//						out << row+offseti << " -" << (*it).column()+offsetj << "\t" << (*it).value() << std::endl;
//			}
//
//			offseti+=matrix_sizes[component][ii];
//			offsetj+=matrix_sizes[component][ii];
//		}
//}


//template <int dim>
//void ADIBlockTridiag<dim>::output_matrix_spy (
//		std::stringstream &filename
//		) const
//{
//	//**********************************************************************
//	// This routine puts out the object in a format readable by visit or
//	// paraview, but without values, i.e. only as a spy pattern.
//	//**********************************************************************
//
//	filename << "_spy.vtk";
//
//	std::ofstream out ((filename.str()).c_str());
//
//	unsigned int offseti=0, offsetj=0;
//
//	for (unsigned int component=0; component < matrices.size(); ++component)
//		for (unsigned int ii=0; ii < matrices[component].size(); ++ii)
//		{
////			for (unsigned int i=0; i<matrix_sizes[component][ii]; ++i)
////				for (unsigned int j=0; j<matrix_sizes[component][ii]; ++j)
////					if (fabs(matrices[component][ii].el(i,j)) > 1e-17)
////						out << i+offseti << " -" << j+offsetj << std::endl;
//
//			for (unsigned int row=0; row<matrix_sizes[component][ii]; ++row)
//			{
//				SparseMatrix<double>::const_iterator it = matrices[component][ii].begin(row);
//				SparseMatrix<double>::const_iterator end = matrices[component][ii].end(row);
//				for (; it!=end; ++it)
//					if (fabs((*it).value()) > 10e-17)
//						out << row+offseti << " -" << (*it).column()+offsetj << std::endl;
//			}
//
//			offseti+=matrix_sizes[component][ii];
//			offsetj+=matrix_sizes[component][ii];
//		}
//}


// template <int dim>
// void ADIBlockTridiag<dim>::output_matrix_matlab (
// 		std::stringstream &filename
// 		) const
// {
// 	filename << "_matlab.vtk";
// 
// 	std::ofstream out ((filename.str()).c_str());
// 
// 	unsigned int offseti=0, offsetj=0;
// 
// 	for (unsigned int component=0; component < matrices.size(); ++component)
// 		for (unsigned int ii=0; ii < matrices[component].size(); ++ii)
// 		{
// //			for (unsigned int i=0; i<matrix_sizes[component][ii]; ++i)
// //				for (unsigned int j=0; j<matrix_sizes[component][ii]; ++j)
// //					if (fabs(matrices[component][ii].el(i,j)) > 1e-17)
// //						out << i+offseti+1 << "\t" << j+offsetj+1 << "\t" << matrices[component][ii].el(i,j) << std::endl;
// 
// 			for (unsigned int row=0; row<matrix_sizes[component][ii]; ++row)
// 			{
// 				SparseMatrix<double>::const_iterator it = matrices[component][ii].begin(row);
// 				SparseMatrix<double>::const_iterator end = matrices[component][ii].end(row);
// 				for (; it!=end; ++it)
// 					if (fabs((*it).value()) > 10e-17)
// 						out << row+offseti+1 << "\t" << (*it).column()+offsetj+1 << "\t" << (*it).value() << std::endl;
// 			}
// 
// 			offseti+=matrix_sizes[component][ii];
// 			offsetj+=matrix_sizes[component][ii];
// 		}
// }


//template <int dim>
//bool ADIBlockTridiag<dim>::operator== (
//		const ADIBlockTridiag<dim>	&mat
//		)
//{
//	//**********************************************************************
//	// Comparison of ADIMatrix-Objects, very costly, should only be called
//	// for testing purposes. This routine simply compares all the member
//	// variables.
//	//**********************************************************************
//
//	if (direction != mat.direction
//			|| comp_dofs_per_cell != mat.comp_dofs_per_cell
//			|| reordering != mat.reordering
//			|| matrices.size() != mat.matrices.size()
//			|| dof2matrix.size() != mat.dof2matrix.size())
//		return false;
//
//	for (unsigned int i=0; i < dof2matrix.size(); ++i)
//		if (dof2matrix[i] != mat.dof2matrix[i] || dof2matrixdof[i] != mat.dof2matrixdof[i])
//				return false;
//
//	for (unsigned int component=0; component < matrices.size(); ++component)
//		for (unsigned int i=0; i < matrices[component].size(); ++i)
//			for (SparseMatrix< double >::const_iterator entry1 = matrices[component][i].begin(), entry2 = mat.matrices[component][i].begin(); entry1 < matrices[component][i].end(); ++entry1, ++entry2)
//				if (!(*entry1 == *entry2))
//					return false;
//
//	return true;
//}


template <int dim>
unsigned int ADIBlockTridiag<dim>::n () const
{
	return dim_n;
}


template <int dim>
unsigned int ADIBlockTridiag<dim>::m () const
{
	return dim_n;
}

template class ADIBlockTridiag<3>;

} /* namespace MaxwellProblem */
