#include "ADIMassMatrix.h"

namespace MaxwellProblem {

template <int dim>
ADIMassMatrix<dim>::ADIMassMatrix ()
:
direction(405),
dofs_per_cell(0),
n_matrices(0),
reordering(false),
n_threads(MultithreadInfo::n_threads())
{}


template <int dim>
void ADIMassMatrix<dim>::initialize	(
		const ADIGeometryInfo<dim>			&geometry_info,
		unsigned int						dir,
		unsigned int						system_direction
		)
{
	/*
	 * Initialization routine for ADIMassMatrices. This sets up all the
	 * member variables by using information given by the geometry_info
	 * object and initializes the matrices vector and the FullMatrix objects
	 * forming this vector.
	 */

	/*
	 * set direction such that the adding-routine of the ADIMatrix objects know whether they should
	 * use transform or not
	 */
	direction = dir;

	if (system_direction==direction)
		reordering = false;
	else
		reordering = true;

	/*
	 * get basic geometric infos from the geometry_info-object
	 */
	n_matrices = geometry_info.get_n_elements()*geometry_info.get_n_components();

	dofs_per_cell = geometry_info.get_comp_dofs_per_cell();

	/*
	 * initialize matrices vector
	 */
	matrices.resize(n_matrices);

	/*
	 * initialize the FullMatrix objects corresponding to the element mass matrices.
	 */
	for (unsigned int i=0; i < n_matrices; ++i)
		matrices[i] = FullMatrix< double >(dofs_per_cell, dofs_per_cell);

	offset.resize(matrices.size());

	unsigned int temp = 0;

	for (unsigned int matrix=0; matrix < matrices.size(); ++matrix)
	{
		offset[matrix]=temp;
		temp += matrices[matrix].n();
	}
}


template <int dim>
void ADIMassMatrix<dim>::vmult (
		Vector< double > 					&vec_out,
		const Vector< double > 				&vec_in,
		const std::vector< unsigned int > 	&transform
		) const
{
	Assert(vec_in.size() == vec_out.size(), ExcInternalError());

	for (unsigned int matrix=0; matrix < matrices.size(); ++matrix)
			tasks.push(matrix);

	auto thread_func = &ADIMassMatrix<dim>::vmult_function;

	std::vector< std::thread > threads;
	threads.reserve(n_threads);

	for (unsigned int ii=0; ii<n_threads; ++ii)
		threads.push_back(std::thread(thread_func, this, std::ref(vec_out), std::cref(vec_in), std::cref(transform)));

	for (unsigned int ii=0; ii<n_threads; ++ii)
		threads[ii].join();
}


template <int dim>
void ADIMassMatrix<dim>::vmult_function (
		Vector< double > 					&vec_out,
		const Vector< double > 				&vec_in,
		const std::vector< unsigned int >	&transform
		) const
{
	while (true)
	{
		mtx.lock();

		if(tasks.empty())
		{
				mtx.unlock();
				break;
		}

		unsigned int current_task = tasks.top();
		tasks.pop();

		mtx.unlock();

		/*
		 * initialize temporary vectors corresponding to the vectors of the subproblems
		 * and the offset-variable keeping track of the dofs of the current subproblem.
		 */
		Vector< double > temp_in, temp_out, temp_transform;

		temp_in.reinit(matrices[current_task].n());
		temp_out.reinit(matrices[current_task].n());
		temp_transform.reinit(matrices[current_task].n());

		/*
		 * use the vmult-routine of the SparseMatrix class to do the actual vector
		 * multiplication on the subproblems. If necessary, reorder the dofs beforehand
		 * and retransform them afterwards.
		 * 
		 * Comment: In principle, there should never be a reason to use the transformation
		 * as the mass matrix can always be assembled w.r.t. the system direction (?)
		 */
		if (reordering)
		{
			for (unsigned int j=0; j< matrices[current_task].n(); ++j)
			{
				temp_in[j] = vec_in[transform_dofs(offset[current_task] + j, transform)];
				temp_transform[j] = transform_dofs(offset[current_task] + j, transform);
			}

			matrices[current_task].vmult(temp_out,temp_in);

			for (unsigned int j=0; j< matrices[current_task].n(); ++j)
				vec_out[temp_transform[j]] = temp_out[j];
		}
		else
		{
			for (unsigned int j=0; j< matrices[current_task].n(); ++j)
				temp_in[j] = vec_in[offset[current_task] + j];

			matrices[current_task].vmult(temp_out,temp_in);

			for (unsigned int j=0; j< matrices[current_task].n(); ++j)
				vec_out[offset[current_task] + j] = temp_out[j];
		}
	}
}


template <int dim>
void ADIMassMatrix<dim>::invert ()
{
	for (unsigned int matrix=0; matrix < matrices.size(); ++matrix)
		tasks.push(matrix);

	auto thread_func = &ADIMassMatrix<dim>::invert_function;

	std::vector< std::thread > threads;
	threads.reserve(n_threads);

	for (unsigned int ii=0; ii<n_threads; ++ii)
		threads.push_back(std::thread(thread_func, this));

	for (unsigned int ii=0; ii<n_threads; ++ii)
		threads[ii].join();
}


template <int dim>
void ADIMassMatrix<dim>::invert_function ()
{
	while (true)
	{
		mtx.lock();

		if(tasks.empty())
		{
				mtx.unlock();
				break;
		}

		unsigned int current_task = tasks.top();
		tasks.pop();

		mtx.unlock();

		matrices[current_task].gauss_jordan();
	}
}


template <int dim>
void ADIMassMatrix<dim>::output_matrix (
		std::stringstream &filename
		) const
{
	/*
	 * This routine puts out the object in a format readable by visit or
	 * paraview, but without values, i.e. only as a spy pattern.
	 * Nevertheless, in the file itself, the values are written down.
	 */

	filename << ".vtk";

	std::ofstream out ((filename.str()).c_str());

	unsigned int offseti=0, offsetj=0;

	for (unsigned int ii=0; ii < matrices.size(); ++ii)
	{
		for (unsigned int row=0; row<matrices[ii].n(); ++row)
		{
			FullMatrix<double>::const_iterator it = matrices[ii].begin(row);
			FullMatrix<double>::const_iterator end = matrices[ii].end(row);
			for (; it!=end; ++it)
				if (fabs((*it).value()) > 10e-17)
					out << row+offseti << " -" << (*it).column()+offsetj << "\t" << (*it).value() << std::endl;
		}

		offseti+=matrices[ii].n();
		offsetj+=matrices[ii].m();
	}
}


template <int dim>
void ADIMassMatrix<dim>::output_matrix_spy (
		std::stringstream &filename
		) const
{
	/*
	 * This routine puts out the object in a format readable by visit or
	 * paraview, but without values, i.e. only as a sparsity pattern.
	 */

	filename << "_spy.vtk";

	std::ofstream out ((filename.str()).c_str());

	unsigned int offseti=0, offsetj=0;

	std::string stri, strj;

	for (unsigned int ii=0; ii < matrices.size(); ++ii)
	{
		for (unsigned int row=0; row<matrices[ii].n(); ++row)
		{
			FullMatrix<double>::const_iterator it = matrices[ii].begin(row);
			FullMatrix<double>::const_iterator end = matrices[ii].end(row);
			for (; it!=end; ++it)
				if (fabs((*it).value()) > 10e-17)
					out << row+offseti << " -" << (*it).column()+offsetj << std::endl;
		}

		offseti+=matrices[ii].n();
		offsetj+=matrices[ii].m();
	}
}


template <int dim>
void ADIMassMatrix<dim>::output_matrix_matlab (
		std::stringstream &filename
		) const
{
	/*
	 * This routine puts out the object in a format readable by visit or
	 * paraview, but without values, i.e. only as a spy pattern.
	 * Nevertheless, in the file itself, the values are written down.
	 */

	filename << "_matlab.vtk";

	std::ofstream out ((filename.str()).c_str());

	unsigned int offseti=0, offsetj=0;

	for (unsigned int ii=0; ii < matrices.size(); ++ii)
	{
		for (unsigned int row=0; row<matrices[ii].n(); ++row)
		{
			FullMatrix<double>::const_iterator it = matrices[ii].begin(row);
			FullMatrix<double>::const_iterator end = matrices[ii].end(row);
			for (; it!=end; ++it)
				if (fabs((*it).value()) > 10e-17)
					out << row+offseti+1 << "\t" << (*it).column()+offsetj+1 << "\t" << (*it).value() << std::endl;
		}

		offseti+=matrices[ii].n();
		offsetj+=matrices[ii].m();
	}
}


template <int dim>
double& ADIMassMatrix<dim>::operator() (
		const types::global_dof_index	dof_i,
		const types::global_dof_index	dof_j
		)
{
	/*
	 * This routine provides read-write access on the entries of the mass
	 * matrix with global indices (dof_i, dof_j).
	 */

	unsigned int matrix_index = dof_i / dofs_per_cell;

	unsigned int new_dof_i = dof_i % dofs_per_cell;
	unsigned int new_dof_j = dof_j % dofs_per_cell;

// 	Assert((unsigned int)fabs(dof_i-dof_j) <= dofs_per_cell, ExcInternalError());

	return matrices[matrix_index](new_dof_i, new_dof_j);
}


template <int dim>
void ADIMassMatrix<dim>::copy_from	(
		const ADIMassMatrix<dim>			&mat
								 )
{
	direction = mat.direction;
	dofs_per_cell = mat.dofs_per_cell;
	n_matrices = mat.n_matrices;
	reordering = mat.reordering;

	matrices = mat.matrices;

	offset = mat.offset;
}


template <int dim>
void ADIMassMatrix<dim>::operator=	(
		const ADIMassMatrix<dim>			&mat
								 )
{
	direction = mat.direction;
	dofs_per_cell = mat.dofs_per_cell;
	n_matrices = mat.n_matrices;
	reordering = mat.reordering;

	matrices = mat.matrices;

	offset = mat.offset;
}


template <int dim>
unsigned int ADIMassMatrix<dim>::transform_dofs	(
		const unsigned int					dof,
		const std::vector< unsigned int >	&transform
		) const
{
	/*
	 * This routine transforms dofs from the system_direction to the
	 * direction of this object.
	 */

	return transform[dof];
}


template <int dim>
unsigned int ADIMassMatrix<dim>::get_direction () const
{
	/*
	 * This routine returns the direction of the object.
	 */

	return direction;
}


template <int dim>
const FullMatrix< double >& ADIMassMatrix<dim>::get_element_matrix (
		types::global_dof_index i
		) const
{
	/*
	 * This routine returns the FullMatrix object corresponding to the mass
	 * matrix of element i.
	 */

	return matrices[i];
}


template <int dim>
unsigned int ADIMassMatrix<dim>::get_n_element_matrices () const
{
	return n_matrices;
}


template class ADIMassMatrix<3>;

} /* namespace MaxwellProblem */
