#ifndef ADIMASSMATRIX_H_
#define ADIMASSMATRIX_H_

#include <deal.II/numerics/matrix_tools.h>
#include <deal.II/base/multithread_info.h>

#include <iostream>
#include <fstream>

#include <vector>
#include <thread>
#include <mutex>

#include "../reordering/ADIGeometryInfo.h"
#include "../global.h"

namespace MaxwellProblem {

using namespace dealii;

/*
 * This class implements a mass matrix structure tailored to the ADIMatrix
 * class used for the ADI-method.
 * Since for the ADI-method, we need two orderings of the cells,
 * transformation is necessary, if only one mass matrix object should be
 * in memory (and no copy of it with changed sequence of the cells).
 * Unfortunately, since deal.II only admits addition of matrices connected
 * to the very same sparsity pattern, this makes it necessary to either
 * store two mass matrices with changed sequence or use another data
 * structure like this class. One drawback is, unfortunately, loss of
 * efficiency, since the element mass matrices are stored in FullMatrices
 * and adding of FullMatrices into SparseMatrices is quite costly.
 *
 * The element mass matrices are, as already stated, stored as FullMatrices
 * in a vector of length n_elements. This, of course, introduces some book-
 * keeping, which is done by the routines of this class.
 *
 * This class is made such that it is compatible to the ADIMatrix class and
 * that only one mass matrix object has to be stored even though the system
 * is split into two directional parts.
 */

template <int dim>
class ADIMassMatrix
{
public:
								ADIMassMatrix				();
	void						initialize	 				(const ADIGeometryInfo<dim>						&geometry_info,
	 	 	 	 	 	 	 	 	 	 	 	 	 	 	 unsigned int 									dir,
	 	 	 	 	 	 	 	 	 	 	 	 	 	 	 unsigned int 									system_dir);
	void						output_matrix				(std::stringstream								&filename) const;
	void						output_matrix_spy			(std::stringstream								&filename) const;
	void						output_matrix_matlab		(std::stringstream								&filename) const;
	unsigned int		 		get_direction				() const;
	double&						operator()					(const types::global_dof_index					dof_i,
														 	 const types::global_dof_index					dof_j);
	const FullMatrix< double >& get_element_matrix			(types::global_dof_index 						i) const;
	unsigned int				get_n_element_matrices		() const;
	void						vmult 						(Vector< double > 								&vec_out,
												 	 		 const Vector< double > 						&vec_in,
												 	 		 const std::vector< unsigned int > 				&transform) const;
	void						invert 						();
	void						copy_from					(const ADIMassMatrix<dim>						&mat);
	void						operator=					(const ADIMassMatrix<dim>						&mat);

private:
	void						vmult_function 				(Vector< double > 								&vec_out,
												 	 		 const Vector< double > 						&vec_in,
												 	 		 const std::vector< unsigned int > 				&transform) const;
	void						invert_function 			();
	unsigned int				transform_dofs				(const unsigned int								dof,
														 	 const std::vector< unsigned int >				&transform) const;


	unsigned int 													direction, dofs_per_cell, n_matrices;

	bool															reordering;

	std::vector< FullMatrix< double > >								matrices;

	std::vector <unsigned int>										offset;

	const unsigned int												n_threads;

	mutable std::stack< unsigned int >								tasks;
	mutable std::mutex												mtx;
};

} /* namespace MaxwellProblem */

#endif /* ADIMASSMATRIX_H_ */
