#include "ADIMatrix.h"

namespace MaxwellProblem {

template <int dim>
ADIMatrix<dim>::ADIMatrix ()
:
direction(404),
comp_dofs_per_cell(0),
dim_n(0),
dim_m(0),
reordering(false),
n_threads(MultithreadInfo::n_cores())
// n_threads(1)
{}

template <int dim>
ADIMatrix<dim> & ADIMatrix<dim>::operator*= (
		const double coeff
		)
{
	for (unsigned int component=0; component < dim; ++component)
		for (unsigned int i=0; i < matrices[component].size(); ++i)
			this->matrices[component][i] *= coeff;
	return *this;
}


template <int dim>
void ADIMatrix<dim>::initialize (
		const ADIGeometryInfo< dim >	&geometry_info,
		const unsigned int				ADI_direction,
		const unsigned int				system_direction
		)
{
	//**********************************************************************
	// Initialization routine for ADIMatrices. This sets up all the member
	// variables by using information given by the geometry_info object and
	// reinits the matrices by their corresponding SparsityPattern in the
	// ADISparsityPattern object.
	//**********************************************************************

	// TODO die ganze Buchhaltungsgeschichte evtl effizienter gestalten,

	// get basic geometric infos from the geometry_info-object.
	unsigned int n_elements = geometry_info.get_n_elements();
	comp_dofs_per_cell = geometry_info.get_comp_dofs_per_cell();

	// set direction such that the adding- and vmult-Routines know whether they should use transform
	// or not.
	direction = ADI_direction;

	if (system_direction==direction)
		reordering = false;
	else
		reordering = true;

	// get the number of rows to know how many matrices have to be stored.
	std::vector< unsigned int > n_rows(dim);
	for (unsigned int i=0; i<dim; ++i)
		n_rows[i] = geometry_info.get_n_rows(i);

	// resize the matrices-vector and the index-transforms from global dof to matrix, resp. (matrix-)local dofs.
	matrices.resize(dim);
	matrix_sizes.resize(dim);
	n_dofs_per_component.resize(dim);
	dof2component.resize(n_elements*dim*2*comp_dofs_per_cell);
	dof2matrix.resize(n_elements*dim*2*comp_dofs_per_cell);
	dof2matrixdof.resize(n_elements*dim*2*comp_dofs_per_cell);

	// reinit the matrices. The bookkeeping stuff is done by the ADISparsityPattern object.
	for (unsigned int component=0; component < dim; ++component)
	{
		matrices[component].resize(n_rows[component]);
		matrix_sizes[component].resize(n_rows[component]);
		for (unsigned int i=0; i < n_rows[component]; ++i)
		{
			matrices[component][i].initialize(geometry_info, geometry_info.get_n_el_per_row(component, i));
			matrix_sizes[component][i] = matrices[component][i].n();
			n_dofs_per_component[component] += matrix_sizes[component][i];
		}
		dim_n += n_dofs_per_component[component];
		dim_m += n_dofs_per_component[component];
	}

	// compute the dof2matrix and dof2matrixdof vectors giving the transformation from global dof to
	// matrix or (matrix-)local dof.
	unsigned int jj = 0;
	for (unsigned int component = 0; component < dim; ++component)
	{
		for (unsigned int i=0; i < n_rows[component]; ++i)
			for (unsigned int j=0; j < matrix_sizes[component][i]; ++j)
			{
				dof2component	[jj]	=	component;
				dof2matrix		[jj]	=	i;
				dof2matrixdof	[jj]	=	j;
				++jj;
			}
	}

	offset.resize(dim);
	for (unsigned int component=0; component < matrices.size(); ++component)
		offset[component].resize(n_rows[component]);

	unsigned int temp = 0;

	for (unsigned int component=0; component < matrices.size(); ++component)
		for (unsigned int matrix=0; matrix < matrices[component].size(); ++matrix)
		{
			offset[component][matrix]=temp;
			temp += matrix_sizes[component][matrix];
		}
}


template <int dim>
void ADIMatrix<dim>::copy_from (
		const double				coeff,
		const ADIMatrix<dim>	&mat
		)
{
	//**********************************************************************
	// This routine simply copies all the member variables of another
	// ADIMatrix into the current object and multiplies the matrix entries
	// by a constant coefficient coeff.
	// This routine is quite inefficient, probably because of the need to
	// reinit the matrices everytime it is called. This can be alleviated
	// by sacrificing memory and constructing one system-matrix object for
	// each direction and then only admit copying from ADIMatrix objects
	// with the same direction, such that no calls to reinit (and also no
	// copying of some other stuff) are necessary, since both objects share
	// the same structure.
	//**********************************************************************

// 	if (direction != mat.direction)
// 	{
		
		copy_from(mat);
		if (coeff != 1)
			for (unsigned int component=0; component < matrices.size(); ++component)
				for (unsigned int i=0; i < matrices[component].size(); ++i)
					matrices[component][i] *= coeff;
// 	}
// 	else
// 	{
// 		for (unsigned int component=0; component < matrices.size(); ++component)
// 			for (unsigned int i=0; i < matrices[component].size(); ++i)
// 				matrices[component][i].copy_from(mat.matrices[component][i]);
// 		if (coeff != 1)
// 			for (unsigned int component=0; component < matrices.size(); ++component)
// 				for (unsigned int i=0; i < matrices[component].size(); ++i)
// 					matrices[component][i] *= coeff;
// 	}
}


template <int dim>
void ADIMatrix<dim>::copy_from (
		const ADIMatrix<dim>	&mat
		)
{
	//**********************************************************************
	// see the corresponding routine with variable coefficient.
	//**********************************************************************

// 	if (direction != mat.direction)
// 	{
		direction = mat.direction;
		comp_dofs_per_cell = mat.comp_dofs_per_cell;
		reordering = mat.reordering;
		dim_n = mat.dim_n;
		dim_m = mat.dim_m;

		dof2component=mat.dof2component;
		dof2matrix=mat.dof2matrix;
		dof2matrixdof=mat.dof2matrixdof;
		n_dofs_per_component=mat.n_dofs_per_component;
		
		matrix_sizes=mat.matrix_sizes;
		offset=mat.offset;

		matrices.resize(mat.matrices.size());

		for (unsigned int component=0; component < matrices.size(); ++component)
		{
			matrices[component].resize(mat.matrices[component].size());

			for (unsigned int i=0; i < matrices[component].size(); ++i)
				matrices[component][i].copy_from(mat.matrices[component][i]);
		}
// 	}
// 	else
// 	{
// 		for (unsigned int component=0; component < matrices.size(); ++component)
// 			for (unsigned int i=0; i < matrices[component].size(); ++i)
// 				matrices[component][i].copy_from(mat.matrices[component][i]);
// 	}
}


template <int dim>
void ADIMatrix<dim>::add (
		const std::vector< types::global_dof_index >	&dofs,
		const std::vector< types::global_dof_index >	&neighbor_dofs,
		const FullMatrix< double >						&mat
		)
{
	//**********************************************************************
	// This routine copies the entries of a FullMatrix object into the
	// ADIMatrix at the place given by the two vectors dofs and neighbor_dofs.
	// This is mainly important for assemblage of the system matrices and
	// currently also for the adding of ADIMassMatrix objects, even though
	// this is inefficient since some overhead has to be done for determining
	// the place where the corresponding entry lies in the SparseMatrix objects.
	//**********************************************************************

	// make sure that the dimensions of the dof-vectors and the FullMatrix coincide.
	Assert (mat.n()==dofs.size() && mat.m()==neighbor_dofs.size(), ExcInternalError());

	unsigned int component = dof_to_component(dofs[0]);

	// get the index of the matrix to which the dofs belong (maybe an Assertion should
	// be executed to check, if the other dofs belong to the same matrix, but by design,
	// this should always be the case, since only the ADIMassMatrix calls (and should call)
	// this routine).
	unsigned int matrix_index = dof_to_matrix(dofs[0]);

	// transform the global dofs given as parameters into (matrix-)local dofs.
	std::vector< types::global_dof_index > new_dofs(dofs), new_neighbor_dofs(neighbor_dofs);

	dofs_to_matrix_dofs(new_dofs);
	dofs_to_matrix_dofs(new_neighbor_dofs);

	// do the adding by using the corresponding routine of the matrix-object
	matrices[component][matrix_index].add(dofs, neighbor_dofs, mat);
}


template <int dim>
void ADIMatrix<dim>::add (
		const std::vector< types::global_dof_index >	&dofs,
		const FullMatrix< double >						&mat
		)
{
	//**********************************************************************
	// This is the same routine as the one with another parameter,
	// neighbor_dofs, but for quadratic FullMatrices.
	//**********************************************************************

	Assert (mat.n()==dofs.size() && mat.m()==dofs.size(), ExcInternalError());

	unsigned int component = dof_to_component(dofs[0]);

	unsigned int matrix_index = dof_to_matrix(dofs[0]);

	std::vector< types::global_dof_index > new_dofs(dofs);

	for (unsigned int i=0; i<new_dofs.size(); ++i)
		new_dofs[i] = dofs[i];

	dofs_to_matrix_dofs(new_dofs);

	matrices[component][matrix_index].add(new_dofs, mat);
}


template <int dim>
void ADIMatrix<dim>::add (
		types::global_dof_index	dof_i,
		types::global_dof_index	dof_j,
		const double			val
		)
{
	//**********************************************************************
	// This routine adds the value given by val into the place given by
	// global coordinates (dof_i, dof_j).
	// It's not really used, but could be (in a modified version) used in a
	// more efficient approach to adding ADIMassMatrix objects into
	// ADIMatrix-objects. Then, the index of the SparseMatrix val-vector
	// should be computed by the ADISparsityPattern object based on the
	// given dof_i and dof_j and val should be directly copied into this
	// val-vector.
	//**********************************************************************

	if (fabs(val) > 1e-17)
	{
		unsigned int component = dof_to_component(dof_i);

		unsigned int matrix_index = dof_to_matrix(dof_i);

		unsigned int new_dof_i = dof_to_matrix_dof(dof_i);
		unsigned int new_dof_j = dof_to_matrix_dof(dof_j);
                
//                 if(new_dof_i == 405 && new_dof_j == 1194)
//                 {
//                     std::cout << "component: " << component << 
//                     std::endl << "matrix_index: " << matrix_index << 
//                     std::endl << "dof_i: " << dof_i << 
//                     std::endl << "dof_j: " << dof_j << 
//                     std::endl << "val: " << val << 
//                     std::endl << std::endl;
//                 }

//		std::cout << "0 -- " << new_dof_i << " -- " << new_dof_j << " -- " << val << std::endl;

//		if (new_dof_i==8)
//			if (new_dof_j==9)
//				std::cout << "Value " << val << " at global position (" << dof_i << "," << dof_j << ") will be added to Matrix " << matrix_index << " of component " << component << " at position (" << new_dof_i << "," << new_dof_j << ")." << std::endl;
//
//				if(abs(dof_i-dof_j)>=32)
//					std::cout << "Error! -- " << abs(dof_i-dof_j)<< std::endl;

		matrices[component][matrix_index].add(new_dof_i, new_dof_j, val);
	}
}

template <int dim>
void ADIMatrix<dim>::add (
		const ADIMassMatrix< dim >			&mat,
		const std::vector< unsigned int >	&transform
		)
{
	//**********************************************************************
	// This routine copies the entries of a ADIMassMatrix object into the
	// ADIMatrix. It is crucial to the assemblage process, but unfortunately,
	// at this moment (14.03.16), not very efficient, since the adding
	// routine used for doing the actual adding is not efficient as well.
	// See the routine adding FullMatrices to the object for further
	// explanations.
	//**********************************************************************

	// Koennte man vermutlich zu Ungunsten von assemble-time und memory effizienter gestalten.
	// Dazu in ADISparsityPattern zu jedem prototype eine Liste erstellen, die (matrix-)lokale
	// (i,j) Koordinaten in Indizes fuer den val-vector der SparseMatrices (aus matrices) umwandelt.
	// Dann also zunaechst wie gehabt in lokale (i,j) Koordinaten umwandeln (dof_to_matrixdof), dann
	// ueber ADISparsityPattern in Indices und dann direkt in den val-vector schreiben. Dazu muss
	// allerdings eine SparseMatrix-Routine dafuer implementiert werden, die direkt auf val operiert.
	// Lohnt sich nur, falls variable timesteps zugelassen werden, da dann in jedem Schritt erneutes
	// assemblieren noetig ist.
	// Um den Index-vector (vector<vector>) aufzubauen und abzuspeichern muss natuerlich Memory
	// und zu Beginn einmal Aufwand geopfert werden.

	// traverse all elements and add the corresponding mass matrix stored in the ADIMassMatrix (as
	// FullMatrix) into the corresponding place of this object. If this object has another direction
	// than the ADIMassMatrix object, the dofs corresponding to the current element are transformed.
	std::vector< unsigned int > dofs(comp_dofs_per_cell);

	unsigned int n_mass_matrices = mat.get_n_element_matrices();

	for (unsigned int i=0; i < comp_dofs_per_cell; ++i)
		dofs[i] = i;

	if (direction == mat.get_direction())
	{
		for (unsigned int i=0; i < n_mass_matrices; ++i)
		{
//			//TODO delete
//			for (unsigned int j=0; j< dofs.size(); ++j)
//				std::cout << dofs[j] << std::endl;
//
//			std::cout << std::endl << i;
//			//TODO delete

			add(dofs, mat.get_element_matrix(i));

//			// TODO Performance gegen obige Variante testen.
//			for (unsigned int ii=0; ii<mat.get_element_matrix(i).n();++ii)
//				for (unsigned int jj=0; jj<mat.get_element_matrix(i).m();++jj)
//				{
//					add(dofs[ii], dofs[jj], mat.get_element_matrix(i)(ii,jj));
//				}
//			// TODO deĺete

			for (unsigned j=0; j < comp_dofs_per_cell; ++j)
				dofs[j] += comp_dofs_per_cell;
		}
	}
	else
	{
		std::vector< unsigned int > transformed_dofs(comp_dofs_per_cell);

		for (unsigned int i=0; i < n_mass_matrices; ++i)
		{
			for (unsigned j=0; j < comp_dofs_per_cell; ++j)
				transformed_dofs[j] = transform_dofs(i * comp_dofs_per_cell + j, transform);
//				transformed_dofs[j] = transform[i * dofs_per_cell + j];

			add(transformed_dofs, mat.get_element_matrix(i));

//			// TODO Performance gegen obige Variante testen.
//			for (unsigned int ii=0; ii<mat.get_element_matrix(i).n();++ii)
//				for (unsigned int jj=0; jj<mat.get_element_matrix(i).m();++jj)
//				{
//					add(transformed_dofs[ii], transformed_dofs[jj], mat.get_element_matrix(i)(ii,jj));
//				}
//			// TODO deĺete?

			for (unsigned j=0; j < comp_dofs_per_cell; ++j)
				dofs[j] += comp_dofs_per_cell;
		}
	}
}


template <int dim>
void ADIMatrix<dim>::add (
		const double				coeff,
		const ADIMatrix< dim >	&mat
		)
{
	//**********************************************************************
	// This routine adds coeff*mat to the current object, where mat is
	// another ADIMatrix object with the same direction (otherwise an
	// Exception is thrown, since adding would be very inefficient and is
	// not intended).
	// This is done by simply calling the corresponding routine of the
	// SparseMatrix class for each object in the matrices vector.
	//**********************************************************************

	Assert(direction == mat.get_direction(), ExcInternalError());

	for (unsigned int component=0; component < matrices.size(); ++component)
		for (unsigned int i=0; i < matrices[component].size(); ++i)
			matrices[component][i].add(coeff, mat.matrices[component][i]);
}


template <int dim>
void ADIMatrix<dim>::add (
		const ADIMatrix< dim > &mat
		)
{
	//**********************************************************************
	// see the corresponding routine with variable coefficient.
	//**********************************************************************

	add(1,mat);
}


template <int dim>
unsigned int ADIMatrix<dim>::get_direction () const
{
	//**********************************************************************
	// This routine just returns the direction of the object.
	//**********************************************************************

	return direction;
}


//template <int dim>
//void ADIMatrix<dim>::mmult (
//		ADIMatrix< dim > 			&C,
//		const ADIMatrix< dim > 	&B
//		) const
//{
//	Assert(B.direction == direction, ExcInternalError());
//	Assert(B.dim_n == dim_n, ExcInternalError());
//	Assert(C.direction == direction, ExcInternalError());
//	Assert(C.dim_n == dim_n, ExcInternalError());
//
//	// loop over all matrices, i.e., subproblems
//	for (unsigned int component=0; component < dim; ++component)
//		for (unsigned int i=0; i < matrices[component].size(); ++i)
//			matrices[component][i].mmult(C.matrices[component][i], B.matrices[component][i], Vector< double >(), false);
//}
//
//
//template <int dim>
//void ADIMatrix<dim>::Tmmult (
//		ADIMatrix< dim >		&C,
//		const ADIMatrix< dim > &B
//		) const
//{
//	Assert(B.direction == direction, ExcInternalError());
//	Assert(B.dim_n == dim_n, ExcInternalError());
//	Assert(C.direction == direction, ExcInternalError());
//	Assert(C.dim_n == dim_n, ExcInternalError());
//
//	// loop over all matrices, i.e., subproblems
//	for (unsigned int component=0; component < dim; ++component)
//		for (unsigned int i=0; i < matrices[component].size(); ++i)
//			matrices[component][i].Tmmult(C.matrices[component][i], B.matrices[component][i], Vector< double >(), false);
//}


//template <int dim>
//void ADIMatrix<dim>::sequential_invert ()
//{
//	for (unsigned int component=0; component < dim; ++component)
//		for (unsigned int matrix=0; matrix < matrix_sizes[component].size(); ++matrix)
//			matrices[component][matrix].invert();
//}


template <int dim>
void ADIMatrix<dim>::invert ()
{
	for (unsigned int component=0; component < dim; ++component)
		for (unsigned int matrix=0; matrix < matrix_sizes[component].size(); ++matrix)
			matrices[component][matrix].invert();
}


// template <int dim>
// void ADIMatrix<dim>::invert ()
// {
// 	for (unsigned int component=0; component < matrices.size(); ++component)
// 	{
// 		for (unsigned int matrix=0; matrix < matrices[component].size(); ++matrix)
// 			tasks.push(std::make_pair(component, matrix));
// 	}
// 
// 	auto thread_func = &ADIMatrix<dim>::invert_function;
// 
// 	std::vector< std::thread > threads;
// 
// 	for (unsigned int ii=0; ii<n_threads; ++ii)
// 	{
// 		threads.push_back(std::thread(thread_func, this));
// //                 std::cout << threads.size() << " -- ";
// 	}
// //         std::cout << std::endl << std::endl;
// 
// 	for (unsigned int ii=0; ii<n_threads; ++ii)
// 		threads[ii].join();
// }
// 
// 
// template <int dim>
// void ADIMatrix<dim>::invert_function ()
// {
// 	while (true)
// 	{
// 		mtx.lock();
// 
// 		if(tasks.empty())
// 		{
// 				mtx.unlock();
// 				break;
// 		}
// 
// 		std::pair<unsigned int, unsigned int> current_task = tasks.top();
// 		tasks.pop();
// 
// 		mtx.unlock();
// 
// 		matrices[current_task.first][current_task.second].invert();
// 	}
// }


//template <int dim>
//void ADIMatrix<dim>::sequential_inverse_vmult (
//		Vector< double > 					&vec_out,
//		const Vector< double > 				&vec_in,
//		const std::vector< unsigned int > 	&transform
//		) const
//{
//	//**********************************************************************
//	// This routine performs vector multiplication and stores ADIMatrix*vec_in
//	// into the vector vec_out, where ADIMatrix is the calling instance of
//	// this class.
//	//**********************************************************************
//
//	//TODO Evtl. Alternative implementieren, bei der gesamter vec_in transformiert und
//	// zum Schluss gesamter vec_out zuruecktransformiert wird.
//
//	Assert(vec_in.size() == vec_out.size(), ExcInternalError());
//
//	// initialize temporary vectors corresponding to the vectors of the subproblems
//	// and the offset-variable keeping track of the dofs of the current subproblem.
//	unsigned int offset = 0, former_matrix_size = 0;
//	Vector< double > temp_in, temp_out, temp_transform;
//
//	// loop over all matrices, i.e., subproblems
//	for (unsigned int component=0; component < dim; ++component)
//		for (unsigned int i=0; i < matrices[component].size(); ++i)
//		{
//			// if we have a L-shaped domain, matrix sizes could vary, so we keep track of
//			// this and resize the temp-vectors correspondingly.
//			if (former_matrix_size != matrix_sizes[component][i])
//			{
//				temp_in.reinit(matrix_sizes[component][i]);
//				temp_out.reinit(matrix_sizes[component][i]);
//				temp_transform.reinit(matrix_sizes[component][i]);
//			}
//
//			// use the vmult-routine of the SparseMatrix class to do the actual vector
//			// multiplication on the subproblems. If necessary, reorder the dofs beforehand
//			// and retransform them afterwards.
//			if (reordering)
//			{
//				for (unsigned int j=0; j < matrix_sizes[component][i]; ++j)
//				{
//					temp_in[j] = vec_in[transform_dofs(offset + j, transform)];
//					temp_transform[j] = transform_dofs(offset + j, transform);
//	//				temp_in[j] = vec_in[transform[offset + j]];
//	//				temp_transform[j] = transform[offset + j];
//				}
//
//				matrices[component][i].inverse_vmult(temp_out,temp_in);
//
//				for (unsigned int j=0; j < matrix_sizes[component][i]; ++j)
//					vec_out[temp_transform[j]] = temp_out[j];
//			}
//			else
//			{
//				for (unsigned int j=0; j < matrix_sizes[component][i]; ++j)
//					temp_in[j] = vec_in[offset + j];
//
//				matrices[component][i].inverse_vmult(temp_out,temp_in);
//
//				for (unsigned int j=0; j < matrix_sizes[component][i]; ++j)
//					vec_out[offset + j] = temp_out[j];
//			}
//
//			// increment the offset variable and store the size of the current matrix so it
//			// can be compared to the next one to make sure that the temp vectors are resized
//			// if necessary.
//			offset += matrix_sizes[component][i];
//			former_matrix_size = matrix_sizes[component][i];
//		}
//}


template <int dim>
void ADIMatrix<dim>::inverse_vmult (
		Vector< double > 					&vec_out,
		const Vector< double > 				&vec_in,
		const std::vector< unsigned int > 	&transform
		) const
{

	Assert(vec_in.size() == vec_out.size(), ExcInternalError());

	// TODO nur fuer Threading, sollte man evtl als abgeleitete Klasse oder so machen..
	for (unsigned int component=0; component < matrices.size(); ++component)
			for (unsigned int matrix=0; matrix < matrices[component].size(); ++matrix)
					tasks.push(std::make_pair(component, matrix));
	// TODO nur fuer Threading

	auto thread_func = &ADIMatrix<dim>::inverse_vmult_function;

	std::vector< std::thread > threads;

	for (unsigned int ii=0; ii<n_threads; ++ii)
	{
		threads.push_back(std::thread(thread_func, this, std::ref(vec_out), std::cref(vec_in), std::cref(transform)));
//                 std::cout << threads.size() << " -- ";
	}
//         std::cout << std::endl << std::endl;

	for (unsigned int ii=0; ii<n_threads; ++ii)
		threads[ii].join();
}


template <int dim>
void ADIMatrix<dim>::inverse_vmult_function (
		Vector< double > 					&vec_out,
		const Vector< double > 				&vec_in,
		const std::vector< unsigned int >	&transform
		) const
{
	while (true)
	{
		mtx.lock();

		if(tasks.empty())
		{
				mtx.unlock();
				break;
		}

		std::pair<unsigned int, unsigned int> current_task = tasks.top();
		tasks.pop();

		mtx.unlock();

		// initialize temporary vectors corresponding to the vectors of the subproblems
		// and the offset-variable keeping track of the dofs of the current subproblem.
		Vector< double > temp_in, temp_out, temp_transform;

		temp_in.reinit(matrix_sizes[current_task.first][current_task.second]);
		temp_out.reinit(matrix_sizes[current_task.first][current_task.second]);
		temp_transform.reinit(matrix_sizes[current_task.first][current_task.second]);

		// use the vmult-routine of the SparseMatrix class to do the actual vector
		// multiplication on the subproblems. If necessary, reorder the dofs beforehand
		// and retransform them afterwards.
		if (reordering)
		{
			for (unsigned int j=0; j< matrix_sizes[current_task.first][current_task.second]; ++j)
			{
				temp_in[j] = vec_in[transform_dofs(offset[current_task.first][current_task.second] + j, transform)];
				temp_transform[j] = transform_dofs(offset[current_task.first][current_task.second] + j, transform);
//				temp_in[j] = vec_in[transform[offset + j]];
//				temp_transform[j] = transform[offset + j];
			}

			matrices[current_task.first][current_task.second].inverse_vmult(temp_out,temp_in);

			for (unsigned int j=0; j< matrix_sizes[current_task.first][current_task.second]; ++j)
				vec_out[temp_transform[j]] = temp_out[j];
		}
		else
		{
			for (unsigned int j=0; j< matrix_sizes[current_task.first][current_task.second]; ++j)
				temp_in[j] = vec_in[offset[current_task.first][current_task.second] + j];

			matrices[current_task.first][current_task.second].inverse_vmult(temp_out,temp_in);

			for (unsigned int j=0; j< matrix_sizes[current_task.first][current_task.second]; ++j)
				vec_out[offset[current_task.first][current_task.second] + j] = temp_out[j];
		}
	}
}


//template <int dim>
//void ADIMatrix<dim>::sequential_vmult (
//		Vector< double > 					&vec_out,
//		const Vector< double > 				&vec_in,
//		const std::vector< unsigned int > 	&transform
//		) const
//{
//	//**********************************************************************
//	// This routine performs vector multiplication and stores ADIMatrix*vec_in
//	// into the vector vec_out, where ADIMatrix is the calling instance of
//	// this class.
//	//**********************************************************************
//
//	//TODO Evtl. Alternative implementieren, bei der gesamter vec_in transformiert und
//	// zum Schluss gesamter vec_out zuruecktransformiert wird.
//
//	Assert(vec_in.size() == vec_out.size(), ExcInternalError());
//
//	// initialize temporary vectors corresponding to the vectors of the subproblems
//	// and the offset-variable keeping track of the dofs of the current subproblem.
//	unsigned int offset = 0, former_matrix_size = 0;
//	Vector< double > temp_in, temp_out, temp_transform;
//
//	// loop over all matrices, i.e., subproblems
//	for (unsigned int component=0; component < dim; ++component)
//		for (unsigned int i=0; i < matrices[component].size(); ++i)
//		{
//			// if we have a L-shaped domain, matrix sizes could vary, so we keep track of
//			// this and resize the temp-vectors correspondingly.
//			if (former_matrix_size != matrix_sizes[component][i])
//			{
//				temp_in.reinit(matrix_sizes[component][i]);
//				temp_out.reinit(matrix_sizes[component][i]);
//				temp_transform.reinit(matrix_sizes[component][i]);
//			}
//
//			// use the vmult-routine of the SparseMatrix class to do the actual vector
//			// multiplication on the subproblems. If necessary, reorder the dofs beforehand
//			// and retransform them afterwards.
//			if (reordering)
//			{
//				for (unsigned int j=0; j < matrix_sizes[component][i]; ++j)
//				{
//					temp_in[j] = vec_in[transform_dofs(offset + j, transform)];
//					temp_transform[j] = transform_dofs(offset + j, transform);
//				}
//
//				matrices[component][i].vmult(temp_out,temp_in);
//
//				for (unsigned int j=0; j < matrix_sizes[component][i]; ++j)
//					vec_out[temp_transform[j]] = temp_out[j];
//			}
//			else
//			{
//				for (unsigned int j=0; j < matrix_sizes[component][i]; ++j)
//					temp_in[j] = vec_in[offset + j];
//
//				matrices[component][i].vmult(temp_out,temp_in);
//
//				for (unsigned int j=0; j < matrix_sizes[component][i]; ++j)
//					vec_out[offset + j] = temp_out[j];
//			}
//
//			// increment the offset variable and store the size of the current matrix so it
//			// can be compared to the next one to make sure that the temp vectors are resized
//			// if necessary.
//			offset += matrix_sizes[component][i];
//			former_matrix_size = matrix_sizes[component][i];
//		}
//}


template <int dim>
void ADIMatrix<dim>::vmult (
		Vector< double > 					&vec_out,
		const Vector< double > 				&vec_in,
		const std::vector< unsigned int > 	&transform
		) const
{
	Assert(vec_in.size() == vec_out.size(), ExcInternalError());

	// TODO nur fuer Threading, sollte man evtl als abgeleitete Klasse oder so machen..
	for (unsigned int component=0; component < matrices.size(); ++component)
			for (unsigned int matrix=0; matrix < matrices[component].size(); ++matrix)
					tasks.push(std::make_pair(component, matrix));
	// TODO nur fuer Threading

	auto thread_func = &ADIMatrix<dim>::vmult_function;

	std::vector< std::thread > threads;

	for (unsigned int ii=0; ii<n_threads; ++ii)
        {
		threads.push_back(std::thread(thread_func, this, std::ref(vec_out), std::cref(vec_in), std::cref(transform)));
//                 std::cout << threads.size() << " -- ";
        }
//         std::cout << std::endl << std::endl;

	for (unsigned int ii=0; ii<n_threads; ++ii)
		threads[ii].join();
}


template <int dim>
void ADIMatrix<dim>::vmult_function (
		Vector< double > 					&vec_out,
		const Vector< double > 				&vec_in,
		const std::vector< unsigned int >	&transform
		) const
{
	while (true)
	{
		mtx.lock();

		if(tasks.empty())
		{
				mtx.unlock();
				break;
		}

		std::pair<unsigned int, unsigned int> current_task = tasks.top();
		tasks.pop();

		mtx.unlock();

		// initialize temporary vectors corresponding to the vectors of the subproblems
		// and the offset-variable keeping track of the dofs of the current subproblem.
		Vector< double > temp_in, temp_out, temp_transform;

		temp_in.reinit(matrix_sizes[current_task.first][current_task.second]);
		temp_out.reinit(matrix_sizes[current_task.first][current_task.second]);
		temp_transform.reinit(matrix_sizes[current_task.first][current_task.second]);

		// use the vmult-routine of the SparseMatrix class to do the actual vector
		// multiplication on the subproblems. If necessary, reorder the dofs beforehand
		// and retransform them afterwards.
		if (reordering)
		{
			for (unsigned int j=0; j< matrix_sizes[current_task.first][current_task.second]; ++j)
			{
				temp_in[j] = vec_in[transform_dofs(offset[current_task.first][current_task.second] + j, transform)];
				temp_transform[j] = transform_dofs(offset[current_task.first][current_task.second] + j, transform);
			}

			matrices[current_task.first][current_task.second].vmult(temp_out,temp_in);

			for (unsigned int j=0; j< matrix_sizes[current_task.first][current_task.second]; ++j)
				vec_out[temp_transform[j]] = temp_out[j];
		}
		else
		{
			for (unsigned int j=0; j< matrix_sizes[current_task.first][current_task.second]; ++j)
				temp_in[j] = vec_in[offset[current_task.first][current_task.second] + j];

			matrices[current_task.first][current_task.second].vmult(temp_out,temp_in);

			for (unsigned int j=0; j< matrix_sizes[current_task.first][current_task.second]; ++j)
				vec_out[offset[current_task.first][current_task.second] + j] = temp_out[j];
		}
	}
}


template <int dim>
void ADIMatrix<dim>::set_reordering_flag (
		const bool reord_flag
		)
{
	reordering = reord_flag;
}


template <int dim>
void ADIMatrix<dim>::set_direction (
		unsigned dir
		)
{
	direction = dir;
}


template <int dim>
void ADIMatrix<dim>::set_reordering_flag_from_system_dir (
		unsigned int system_dir
		)
{
	//**********************************************************************
	// This routine sets the reordering_flag, necessary to decide if
	// transformation of dofs is necessary. This is crucial as otherwise the
	// program will produce wrong results or even won't work at all. It has
	// to be called after the last call to the reorder_dofs routine.
	//**********************************************************************

	if (system_dir==direction)
		reordering = false;
	else
		reordering = true;
}


template <int dim>
void ADIMatrix<dim>::save_matrix (
		std::stringstream &filename
		) const
{
	filename << ".mat";

	std::ofstream out ((filename.str()).c_str());
	
	out << "direction:" << std::endl << direction << std::endl 
		<< "comp_dofs_per_cell:" << std::endl << comp_dofs_per_cell << std::endl 
		<< "dim_n:" << std::endl << dim_n << std::endl 
		<< "dim_m:" << std::endl << dim_m << std::endl 
		<< "reordering:" << std::endl << reordering << std::endl << std::endl;
	
	{
		out << "matrix_sizes:" << std::endl << matrix_sizes.size() << std::endl;
		auto r_it = matrix_sizes.begin(), end_r_it = matrix_sizes.end();
		
		for (; r_it != end_r_it; ++r_it)
		{
			out << (*r_it).size() << std::endl;
			auto c_it = (*r_it).begin(), end_c_it = (*r_it).end();
			for (; c_it != end_c_it; ++c_it)
				out << *c_it << "\t";
			out << std::endl;
		}
		out << std::endl;
	}
	
	{
		out << "offset:" << std::endl << offset.size() << std::endl;
		auto r_it = offset.begin(), end_r_it = offset.end();
		
		for (; r_it != end_r_it; ++r_it)
		{
			out << (*r_it).size() << std::endl;
			auto c_it = (*r_it).begin(), end_c_it = (*r_it).end();
			for (; c_it != end_c_it; ++c_it)
				out << *c_it << "\t";
			out << std::endl;
		}
		out << std::endl;
	}
	
	{
		out << "dof2component:" << std::endl << dof2component.size() << std::endl;
		auto it = dof2component.begin(), end_it = dof2component.end();
		
		for (; it != end_it; ++it)
			out << *it << "\t";
		out << std::endl << std::endl;
	}
	
	{
		out << "n_dofs_per_component:" << std::endl << n_dofs_per_component.size() << std::endl;
		auto it = n_dofs_per_component.begin(), end_it = n_dofs_per_component.end();
		
		for (; it != end_it; ++it)
			out << *it << "\t";
		out << std::endl << std::endl;
	}
	
	{
		out << "dof2matrix:" << std::endl << dof2matrix.size() << std::endl;
		auto it = dof2matrix.begin(), end_it = dof2matrix.end();
		
		for (; it != end_it; ++it)
			out << *it << "\t";
		out << std::endl << std::endl;
	}
	
	{
		out << "dof2matrixdof:" << std::endl << dof2matrixdof.size() << std::endl;
		auto it = dof2matrixdof.begin(), end_it = dof2matrixdof.end();
		
		for (; it != end_it; ++it)
			out << *it << "\t";
		out << std::endl << std::endl;
	}
	
	{
		out << "data:" << std::endl << matrices.size() << std::endl;
		auto r_it = matrices.begin(), end_r_it = matrices.end();
		
		for (; r_it != end_r_it; ++r_it)
		{
			out << (*r_it).size() << std::endl;
			
			auto c_it = (*r_it).begin(), end_c_it = (*r_it).end();
		
			for (; c_it != end_c_it; ++c_it)
			{
				(*c_it).save_matrix(out);
			
// 				out << std::endl;
			}
			
// 			out << std::endl;
		}
	}
}


template <int dim>
void ADIMatrix<dim>::load_matrix (
		std::stringstream &filename
		)
{
	filename << ".mat";

	std::ifstream in ((filename.str()).c_str());
	if (!in)
	{
		std::cout << "Unable to open file.";
		exit(1);
	}
	
	char buff[256];
	
	in.getline(buff, 256);
	if (!strcmp(buff, "direction:"))
	{
		in.getline(buff, 256);
		direction = atoi(buff);
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}
	
	in.getline(buff, 256);
	if (!strcmp(buff, "comp_dofs_per_cell:"))
	{
		in.getline(buff, 256);
		comp_dofs_per_cell = atoi(buff);
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}
	
	in.getline(buff, 256);
	if (!strcmp(buff, "dim_n:"))
	{
		in.getline(buff, 256);
		dim_n = atoi(buff);
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}
	
	in.getline(buff, 256);
	if (!strcmp(buff, "dim_m:"))
	{
		in.getline(buff, 256);
		dim_m = atoi(buff);
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}
	
	in.getline(buff, 256);
	if (!strcmp(buff, "reordering:"))
	{
		in.getline(buff, 256);
		reordering = atoi(buff);
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}
	
	char delim = '\t';

	in.getline(buff, 256);
	while (!strcmp(buff, ""))
		in.getline(buff, 256);
		
	if (!strcmp(buff, "matrix_sizes:"))
	{
		in.getline(buff, 256);
		matrix_sizes.resize(atoi(buff));
		
		auto r_it = matrix_sizes.begin(), end_r_it = matrix_sizes.end();
		for (; r_it != end_r_it; ++r_it)
		{
			in.getline(buff, 256);
			while(!strcmp(buff, ""))
				in.getline(buff, 256);
			(*r_it).resize(atoi(buff));
			auto c_it = (*r_it).begin(), end_c_it = (*r_it).end();
			for (; c_it != end_c_it; ++c_it)
			{
				in.getline(buff, 256, delim);
				*c_it = atoi(buff);
			}
		}
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}

	in.getline(buff, 256);
	while (!strcmp(buff, ""))
		in.getline(buff, 256);
		
	if (!strcmp(buff, "offset:"))
	{
		in.getline(buff, 256);
		offset.resize(atoi(buff));
		
		auto r_it = offset.begin(), end_r_it = offset.end();
		for (; r_it != end_r_it; ++r_it)
		{
			in.getline(buff, 256);
			while(!strcmp(buff, ""))
				in.getline(buff, 256);
			(*r_it).resize(atoi(buff));
			auto c_it = (*r_it).begin(), end_c_it = (*r_it).end();
			for (; c_it != end_c_it; ++c_it)
			{
				in.getline(buff, 256, delim);
				*c_it = atoi(buff);
			}
		}
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}

	in.getline(buff, 256);
	while (!strcmp(buff, ""))
		in.getline(buff, 256);
		
	if (!strcmp(buff, "dof2component:"))
	{
		in.getline(buff, 256);
		dof2component.resize(atoi(buff));
		
		auto it = dof2component.begin(), end_it = dof2component.end();
		for (; it != end_it; ++it)
		{
			in.getline(buff, 256, delim);
				*it = atoi(buff);
		}
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}

	in.getline(buff, 256);
	while (!strcmp(buff, ""))
		in.getline(buff, 256);
		
	if (!strcmp(buff, "n_dofs_per_component:"))
	{
		in.getline(buff, 256);
		n_dofs_per_component.resize(atoi(buff));
		
		auto it = n_dofs_per_component.begin(), end_it = n_dofs_per_component.end();
		for (; it != end_it; ++it)
		{
			in.getline(buff, 256, delim);
				*it = atoi(buff);
		}
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}

	in.getline(buff, 256);
	while (!strcmp(buff, ""))
		in.getline(buff, 256);
		
	if (!strcmp(buff, "dof2matrix:"))
	{
		in.getline(buff, 256);
		dof2matrix.resize(atoi(buff));
		
		auto it = dof2matrix.begin(), end_it = dof2matrix.end();
		for (; it != end_it; ++it)
		{
			in.getline(buff, 256, delim);
				*it = atoi(buff);
		}
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}

	in.getline(buff, 256);
	while (!strcmp(buff, ""))
		in.getline(buff, 256);
		
	if (!strcmp(buff, "dof2matrixdof:"))
	{
		in.getline(buff, 256);
		dof2matrixdof.resize(atoi(buff));
		
		auto it = dof2matrixdof.begin(), end_it = dof2matrixdof.end();
		for (; it != end_it; ++it)
		{
			in.getline(buff, 256, delim);
				*it = atoi(buff);
		}
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}

	in.getline(buff, 256);
	while (!strcmp(buff, ""))
		in.getline(buff, 256);
		
	if (!strcmp(buff, "data:"))
	{
		in.getline(buff, 256);
		matrices.resize(atoi(buff));
		
		auto r_it = matrices.begin(), end_r_it = matrices.end();
		for (; r_it != end_r_it; ++r_it)
		{
			in.getline(buff, 256);
			while(!strcmp(buff, ""))
				in.getline(buff, 256);
			(*r_it).resize(atoi(buff));
			auto c_it = (*r_it).begin(), end_c_it = (*r_it).end();
			for (; c_it != end_c_it; ++c_it)
				(*c_it).load_matrix(in);
		}
	}
	else
	{
		std::cout << (filename.str()).c_str() << " is not a valid matrix file. Aborting." << std::endl << std::endl;
		exit(1);
	}
}

    
//template <int dim>
//void ADIMatrix<dim>::output_matrix (
//		std::stringstream &filename
//		) const
//{
//	//**********************************************************************
//	// This routine puts out the object in a format readable by visit or
//	// paraview, but without values, i.e. only as a spy pattern.
//	// Nevertheless, in the file itself, the values are written down.
//	//**********************************************************************
//
//	filename << ".vtk";
//
//	std::ofstream out ((filename.str()).c_str());
//
//	unsigned int offseti=0, offsetj=0;
//
//	for (unsigned int component=0; component < matrices.size(); ++component)
//		for (unsigned int ii=0; ii < matrices[component].size(); ++ii)
//		{
////			for (unsigned int i=0; i<matrix_sizes[component][ii]; ++i)
////				for (unsigned int j=0; j<matrix_sizes[component][ii]; ++j)
////					if (fabs(matrices[component][ii].el(i,j)) > 1e-17)
////						out << i+offseti << " -" << j+offsetj << "\t" << matrices[component][ii].el(i,j) << std::endl;
//
//			for (unsigned int row=0; row<matrix_sizes[component][ii]; ++row)
//			{
//				SparseMatrix<double>::const_iterator it = matrices[component][ii].begin(row);
//				SparseMatrix<double>::const_iterator end = matrices[component][ii].end(row);
//				for (; it!=end; ++it)
//					if (fabs((*it).value()) > 1e-17)
//						out << row+offseti << " -" << (*it).column()+offsetj << "\t" << (*it).value() << std::endl;
//			}
//
//			offseti+=matrix_sizes[component][ii];
//			offsetj+=matrix_sizes[component][ii];
//		}
//}
//
//
//template <int dim>
//void ADIMatrix<dim>::output_matrix_spy (
//		std::stringstream &filename
//		) const
//{
//	//**********************************************************************
//	// This routine puts out the object in a format readable by visit or
//	// paraview, but without values, i.e. only as a spy pattern.
//	//**********************************************************************
//
//	filename << "_spy.vtk";
//
//	std::ofstream out ((filename.str()).c_str());
//
//	unsigned int offseti=0, offsetj=0;
//
//	for (unsigned int component=0; component < matrices.size(); ++component)
//		for (unsigned int ii=0; ii < matrices[component].size(); ++ii)
//		{
////			for (unsigned int i=0; i<matrix_sizes[component][ii]; ++i)
////				for (unsigned int j=0; j<matrix_sizes[component][ii]; ++j)
////					if (fabs(matrices[component][ii].el(i,j)) > 1e-17)
////						out << i+offseti << " -" << j+offsetj << std::endl;
//
//			for (unsigned int row=0; row<matrix_sizes[component][ii]; ++row)
//			{
//				SparseMatrix<double>::const_iterator it = matrices[component][ii].begin(row);
//				SparseMatrix<double>::const_iterator end = matrices[component][ii].end(row);
//				for (; it!=end; ++it)
//					if (fabs((*it).value()) > 1e-17)
//						out << row+offseti << " -" << (*it).column()+offsetj << std::endl;
//			}
//
//			offseti+=matrix_sizes[component][ii];
//			offsetj+=matrix_sizes[component][ii];
//		}
//}
//
//
//template <int dim>
//void ADIMatrix<dim>::output_matrix_matlab (
//		std::stringstream &filename
//		) const
//{
//	//**********************************************************************
//	// This routine puts out the object in a format readable by visit or
//	// paraview, but without values, i.e. only as a spy pattern.
//	// Nevertheless, in the file itself, the values are written down.
//	//**********************************************************************
//
//	filename << "_matlab.vtk";
//
//	std::ofstream out ((filename.str()).c_str());
//
//	unsigned int offseti=0, offsetj=0;
//
//	for (unsigned int component=0; component < matrices.size(); ++component)
//		for (unsigned int ii=0; ii < matrices[component].size(); ++ii)
//		{
////			for (unsigned int i=0; i<matrix_sizes[component][ii]; ++i)
////				for (unsigned int j=0; j<matrix_sizes[component][ii]; ++j)
////					if (fabs(matrices[component][ii].el(i,j)) > 1e-17)
////						out << i+offseti+1 << "\t" << j+offsetj+1 << "\t" << matrices[component][ii].el(i,j) << std::endl;
//
//			for (unsigned int row=0; row<matrix_sizes[component][ii]; ++row)
//			{
//				SparseMatrix<double>::const_iterator it = matrices[component][ii].begin(row);
//				SparseMatrix<double>::const_iterator end = matrices[component][ii].end(row);
//				for (; it!=end; ++it)
//					if (fabs((*it).value()) > 1e-17)
//						out << row+offseti+1 << "\t" << (*it).column()+offsetj+1 << "\t" << (*it).value() << std::endl;
//			}
//
//			offseti+=matrix_sizes[component][ii];
//			offsetj+=matrix_sizes[component][ii];
//		}
//}


template <int dim>
unsigned int ADIMatrix<dim>::dof_to_component (
		const unsigned int dof
		)
{
	return dof2component[dof];
}


template <int dim>
unsigned int ADIMatrix<dim>::dof_to_matrix (
		const unsigned int	dof
		)
{
	return dof2matrix[dof];
}


template <int dim>
unsigned int ADIMatrix<dim>::dof_to_matrix_dof (
		const unsigned int dof
		)
{
	return dof2matrixdof[dof];
}


template <int dim>
void ADIMatrix<dim>::dofs_to_matrix_dofs (
		std::vector< types::global_dof_index >	&dofs
		)
{
	for (unsigned int i=0; i < dofs.size(); ++i)
		dofs[i] = dof_to_matrix_dof(dofs[i]);
}


template <int dim>
unsigned int ADIMatrix<dim>::transform_dofs (
		const unsigned int					dof,
		const std::vector< unsigned int >	&transform
		) const
{
	return transform[dof];
}


template <int dim>
void ADIMatrix<dim>::transform_dofs (
		std::vector< unsigned int >			&transformed_dofs,
		const std::vector< unsigned int >	&dofs,
		const std::vector< unsigned int >	&transform
		)
{
	Assert(transformed_dofs.size() == dofs.size(), ExcInternalError());
	for (unsigned int i=0; i < dofs.size(); ++i)
		transformed_dofs[i] = transform_dofs(dofs[i], transform);
}


//template <int dim>
//bool ADIMatrix<dim>::operator== (
//		const ADIMatrix<dim>	&mat
//		)
//{
//	//**********************************************************************
//	// Comparison of ADIMatrix-Objects, very costly, should only be called
//	// for testing purposes. This routine simply compares all the member
//	// variables.
//	//**********************************************************************
//
//	if (direction != mat.direction
//			|| comp_dofs_per_cell != mat.comp_dofs_per_cell
//			|| reordering != mat.reordering
//			|| matrices.size() != mat.matrices.size()
//			|| dof2matrix.size() != mat.dof2matrix.size())
//		return false;
//
//	for (unsigned int i=0; i < dof2matrix.size(); ++i)
//		if (dof2matrix[i] != mat.dof2matrix[i] || dof2matrixdof[i] != mat.dof2matrixdof[i])
//				return false;
//
//	for (unsigned int component=0; component < matrices.size(); ++component)
//		for (unsigned int i=0; i < matrices[component].size(); ++i)
//			for (SparseMatrix< double >::const_iterator entry1 = matrices[component][i].begin(), entry2 = mat.matrices[component][i].begin(); entry1 < matrices[component][i].end(); ++entry1, ++entry2)
//				if (!(*entry1 == *entry2))
//					return false;
//
//	return true;
//}

template class ADIMatrix<3>;

} /* namespace MaxwellProblem */
