#include "ADIGeometryInfo.h"

namespace MaxwellProblem {

template <int dim>
ADIGeometryInfo<dim>::ADIGeometryInfo ()
:
dofs_per_cell(0),
comp_dofs_per_cell(0),
n_elements(0),
n_components(0),
n_prototypes(0),
n_rows(dim),
n_el_per_row(dim),
matrix2sparsity(dim),
sorted_iters(dim)
{}

template <int dim>
void ADIGeometryInfo<dim>::initialize (
		const DoFHandler< dim >	&dof_handler
		)
{
	/*
	 * get basic geometric information from the DoFHandler
	 */
	n_elements = dof_handler.get_triangulation().n_active_cells();
	dofs_per_cell = dof_handler.get_fe().dofs_per_cell;
	n_components = dof_handler.get_fe().n_components();
	comp_dofs_per_cell = dofs_per_cell/n_components;

	/*
	 * this routine gathers basic geometric information and computes
	 * the reordered cell_iterators needed for reordering the DoFs
	 */
	compute_reordering(dof_handler);

	/*
	 *----------------------------------------------------------------------
	 * compute the ordered list of prototypes, i.e. the different sizes of
	 * cell-rows appearing for both directions.
	 *----------------------------------------------------------------------
	 *
	 * filter out duplicates by first storing all prototypes in a set
	 */
	std::set< unsigned int > tempset;

	for (unsigned int derivative_dir = 0; derivative_dir < dim; ++derivative_dir)
		for (unsigned int i=0; i < n_rows[derivative_dir]; ++i)
			tempset.insert(n_el_per_row[derivative_dir][i]);
	
	n_prototypes = tempset.size();

	/*
	 * store prototypes into the designated vector..
	 */
	prototypes.resize(n_prototypes);

	std::set< unsigned int >::iterator it=tempset.begin(), end=tempset.end();

	for (unsigned int i=0; it != end; ++it, ++i)
		prototypes[i] = *it;

	/*
	 * ..and afterwards sort them
	 */
	sort(prototypes.begin(), prototypes.end());

	/*
	 * compute the array holding the corresponding prototype for the i-th
	 * matrix in the corresponding direction. This is done by going through
	 * the array containing the row-size information and comparing its
	 * content with the prototypes. If both coincide, the index of the
	 * prototype is written into the current index and the next index will
	 * be considered.
	 */
	for (unsigned int derivative_dir=0; derivative_dir < dim; ++derivative_dir)
	{
		matrix2sparsity[derivative_dir].resize(n_rows[derivative_dir]);
		for (unsigned int i=0; i < n_rows[derivative_dir]; ++i)
			for (unsigned int j=0; j < n_prototypes; ++j)
				if (n_el_per_row[derivative_dir][i] == prototypes[j])
				{
					matrix2sparsity[derivative_dir][i]=j;
					break;
				}
	}
}


template<int dim>
void ADIGeometryInfo<dim>::compute_reordering (
		const DoFHandler< dim >	&dof_handler
		)
{
	/*
	 * This routine gathers basic geometric information, computes
	 * the reordered cell_iterators needed for reordering the DoFs
	 * and the DoF-wise transform and retransform arrays.
	 *
	 * It does NOT actually perform the renumbering, since this class is
	 * only supposed to gather the information needed for assembling the
	 * system. If it did the renumbering, it would have to also do the
	 * assemblage, since this has to be done with the correct dof order.
	 */

	/*
	 *----------------------------------------------------------------------
	 * Compute the arrays containing the sorted iterators used for reorder-
	 * ing the DoFs in the ADIDGMethod-class to achieve block-tridiagonal
	 * matrices and the corresponding DoF-wise transform.
	 *----------------------------------------------------------------------
	 *
	 * prepare the array which will be sorted by the first argument (non-
	 * constant coordinate of the midpoint of the cells) later.
	 */
	std::vector< std::pair< Point<dim>, typename DoFHandler<dim>::active_cell_iterator > > cell_ordering;
	cell_ordering.reserve(n_elements);

	/*
	 * push center and iterator of each cell into the array which will be sorted
	 */
	typename DoFHandler<dim>::active_cell_iterator
	cell = dof_handler.begin_active(),
	endc = dof_handler.end();
	unsigned int i=0;
	for (; cell!=endc; ++cell, ++i)
		cell_ordering.push_back(std::make_pair(cell->center(), cell));

	for (unsigned int derivative_dir = 0; derivative_dir < dim; ++derivative_dir)
	{
		/*
		 * create an object of type Compare initialized with the current direction,
		 * which will determine the relation by which the array will be sorted..
		 */
		Compare<dim, typename DoFHandler<dim>::active_cell_iterator> comp (derivative_dir);

		/*
		 * ..and sort it
		 */
		std::sort(cell_ordering.begin(), cell_ordering.end(), comp);

		/*
		 * store the sorted iterators in the designated array
		 */
		sorted_iters[derivative_dir].resize(n_elements);

//		Point<dim> former_point = cell_ordering[0].first;

// 		std::cout << std::endl << std::endl;
		for (unsigned int i=0; i < n_elements; ++i)
		{
			// print sorted cell centers corresponding to current derivative direction.
// 			std::cout << derivative_dir << " -- " << cell_ordering[i].first << " -- " << i%32<< std::endl;
			sorted_iters[derivative_dir][i] = cell_ordering[i].second;
		}
// 		std::cout << std::endl << std::endl;

		/*
		 *----------------------------------------------------------------------
		 * Compute the number of cells for each row of cells
		 *----------------------------------------------------------------------
		 *
		 * this is needed to determine the number of cells per row in current
		 * direction. It is the coordinate of the cell-midpoints that should re-
		 * main constant while traversing along the current direction.
		 * Important: This direction has to coincide with the second direction
		 * corresponding to which the Compare object sorts.
		 */
		unsigned int const_coord_dir = (derivative_dir+(dim-1-1))%dim;

		/*
		 * initialize stuff
		 */
		n_el_per_row[derivative_dir].reserve(n_elements);
		double coord = 0;
		unsigned int current_n_el_per_row = 0;

		/*
		 * determine the number of cells for each row by traversing the sorted
		 * cells and counting the cells until the coordinate, that should be
		 * constant along the row, of the current cell differs from the one before,
		 * i.e. this cell belongs to a new row.
		 */

		std::vector< double > coords;
		coords.reserve(n_elements);
		for (unsigned int i=0; i < n_elements; ++i)
		{
			/*
			 * set current cell iterator
			 */
			cell = sorted_iters[derivative_dir][i];
			
// 			// print sorted cell centers corresponding to current derivative direction.
// 			std::cout << derivative_dir << " -- " << cell->center() << " -- " << i%32<< std::endl;

//			std::cout << cell->center()(const_coord_dir) << std::endl;

			/*
			 * compare the current cell-center with the one before (stored in
			 * coord) and increment the number of cells in this row by one if
			 * they coincide (except if it is the last cell). If they
			 * differ (or the current cell is the last one), we store the
			 * number of cells in this row in the designated vector. In case
			 * of the last cell, we beforehand increment the number of cells
			 * in the current row, since we do the incrementing at the end
			 * of each step.
			 */
			if (i == 0)
			{
				coord = cell->center()(const_coord_dir);
				coords.push_back(coord);
				++current_n_el_per_row;
			}
			else if (fabs(coord - cell->center()(const_coord_dir)) < 1e-14 && i!= n_elements-1)
				++current_n_el_per_row;
			else
			{
				if(i == n_elements-1)
					++current_n_el_per_row;
				n_el_per_row[derivative_dir].push_back(current_n_el_per_row);
				coord = cell->center()(const_coord_dir);
				coords.push_back(coord);
				current_n_el_per_row = 1;
			}
		}

		n_el_per_row[derivative_dir].shrink_to_fit();

		n_rows[derivative_dir] = n_el_per_row[derivative_dir].size();

		coords.shrink_to_fit();

//		std::cout << "coords.size() = " << coords.size() << std::endl << std::endl;

//		for (unsigned int ii=0; ii<coords.size(); ++ii)
//			std::cout << coords[ii] << std::endl;
//		std::cout << std::endl << std::endl;

//		std::cout << derivative_dir << " -- coords.size() = " << coords.size() << std::endl << std::endl;
	}
//	std::cout << std::endl << std::endl;
}


template<int dim>
void ADIGeometryInfo<dim>::compute_transformation (
		std::vector< unsigned int >	&transform,
		const DoFHandler< dim >		&dof_handler,
		const unsigned int			ADI_direction
		) const
{
	unsigned int n_dofs				= dof_handler.n_dofs(),
			 	 n_dofs_comp		= n_dofs/n_components;

	Assert(transform.size() == n_dofs, ExcInternalError());

	std::vector< unsigned int > retransform_cell(n_dofs),
								transform_cell(n_dofs),
								retransform(n_dofs),
								transform_tmp(n_dofs),
								retransform_tmp(n_dofs),
								transform_comp(n_dofs),
								retransform_comp(n_dofs),
								get_component(n_dofs),
								active_components(2);

	unsigned int index=0;

	index = 0;

	compute_component_wise(transform_comp, retransform_comp, dof_handler);

	for (unsigned int i=0; i < n_dofs; ++i)
		get_component[i] = floor(transform_comp[i]/n_dofs_comp);

	for (unsigned int direction = 0; direction < dim; ++direction)
	{
		unsigned int index_local = 0;

		for (unsigned int component = 0; component < n_components; ++component)
			if (get_derivative_direction(ADI_direction, component) == direction)
			{
				active_components.at(index_local) = component;
				++index_local;
			}
		index_local = 0;

		DoFRenumbering::compute_cell_wise(transform_cell, retransform_cell, dof_handler, sorted_iters[direction]);

		std::vector< unsigned int > get_cell_nr(n_dofs);

		for (unsigned int i=0; i < n_dofs; ++i)
			get_cell_nr[i] = floor(transform_cell[i]/dofs_per_cell);

		std::vector< unsigned int > offset(n_elements);

		for (unsigned int element = 0; element < n_elements; ++element)
			offset[element] = 0;

		for (unsigned int element = 0; element < n_elements; ++element)
			for (unsigned int component = 0; component < n_components; ++component)
				for (unsigned int j=0; j < comp_dofs_per_cell; ++j, ++index_local)
					retransform_tmp.at(get_cell_nr[index_local]*dofs_per_cell + get_component[index_local]*comp_dofs_per_cell + j) = index_local;

		for (unsigned int element=0; element < n_elements; ++element)
			for (unsigned int component=0; component < 2; ++component)
				for (unsigned int j=0; j < comp_dofs_per_cell; ++j, ++index)
					retransform.at(index) = retransform_tmp.at(element*dofs_per_cell + active_components.at(component)*comp_dofs_per_cell + j);
	}

	for (unsigned int i=0; i < n_dofs; ++i)
		transform[retransform[i]] = i;
}


template<int dim>
void ADIGeometryInfo<dim>::compute_component_wise (
		std::vector< unsigned int >	&transform,
		std::vector< unsigned int >	&retransform,
		const DoFHandler< dim >		&dof_handler
		) const
{
	typename DoFHandler<dim,dim>::active_cell_iterator start = dof_handler.begin_active();
	const typename DoFHandler<dim,dim>::level_cell_iterator end = dof_handler.end();

	const types::global_dof_index result =
		DoFRenumbering::compute_component_wise<dim,dim,
		typename DoFHandler<dim,dim>::active_cell_iterator,
		typename DoFHandler<dim,dim>::level_cell_iterator>
		(transform, start, end, std::vector<unsigned int>(), false);

	if (result == 0) return;

	for (unsigned int i=0; i < retransform.size(); ++i)
		retransform[transform[i]] = i;
}


template<int dim>
unsigned int ADIGeometryInfo<dim>::get_derivative_direction (
		const unsigned int	ADI_direction,
		const unsigned int component
		) const
{
	unsigned int derivative_direction(0);

	switch (ADI_direction)
	{
	case 1:
		switch (component)
		{
		case 0:
			derivative_direction = 1;
			break;
		case 1:
			derivative_direction = 2;
			break;
		case 2:
			derivative_direction = 0;
			break;
		case 3:
			derivative_direction = 2;
			break;
		case 4:
			derivative_direction = 0;
			break;
		case 5:
			derivative_direction = 1;
			break;
		}
		break;
	case 0:
		switch (component)
		{
		case 0:
			derivative_direction = 2;
			break;
		case 1:
			derivative_direction = 0;
			break;
		case 2:
			derivative_direction = 1;
			break;
		case 3:
			derivative_direction = 1;
			break;
		case 4:
			derivative_direction = 2;
			break;
		case 5:
			derivative_direction = 0;
			break;
		}
		break;

	}

	return derivative_direction;
}


template<int dim>
unsigned int ADIGeometryInfo<dim>::get_coupled_component (
		const unsigned int	ADI_direction,
		const unsigned int	component
		) const
{
	unsigned int coupled_component(0);

	switch (ADI_direction)
	{
	case 0:
		switch (component)
		{
		case 0:
			coupled_component = 2;
			break;
		case 1:
			coupled_component = 0;
			break;
		case 2:
			coupled_component = 1;
			break;
		case 3:
			coupled_component = 1;
			break;
		case 4:
			coupled_component = 2;
			break;
		case 5:
			coupled_component = 0;
			break;
		}
		break;
	case 1:
		switch (component)
		{
		case 0:
			coupled_component = 1;
			break;
		case 1:
			coupled_component = 2;
			break;
		case 2:
			coupled_component = 0;
			break;
		case 3:
			coupled_component = 2;
			break;
		case 4:
			coupled_component = 0;
			break;
		case 5:
			coupled_component = 1;
			break;
		}
		break;

	}

	return coupled_component;
}


template<int dim>
std::pair< unsigned int, unsigned int > ADIGeometryInfo<dim>::get_index_derivative_pair (
		const unsigned int	ADI_direction,
		const unsigned int	component
		) const
{
	return std::make_pair(get_coupled_component(ADI_direction, component), get_derivative_direction(ADI_direction, component));
}


template<int dim>
unsigned int ADIGeometryInfo<dim>::get_dofs_per_cell () const
{
	return dofs_per_cell;
}


template<int dim>
unsigned int ADIGeometryInfo<dim>::get_comp_dofs_per_cell () const
{
	return comp_dofs_per_cell;
}


template<int dim>
unsigned int ADIGeometryInfo<dim>::get_n_elements () const
{
	return n_elements;
}


template<int dim>
unsigned int ADIGeometryInfo<dim>::get_n_el_per_row (
		const unsigned int direction,
		const unsigned int matrix_nr
		) const
{
	return n_el_per_row[direction][matrix_nr];
}


template<int dim>
unsigned int ADIGeometryInfo<dim>::get_n_components () const
{
	return n_components;
}


template<int dim>
unsigned int ADIGeometryInfo<dim>::get_n_prototypes () const
{
	return n_prototypes;
}


template<int dim>
unsigned int ADIGeometryInfo<dim>::get_n_rows (
		const unsigned int direction
		) const
{
	return n_rows[direction];
}


template<int dim>
unsigned int ADIGeometryInfo<dim>::get_prototype (
		const unsigned int i
		) const
{
	return prototypes[i];
}


template<int dim>
unsigned int ADIGeometryInfo<dim>::get_sparsity_pattern_nr (
		const unsigned int direction,
		const unsigned int matrix_nr
		) const
{
	return matrix2sparsity[direction][matrix_nr];
}

// ---------------------------------------------------------------------------------
//
//	Functions useful for investigating the ordering of the DoF
//
// ---------------------------------------------------------------------------------

template <int dim>
xcoord<dim>::xcoord()
	:
	Function<dim>(dim)
{}

template <int dim>
double xcoord<dim>::value(const Point<dim>	&p) const
{
//     Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
//     Assert (dim >= 2, ExcNotImplemented());

    return p[0];
}

template <int dim>
void xcoord<dim>::vector_value(const Point<dim>	&p,
							   Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());
	
	double tmp_value(0);
	tmp_value = value(p);

	values(0) = tmp_value;
    values(1) = tmp_value;
    values(2) = tmp_value;

	values(3) = tmp_value;
    values(4) = tmp_value;
    values(5) = tmp_value;
}

// template <int dim>
// void xcoord<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
// 									std::vector< Vector<double> >	&value_list) const
// {
// 	  Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));
// 
// 	  const unsigned int n_points = points.size();
// 
// 	  for (unsigned int p=0; p<n_points; ++p)
// 		  xcoord<dim>::vector_value (points[p], value_list[p]);
// }

template <int dim>
ycoord<dim>::ycoord()
	:
	Function<dim>(dim)
{}

template <int dim>
double ycoord<dim>::value(const Point<dim>	&p) const
{
//     Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
//     Assert (dim >= 2, ExcNotImplemented());

    return p[1];
}

template <int dim>
void ycoord<dim>::vector_value(const Point<dim>	&p,
							   Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());
	
	double tmp_value(0);
	tmp_value = value(p);

	values(0) = tmp_value;
    values(1) = tmp_value;
    values(2) = tmp_value;

	values(3) = tmp_value;
    values(4) = tmp_value;
    values(5) = tmp_value;
}

// template <int dim>
// void ycoord<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
// 									std::vector< Vector<double> >	&value_list) const
// {
// 	  Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));
// 
// 	  const unsigned int n_points = points.size();
// 
// 	  for (unsigned int p=0; p<n_points; ++p)
// 		  ycoord<dim>::vector_value (points[p], value_list[p]);
// }

template <int dim>
zcoord<dim>::zcoord()
	:
	Function<dim>(dim)
{}

template <int dim>
double zcoord<dim>::value(const Point<dim>	&p) const
{
//     Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
//     Assert (dim >= 2, ExcNotImplemented());

    return p[2];
}

template <int dim>
void zcoord<dim>::vector_value(const Point<dim>	&p,
							   Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());
	
	double tmp_value(0);
	tmp_value = value(p);

	values(0) = tmp_value;
    values(1) = tmp_value;
    values(2) = tmp_value;

	values(3) = tmp_value;
    values(4) = tmp_value;
    values(5) = tmp_value;
}

template <int dim>
component<dim>::component()
	:
	Function<dim>(dim)
{}

// template <int dim>
// double component<dim>::value(const Point<dim>	&p) const
// {
//     Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
//     Assert (dim >= 2, ExcNotImplemented());
// 
//     return -1;
// }

template <int dim>
void component<dim>::vector_value(const Point<dim>	&p,
							   Vector<double>	&values) const
{
    Assert (values.size() == dim, ExcDimensionMismatch (values.size(), dim));
    Assert (dim >= 2, ExcNotImplemented());

	values(0) = 0;
    values(1) = 1;
    values(2) = 2;

	values(3) = 3;
    values(4) = 4;
    values(5) = 5;
}

// template <int dim>
// void component<dim>::vector_value_list(const std::vector< Point<dim> >	&points,
// 									std::vector< Vector<double> >	&value_list) const
// {
// 	  Assert (value_list.size() == points.size(), ExcDimensionMismatch (value_list.size(), points.size()));
// 
// 	  const unsigned int n_points = points.size();
// 
// 	  for (unsigned int p=0; p<n_points; ++p)
// 		  component<dim>::vector_value (points[p], value_list[p]);
// }

template class xcoord<3>;
template class ycoord<3>;
template class zcoord<3>;
template class component<3>;

template class ADIGeometryInfo<3>;

} /* namespace MaxwellProblem */
