#ifndef ADIGEOMETRYINFO_H_
#define ADIGEOMETRYINFO_H_

#include <set>
#include <vector>


#include <deal.II/base/timer.h>
#include <deal.II/dofs/dof_renumbering.h>
#include <deal.II/dofs/dof_accessor.h>

#include <deal.II/base/point.h>
#include <deal.II/base/function.h>
#include <deal.II/lac/vector.h>

#include "Compare.h"

namespace MaxwellProblem {

using namespace dealii;


/*
 * This class is used to gather basic geometric information about the
 * mesh needed by the ADIMatrix and ADISparsityPattern classes. It also
 * collects relations between those classes needed to initialize and
 * link them against each other correspondingly.
 */

template <int dim>
class ADIGeometryInfo
{
public:
																			ADIGeometryInfo				();
	void																	initialize					(const DoFHandler<dim>			&dof_handler);
	void																	compute_transformation		(std::vector< unsigned int >	&transform,
																										 const DoFHandler< dim >		&dof_handler,
																										 const unsigned int				ADI_direction) const;

	unsigned int															get_derivative_direction	(const unsigned int				ADI_direction,
																										 const unsigned int				component) const;
	unsigned int															get_coupled_component		(const unsigned int				ADI_direction,
																										 const unsigned int				component) const;
	std::pair< unsigned int, unsigned int >									get_index_derivative_pair	(const unsigned int				ADI_direction,
																										 const unsigned int				component) const;
	unsigned int															get_dofs_per_cell			() const;
	unsigned int															get_comp_dofs_per_cell		() const;
	unsigned int															get_n_elements				() const;
	unsigned int															get_n_el_per_row			(const unsigned int				direction,
																										 const unsigned int				matrix_nr) const;
	unsigned int															get_n_components			() const;
	unsigned int															get_n_prototypes			() const;
	unsigned int															get_n_rows					(const unsigned int				direction) const;
	unsigned int															get_prototype				(const unsigned int				i) const;
	unsigned int															get_sparsity_pattern_nr		(const unsigned int				direction,
																								 	 	 const unsigned int				matrix_nr) const;

private:
	void																	compute_reordering			(const DoFHandler< dim >		&dof_handler);
	void																	compute_component_wise		(std::vector< unsigned int >	&transform,
																										 std::vector< unsigned int >	&retransform,
																										 const DoFHandler< dim >		&dof_handler) const;

private:
	unsigned int 																	dofs_per_cell,
																					comp_dofs_per_cell,
																					n_elements,
																					n_components,
																					n_prototypes;

	std::vector< unsigned int >														prototypes,
																					n_rows;


	std::vector< std::vector<unsigned int> >										n_el_per_row,
																					matrix2sparsity;

	std::vector< std::vector< typename DoFHandler<dim>::active_cell_iterator > >	sorted_iters;
};


template <int dim>
class xcoord : public Function<dim>
{
public:
	xcoord		 			();
	double value			(const Point<dim>				&p) const;
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
// 	void vector_value_list	(const std::vector<Point<dim> >	&points,
// 							 std::vector<Vector<double> >	&value_list) const;
};

template <int dim>
class ycoord : public Function<dim>
{
public:
	ycoord		 			();
	double value			(const Point<dim>				&p) const;
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
// 	void vector_value_list	(const std::vector<Point<dim> >	&points,
// 							 std::vector<Vector<double> >	&value_list) const;
};

template <int dim>
class zcoord : public Function<dim>
{
public:
	zcoord		 			();
	double value			(const Point<dim>				&p) const;
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
// 	void vector_value_list	(const std::vector<Point<dim> >	&points,
// 							 std::vector<Vector<double> >	&value_list) const;
};

template <int dim>
class component : public Function<dim>
{
public:
	component		 			();
	void vector_value		(const Point<dim>				&p,
					 	 	 Vector<double>					&values) const;
// 	void vector_value_list	(const std::vector<Point<dim> >	&points,
// 							 std::vector<Vector<double> >	&value_list) const;
};

} /* namespace MaxwellProblem */

#endif /* ADIGEOMETRYINFO_H_ */
