#include "Leapfrog.h"

namespace MaxwellProblem {
namespace TimeIntegration{


template<typename MassMatrixtype, typename CurlMatrixtype, typename Vectortype>
Leapfrog<MassMatrixtype, CurlMatrixtype, Vectortype>::Leapfrog(
			const MassMatrixtype &MH_inv,
			const MassMatrixtype &ME_inv,
			const CurlMatrixtype &CH,
			const CurlMatrixtype &CE,
			double timestep)
:
	MH_inverse(&MH_inv),
	ME_inverse(&ME_inv),
	CH(&CH),
	CE(&CE),
	timestep(timestep),
	initial_step(true)
{}


template<typename MassMatrixtype, typename CurlMatrixtype, typename Vectortype>
void Leapfrog<MassMatrixtype, CurlMatrixtype, Vectortype>::integrate_step(
			Vectortype &H,
			Vectortype &E,
			const Vectortype &j_current)
{
	/*
	 * First half step for H
	 */
	if (initial_step)
	{
		/*
		 * Only performed during the first time step or after changing the time step.
		 */
		tmp.reinit(H);
		H_half.reinit(H);

		/*
		 * Computation of the first half step
		 */
		(*CE).vmult(tmp, E);
		(*MH_inverse).vmult(H_half,tmp);
		H_half.sadd(-1. * timestep/2.0, H);

		initial_step = false;
	}
	else 
		H_half.sadd(-1., 2., H);

	/*
	 * Full step for E
	 */
	(*CH).vmult(tmp, H_half);
// 	tmp*=timestep;
	tmp.sadd(timestep, -1. * timestep, j_current);
	(*ME_inverse).vmult_add(E,tmp);

	/*
	 * Second half step for H
	 */
	(*CE).vmult(tmp, E);
	(*MH_inverse).vmult(H,tmp);
	H.sadd(-1. * timestep/2.0, H_half);
}


template<typename MassMatrixtype, typename CurlMatrixtype, typename Vectortype>
void Leapfrog<MassMatrixtype, CurlMatrixtype, Vectortype>::integrate_step(
			Vectortype &H,
			Vectortype &E)
{
	/*
	 * First half step for H
	 */
	if (initial_step)
	{
		/*
		 * Only performed during the first time step or after changing the time step.
		 */
		tmp.reinit(H);
		H_half.reinit(H);

		/*
		 * Computation of the first half step
		 */
		(*CE).vmult(tmp, E);
		(*MH_inverse).vmult(H_half,tmp);
		H_half.sadd(-1. * timestep/2.0, H);

		initial_step = false;
	}
	else 
		H_half.sadd(-1., 2., H);

	/*
	 * Full step for E
	 */
	(*CH).vmult(tmp, H_half);
	tmp*=timestep;
	(*ME_inverse).vmult_add(E,tmp);

	/*
	 * Second half step for H
	 */
	(*CE).vmult(tmp, E);
	(*MH_inverse).vmult(H,tmp);
	H.sadd(-1. * timestep/2.0, H_half);
}


template<typename MassMatrixtype, typename CurlMatrixtype, typename Vectortype>
void Leapfrog<MassMatrixtype, CurlMatrixtype, Vectortype>::set_timestep(double timestep)
{
	this->timestep = timestep;

	initial_step = true;
}


template class Leapfrog<SparseMatrix<double>,SparseMatrix<double>,Vector<double>>;
template class Leapfrog<BlockSparseMatrix<double>,BlockSparseMatrix<double>,BlockVector<double>>;

} /* namespace TimeIntegration */
} /* namespace MaxwellProblem */
