#ifndef LEAPFROG_H_
#define LEAPFROG_H_


#include <deal.II/lac/vector.h>
#include <deal.II/lac/block_vector.h>

#include <deal.II/lac/sparse_matrix.h>
#include <deal.II/lac/block_sparse_matrix.h>


using namespace dealii;

namespace MaxwellProblem {
namespace TimeIntegration{

/*
 * Implementation of the leapfrog (or Verlet) scheme for linear Maxwells' equations
 *
 * This class implements the leapfrog integration for the following system of
 * differential equations:
 *
 *  M_H \dot{H} = - C_E  E
 *  M_E \dot{E} =   C_H  H.
 *
 * The method looks as follows:
 *
 * H_{n+1/2} =    H_n    - M_H^{-1} * \tau/2 *  C_E  E_n
 * E_{n+1}   =    E_n    + M_E^{-1} * \tau   *  C_H  H_{n+1/2}
 * H_{n+1}   = H_{n+1/2} - M_H^{-1} * \tau/2 *  C_E  E_{n+1}
 *
 *
 * M_H^{-1}, M_E^{-1}, C_H and C_E can be of type SparseMatrix or BlockSparseMatrix.
 *
 * Since C_H = C_E^T the class could restrict itself to using C_E. However, Tvmult is not parallelized
 * in deal.ii, so both of them are required.
 */


template<typename MassMatrixtype, typename CurlMatrixtype, typename Vectortype=Vector<double>>
class Leapfrog
{
public:
	Leapfrog(
			const MassMatrixtype &MH_inv,
			const MassMatrixtype &ME_inv,
			const CurlMatrixtype &CH,
			const CurlMatrixtype &CE,
			double timestep);

	void integrate_step(
			Vectortype &H,
			Vectortype &E);

	void integrate_step(
			Vectortype &H,
			Vectortype &E,
			const Vectortype &j_current);

	void set_timestep(double timestep);

private:
	const MassMatrixtype* MH_inverse;
	const MassMatrixtype* ME_inverse;
	const CurlMatrixtype* CH;
	const CurlMatrixtype* CE;

	double timestep;

	Vectortype H_half;
	Vectortype tmp;

	bool initial_step;

};

} /* namespace TimeIntegration */
} /* namespace MaxwellProblem */

#endif /* LEAPFROG_H_ */
