#include "PeacemanRachford.h"

namespace MaxwellProblem {
namespace TimeIntegration {

template <int dim>
PeacemanRachford<dim>::PeacemanRachford(
		double theta
		)
:
		theta(theta)
{}


template <int dim>
void PeacemanRachford<dim>::assemble_explicit_matrix(
		ADIMatrix<dim>					&explicit_matrix,
		const ADIMatrix<dim>				&stiffness_matrix,
		const ADIMassMatrix<dim>			&mass_matrix,
		const std::vector<unsigned int>	&transform,
		double 							timestep
		) const
{
	/*
	 * This routine takes one of the stiffness matrices and the mass matrix
	 * and assembles the explicit system matrix out of them, which is used
	 * to compute the system rhs.
	 * In this version, the result is stored in the other ADIMatrix passed
	 * to the function as explicit_matrix.
	 */

	explicit_matrix.copy_from (theta*timestep, stiffness_matrix);
	explicit_matrix.add (mass_matrix, transform);
}


template <int dim>
void PeacemanRachford<dim>::assemble_explicit_matrix(
		ADIMatrix<dim>					&explicit_matrix,
		const ADIMassMatrix<dim>			&mass_matrix,
		const std::vector<unsigned int>	&transform,
		double 							timestep
		) const
{
	/*
	 * This routine takes one of the stiffness matrices and the mass matrix
	 * and assembles the explicit system matrix out of them, which is used
	 * to compute the system rhs.
	 * In this version, the ADIMatrix-parameter explicit_matrix should
	 * contain the stiffness matrix to use, which will then be overriden by
	 * the explicit system matrix.
	 */

	explicit_matrix *= theta*timestep;
	explicit_matrix.add (mass_matrix, transform);
}


template <int dim>
void PeacemanRachford<dim>::assemble_implicit_matrix(
		ADIMatrix<dim>					&implicit_matrix,
		const ADIMatrix<dim>				&stiffness_matrix,
		const ADIMassMatrix<dim>			&mass_matrix,
		const std::vector<unsigned int>	&transform,
		double 							timestep
		) const
{
	/*
	 * This routine takes one of the stiffness matrices and the mass matrix
	 * and assembles the implicit system matrix out of them.
	 * In this version, the result is stored in the other ADIMatrix passed
	 * to the function as implicit_matrix.
	 */

	implicit_matrix.copy_from (-(1-theta)*timestep, stiffness_matrix);
	implicit_matrix.add (mass_matrix, transform);
}


template <int dim>
void PeacemanRachford<dim>::assemble_implicit_matrix(
		ADIMatrix<dim>					&implicit_matrix,
		const ADIMassMatrix<dim>		&mass_matrix,
		const std::vector<unsigned int>	&transform,
		double 							timestep
		) const
{
	/*
	 * This routine takes one of the stiffness matrices and the mass matrix
	 * and assembles the implicit system matrix out of them.
	 * In this version, the ADIMatrix-parameter implicit_matrix should
	 * contain the stiffness matrix to use, which will then be overriden by
	 * the implicit system matrix.
	 */

	implicit_matrix *= -(1-theta)*timestep;
	implicit_matrix.add (mass_matrix, transform);
}


template <int dim>
void PeacemanRachford<dim>::assemble_system_rhs(
		Vector<double> &system_rhs,
		const ADIMatrix<dim> &explicit_matrix,
		const Vector<double> &rhs,
		const Vector<double> &old_solution,
		const std::vector<unsigned int> &transform,
		double timestep
		) const
{
	/*
	 * This routine takes the explicit system matrix (of the current
	 * halfstep), the solution of the last halfstep and the rhs to compute
	 * the system rhs for this halfstep.
	 */

	explicit_matrix.vmult(system_rhs, old_solution, transform);
	system_rhs.add(-timestep,rhs);
}

template class PeacemanRachford<3>;

} /* namespace TimeIntegration */
} /* namespace MaxwellProblem */
