#ifndef PEACEMANRACHFORD_H_
#define PEACEMANRACHFORD_H_

#include <deal.II/base/timer.h>
#include <deal.II/numerics/vector_tools.h>

#include <vector>

#include "../data_types/ADIMassMatrix.h"
#include "../data_types/ADIMatrix.h"

namespace MaxwellProblem {
namespace TimeIntegration{

using namespace dealii;

/*
 * This class is used to assemble the matrices
 * used in the ADI time integration. The parameter theta controls, on
 * which time integration method the AD-Scheme is based, with the default
 * theta, being 0.5, leading to the Peaceman-Rachford method (based on the
 * Crank-Nicolson-Scheme).
 * In general, the AD-Scheme is based on the Theta-Scheme, i.e.,
 * theta = 0 corresponds to the implicit, theta = 1 to the explicit
 * Euler-Scheme.
 *
 * The Peaceman-Rachford Scheme is given by:
 *		(M - \tau/2 * A) u^{n+1/2}	=	(M + \tau/2 * B) u^{n}
 *		(M - \tau/2 * B) u^{n+1}	=	(M + \tau/2 * A) u^{n+1/2},
 *
 * and the more generel theta AD-Scheme is given by:
 *		(M - (1-\theta) * \tau * A) u^{n+1/2}	=	(M + \theta * \tau * B) u^{n}
 *		(M - (1-\theta) * \tau * B) u^{n+1}		=	(M + \theta * \tau * A) u^{n+1/2},
 *
 * where theta is the parameter described above given to the constructor,
 * tau is the timestep, A and B are the stiffness matrices of the two sub-
 * problems and M is the mass matrix.
 * In the following, if we talk of the system rhs, we mean something like
 * (M + \theta * \tau/2 * A) u, where u is the solution  of the last halfstep
 * and A is the appropriate stiffness matrix (which depends on the halfstep
 * we're in). Finally, if we talk about the (implicit) system matrix, we mean
 * something like (M - (1-\theta) * \tau * A), where again, A is the stiffness
 * matrix of the appropriate subproblem.
 */

template <int dim>
class PeacemanRachford
{
public:
			PeacemanRachford				(double 							theta = 0.5);

	void 	assemble_system_matrix_and_rhs	(ADIMatrix<dim>						&system_matrix,
											 Vector<double>						&system_rhs,
											 const ADIMatrix<dim>				&impl_matrix,
											 const ADIMatrix<dim>				&expl_matrix,
											 const ADIMassMatrix<dim>			&mass_matrix,
											 const Vector<double>				&rhs,
											 const Vector<double>				&old_solution,
											 const std::vector<unsigned int>	&transform,
											 double								timestep,
											 Timer								&copy_timer,
											 Timer								&vmult_timer,
											 Timer								&add_timer) const;

	void 	assemble_system_matrix_and_rhs	(ADIMatrix<dim>						&system_matrix,
										 	 Vector<double>						&system_rhs,
										 	 const ADIMatrix<dim>				&impl_matrix,
										 	 const ADIMatrix<dim>				&expl_matrix,
										 	 const ADIMassMatrix<dim>			&mass_matrix,
										 	 const Vector<double>				&rhs,
										 	 const Vector<double>				&old_solution,
										 	 const std::vector<unsigned int>	&transform,
										 	 double								timestep) const;

	void	assemble_explicit_matrix		(ADIMatrix<dim>						&explicit_matrix,
		 	 	 	 	 	 	 	 	 	 const ADIMatrix<dim>				&stiffness_matrix,
										 	 const ADIMassMatrix<dim>			&mass_matrix,
										 	 const std::vector<unsigned int>	&transform,
										 	 double								timestep) const;

	void	assemble_explicit_matrix		(ADIMatrix<dim>						&explicit_matrix,
		 	 	 	 	 	 	 	 	 	 const ADIMassMatrix<dim>			&mass_matrix,
										 	 const std::vector<unsigned int>	&transform,
										 	 double								timestep) const;

	void	assemble_implicit_matrix		(ADIMatrix<dim>						&implicit_matrix,
		 	 	 	 	 	 	 	 	 	 const ADIMatrix<dim>				&stiffness_matrix,
										 	 const ADIMassMatrix<dim>			&mass_matrix,
										 	 const std::vector<unsigned int>	&transform,
										 	 double								timestep) const;

	void	assemble_implicit_matrix		(ADIMatrix<dim>						&implicit_matrix,
		 	 	 	 	 	 	 	 	 	 const ADIMassMatrix<dim>			&mass_matrix,
										 	 const std::vector<unsigned int>	&transform,
										 	 double								timestep) const;

	void	assemble_system_rhs				(Vector<double>						&system_rhs,
		 	 	 	 	 	 	 	 	 	 const ADIMatrix<dim>				&explicit_matrix,
										 	 const Vector<double>				&rhs,
										 	 const Vector<double>				&old_solution,
										 	 const std::vector<unsigned int>	&transform,
										 	 double								timestep) const;
private:
	double	theta;
};

} /* namespace TimeIntegration */
} /* namespace MaxwellProblem */

#endif /* PEACEMANRACHFORD_H_ */
